﻿//      WaterSimDCDC Regional Water Demand and Supply Model Version 5.0

//       This is support classes for the C# WaterSim API

//       Copyright (C) 2011 , The Arizona Board of Regents
//              on behalf of Arizona State University

//       All rights reserved.

//       Developed by the Decision Center for a Desert City
//       Lead Model Development - David A. Sampson <david.a.sampson@asu.edu>

//       This program is free software: you can redistribute it and/or modify
//       it under the terms of the GNU General Public License version 3 as published by
//       the Free Software Foundation.

//       This program is distributed in the hope that it will be useful,
//       but WITHOUT ANY WARRANTY; without even the implied warranty of
//       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//       GNU General Public License for more details.

//       You should have received a copy of the GNU General Public License
//       along with this program.  If not, please see <http://www.gnu.org/licenses/>.
//
//====================================================================================
using System;
using System.Collections.Generic;

/***************************************************************
 * Water Sim Utilities 
 * Version 11
 * 
 * Keeper Ray Quay
 * Version 7 - Modified ModelPArameter access to underlying routines to access model, added documentation
 * Version 8 - Added Process Manager
 * ************************************************************/
namespace WaterSimDCDC
{
    enum ColoradoRiverRecord { eUnspecified, ePaleoRecord, eBureauRecord, eUserSupplied };


    /*********************************************************
     *   Provider Classes, Enums
     * 
     * 
     * **************************************************************/
    #region provider

    // Speical Range Check
    delegate bool DcheckBase(int value, ref string errorstr, ModelParameterClass aModelParameter);
    delegate bool DcheckProvider(int value, eProvider ep, ref string errorstr, ModelParameterClass aModelParameter);
    // get int delegate
    delegate int Dget();
    // Set int delegate
    delegate void Dset(int value);
    // Getarray delegate
     delegate int[] Dgetarray();
    // Setarayy delegate
     delegate void Dsetarray(int[] value);
    //-------------------------------------------------------------------------------------
    // This is the class used to provde and indexed property to wrap around the provider input array parammeters
    /// <summary>
     /// Base class for All Provider Model Parameters
    /// </summary>
     /// <remarks>, This is the class used to provde and indexed property to wrap around the provider input array parammeters.  This class provides direct access to the Fortan model Parameters, with a layer of rangechacking that throws exception when values violate range check rules.  The constructor for this class is hidden.  Parameters of this class are created by the WaterSim class and managed by the WaterSim.ParamManager, thusthe constructor is hidden.</remarks>
    public class providerArrayProperty
    {
        // get data references
        Dgetarray Fget;
        Dsetarray Fset;
        ParameterManagerClass Fpm;
        eModelParam Femp;
        //------------------
        internal providerArrayProperty(ParameterManagerClass pm,  eModelParam modelparam, Dgetarray getcall, Dsetarray setcall) : base()
        {
            // do tests
            if ((pm==null)|(getcall==null)|(setcall==null)) throw new WaterSim_Exception(WS_Strings.wsdbNullParameter);
            if (!ModelParamClass.valid(modelparam)) throw new WaterSim_Exception(WS_Strings.wsInvalid_EModelPAram); 
            Fget = getcall;
            Fset = setcall;
            Fpm= pm;
            Femp = modelparam;
        }
        //----------------------
        /// <summary>
        /// indexer for indexed array gets and sets based on eProvider enum values
        /// </summary>
        /// <param name="index">eProvider index</param>
        /// <value>indexed Model Parameter value</value>
        /// <returns>value in provider array for eProvider index</returns>
        /// <exception cref="WaterSim_Exception">if set and value viloates a range rule</exception>
        public int this[eProvider index]
        {
            get
            {
                if (!ProviderClass.valid(index)) throw new WaterSim_Exception(WS_Strings.wsInvalidProvider);
                int[] temp = Fget();
                return temp[(int)index];
            }
            set
            {
                if (!ProviderClass.valid(index)) throw new WaterSim_Exception(WS_Strings.wsInvalidProvider);
                string errMessage = "";
                if (Fpm.CheckProviderValueRange(Femp,value, (eProvider)index,ref errMessage))
                {
                    int[] temp = Fget();
                    temp[(int)index] = value;
                    Fset(temp);
                }
                else throw new WaterSim_Exception(errMessage);
            }
        }
        /// <summary>
        /// indexer for indexed array gets and sets
        /// </summary>
        /// <param name="index">index to array</param>
        /// <value>indexed Model Parameter value</value>
        /// <returns>value in provider array for index</returns>
        /// <exception cref="WaterSim_Exception">if set and value viloates a range rule</exception>
        public int this[int index]
        {
            get
            {
                return this[(eProvider)index];
            }
            set
            {
                this[(eProvider)index] = value;
            }
        }        //----------------------
        /// <summary>
        /// Gets Model Parameter Array values 
        /// </summary>
        /// <returns>an Array of Model Values</returns>
        public ProviderIntArray getvalues()
       // public int[] getvalues()
        {
            ProviderIntArray pia = new ProviderIntArray(0);
            pia.Values = Fget();
            return pia;
            //return Fget();
        }
        //----------------------
        /// <summary>
        /// Set Model Parameter array using value array
        /// </summary>
        /// <param name="value">values to set Model Parameter</param>
        /// <exception cref="WaterSim_Exception">if any numbers in value violate a range rule</exception>
        public void setvalues(ProviderIntArray value)
        {
            Fpm.CheckProviderValuesRange(Femp, value);
            Fset(value.Values);
        }
    }
    //---------------------------------------------------------------------------------
    // 
    /// <summary>
    /// A readonly version of providerArrayProperty class
    /// </summary>
    /// <remarks>This is the class used to provde and indexed property to wrap around the provider output array parammeters.  Access to model paramters is designed to be fast.</remarks>
    /// <seealso cref="providerArrayProperty"/>
    public class providerArrayProperty_readonly
    {

        // Get Delegates
        Dgetarray Fget;
        //----------------------
        internal providerArrayProperty_readonly(Dgetarray getcall)
        {
            Fget = getcall;
        }
        //----------------------
         /// <summary>
        /// indexer for indexed array gets based on eProvider enum index
        /// </summary>
        /// <param name="index">index to array</param>
        /// <value> Paramter value based on index</value>
        /// <returns>value in provider array for index</returns>    
        public int this[eProvider index]
        {
            get
            {
                if (!ProviderClass.valid(index)) throw new WaterSim_Exception(WS_Strings.wsInvalidProvider);
                int[] temp = Fget();
                return temp[(int)index];
            }
        }
        //----------------------
        /// <summary>
        /// indexer for indexed array gets and sets
        /// </summary>
        /// <param name="index">index to array</param>
        /// <returns>value in provider array for index</returns>
        /// <value>indexed Model Parameter value</value>
        public int this[int index]
        {
            get
            {
                return this[(eProvider)index];
            }
        }
        //----------------------
        /// <summary>
        /// Gets Model Parameter Array values 
        /// </summary>
        /// <returns>an Array of Model Values</returns>
        /// <value>Model Parameter value array</value>
        public ProviderIntArray getvalues()
        {
            ProviderIntArray pia = new ProviderIntArray(0);
            pia.Values = Fget();
            return pia;
       //     return Fget();
        }
    }
    //=========================================
   /// <summary>
   /// enum values for each of the Water Providers
   /// </summary>
   /// <remarks>This is a key enumerator for all the provide input and output paramters, data stored in 
   /// provider arrays is stored in the order of these enum values.  Most provider data structures allow indexing based
   /// on these enum values and the ProviderClass provides an IEnumerator that returns enum values using a foreach () loop.  The exception to this is the
   /// eProvider.regional value which is not considered a valid provider by most routines</remarks>
    public enum eProvider
    {

        Adaman_Mutual,
        White_Tanks,
        Paradise_Valley,
        Sun_City,
        Sun_City_West,
        Avondale,
        Berneil,
        Buckeye,
        Carefree,
        Cave_Creek,
        Chandler,
        Chaparral_City,
        Surprise,
        Clearwater_Utilities,
        Desert_Hills,
        El_Mirage,
        Gilbert,
        Glendale,
        Goodyear,
        Litchfield_Park,
        Mesa,
        Peoria,
        Phoenix,
        Queen_Creek,
        Rigby,
        Rio_Verde,
        Rose_Valley,
        Scottsdale,
        Sunrise,
        Tempe,
        Tolleson,
        Valley_Utilities,
        West_End,
  // Regional is not consider a valid provider enum by most routines
        Regional
  
    };

  
    //--------------------------------------------------------------------------------- 
    //===========================================
    // ProviderClass

    /// <summary>
    /// This a static class that provides support methods and constants for the eProvider enum
    /// </summary>
    public static class ProviderClass
    {
        // Provider Routines, Constants and enums
        /// <summary>
        /// The last valid provider enum value
        /// </summary>
        /// <value>eProvider enum</value>
        public const eProvider LastProvider = eProvider.West_End;
        /// <summary>
        /// The first valid enum value
        /// </summary>
        /// <value>eProvider enum</value>
        public const eProvider FirstProvider = eProvider.Adaman_Mutual;
        /// <summary>
        /// The number of valid enum values, eProvider.
        /// </summary>
        /// <value>count of valid eProvider enums</value>
        /// <remarks>eProvider.Regional is not considered one of the valie eProvider enum values</remarks>
        public const int NumberOfProviders = (int)eProvider.Regional;
        //---------------------------------------------------------

        //---------------------------------------------------------
        /// <summary>
        /// Enumerable Collection of eProviders
        /// </summary>
        /// <returns>eProvider</returns>
        public static IEnumerable<eProvider> providers()
        {
            for (eProvider i = FirstProvider; i < eProvider.Regional; i++)
            {
                yield return i;
            }
        }
        //-------------------------------------------------------------
        /// <summary>
        /// Determines of index is a valid eProvider array index
        /// </summary>
        /// <param name="index">an index to check</param>
        /// <returns>true if valid, otherwise false</returns>
        public static bool valid(int index)
        {
            return ((index >= 0) & (index < (int)eProvider.Regional));
        }
        //-----------------------------------------------------------------
        /// <summary>
        /// Determines if valid eProvider value
        /// </summary>
        /// <param name="p">an eProvider value to check</param>
        /// <returns>true if valid</returns>
        public static bool valid(eProvider p)
        {
            return (((int)p >= 0) & (p < eProvider.Regional));
        }
        //-----------------------------------------------------------------
        /// <summary>
        /// Determines if valid eProvider fieldname
        /// </summary>
        /// <param name="fieldname">a fieldname to check</param>
        /// <returns>true if valid</returns>
        public static bool valid(string fieldname)
        {
            bool found = false;
            int pindex = -1;
            fieldname = fieldname.ToUpper().Trim();
            for (int i=0;i<FieldNameList.Length; i++)
            {
                if (FieldNameList[i].ToUpper() == fieldname)
                {
                    pindex = i;
                    found = true;
                    break;
                }
            }
            return found;
        }
        //-----------------------------------------------------------------
        internal static void TestAndThrowInvalidProviderException(eProvider p)
        {
            if (!valid(p)) { throw new WaterSim_Exception(WS_Strings.Get(WS_Strings.wsInvalidProvider) + p.ToString()); }
        }

        //-----------------------------------------------------------------
        /// <summary>
        /// Returns an eProvider based on its fieldname
        /// </summary>
        /// <param name="fieldname">a fieldname to look up</param>
        /// <returns>a valid eProvider value</returns>
        /// <exception cref="WaterSim_Exception"> if not a valid fieldname</exception>
        public static eProvider provider(string fieldname)
        {
            if (valid(fieldname))
            {
                int pindex = -1;
                FastFindFieldname(fieldname, ref pindex);
                return (eProvider)pindex;
            }
            else
            {
                throw new WaterSim_Exception("Invlaid Provider Fieldname: "+fieldname);

            }
        }
        //-----------------------------------------------------------------
        /// <summary>
        /// Returns the eProvider value that corresponds to the eProvider array index
        /// </summary>
        /// <param name="index">index for eProvider</param>
        /// <returns>a valid eProvider value</returns>
        /// <exception cref="WaterSim_Exception"> if not a valid index</exception>
        public static eProvider provider(int index)
        {
            if (valid(index))
            {
                return (eProvider)index;
            }
            else
            {
                throw new WaterSim_Exception("Provider index -" + index.ToString() + " is out of range of " + FirstProvider.ToString() + " to " + LastProvider.ToString());

            }
        }
        //-----------------------------------------------------------------
        /// <summary>
        /// Provider Array index for eProvider value
        /// </summary>
        /// <param name="p">provider</param>
        /// <returns>index of provider</returns>
        public static int index(eProvider p)
        {
            TestAndThrowInvalidProviderException(p);
            return (int)p;
        }
        //-----------------------------------------------------------------
       /// <summary>
       /// Returns the index of the eProvider based on is fieldname.  Very slow, First checks if valid and thorws exception of not
       /// </summary>
       /// <param name="fieldname">an eProvider fieldname</param>
       /// <returns>index of eProvider</returns>
       /// <exception cref="WaterSim_Exception"> if no valid fieldname</exception>
        public static int index(string fieldname)
        {
            if (!valid(fieldname)) throw new WaterSim_Exception("Invlaid provider fieldname: fieldname");
           int pindex = -1;
           FastFindFieldname(fieldname,ref pindex);
           return pindex;
        }
        //-----------------------------------------------------------------
        /// <summary>
        /// Returns if Valid and Index.  Much faster, one search no exceptions
        /// </summary>
        /// <param name="fieldname">an eProvider fieldname</param>
        /// <param name="index">ref for index, returns -1 if returns false</param>
        /// <returns>true of valid eProvider fieldname, false otherwise</returns>
        public static bool FastFindFieldname(string fieldname, ref int index)
        {
            bool found = false;
            int pindex = -1;
            fieldname = fieldname.ToUpper().Trim();
            for (int i = 0; i < FieldNameList.Length; i++)
            {
                if (FieldNameList[i].ToUpper() == fieldname)
                {
                    pindex = i;
                    found = true;
                    break;
                }
            }
            index = pindex;
            return found;
        }
        //-------------------------------------------------------------
        /// <summary>
        /// Full text label for eProvider
        /// </summary>
        /// <param name="p">provider</param>
        /// <returns>label</returns>
        public static string Label(eProvider p)
        {
            TestAndThrowInvalidProviderException(p);
            return ProviderNameList[(int)p];
        }
        //-------------------------------------------------------------
       /// <summary>
       /// Fieldname for Provider
       /// </summary>
       /// <param name="p">provider</param>
       /// <returns>fieldname</returns>
        public static string FieldName(eProvider p)
        {
            TestAndThrowInvalidProviderException(p);
            return FieldNameList[(int)p];
        }
        //-------------------------------------------------------------
        //-------------------------------------------------------------
        private static string[] ProviderNameList = new string[NumberOfProviders] {       
  
            "Adaman Mutual",
            "White Tanks",
            "Paradise Valley",
            "Sun City",
            "Sun City West",
            "Avondale",
            "Berneil",
            "Buckeye",
            "Carefree",
            "Cave Creek",
            "Chandler",
            "Chaparral City",
            "Surprise",
            "Clearwater Utilities",
            "Desert Hills",
            "El Mirage",
            "Gilbert",
            "Glendale",
            "Goodyear",
            "Litchfield Park",
            "Mesa",
            "Peoria",
            "Phoenix",
            "Queen Creek",
            "Rigby",
            "Rio Verde",
            "Rose Valley",
            "Scottsdale",
            "Sunrise",
            "Tempe",
            "Tolleson",
            "Valley Utilities",
            "West End"};

        private static string[] FieldNameList = new string[NumberOfProviders]  {       
  
       "ad", //providers.Adaman_Mutual;
       "wt", //providers.White_Tanks;
       "pv", //providers.Paradise_Valley;
       "su", //providers.Sun_City;
       "sw", //providers.Sun_City_West;
       "av", //providers.Avondale;
       "be", //providers.Berneil;
       "bu", //providers.Buckeye;
       "cf", //providers.Carefree;
       "cc", //providers.Cave_Creek;
       "ch", //providers.Chandler;
       "cp", //providers.Chaparral_City;
       "sp", //providers.Surprise;
       "cu", //providers.Clearwater_Utilities;
       "dh", //providers.Desert_Hills;
       "em", //providers.El_Mirage;
       "gi", //providers.Gilbert;
       "gl", //providers.Glendale;
       "go", //providers.Goodyear;
       "lp", //providers.Litchfield_Park;
       "me", //providers.Mesa;
       "pe", //providers.Peoria;
       "ph", //providers.Phoenix;
       "qk", //providers.Queen_Creek;
       "rg", //providers.Rigby;
       "rv", //providers.Rio_Verde;
       "ry", //providers.Rose_Valley;
       "sc", //providers.Scottsdale;
       "sr", //providers.Sunrise;
       "te", //providers.Tempe;
       "to", //providers.Tolleson;
       "vu", //providers.Valley_Utilities;
       "we", //providers.West_End;
    };

    }
    #endregion
    //----------------------------------------------------------------
    //BASIC DATA / PROVIDER ARRAYS
    //--------------------------------------------------------------------------------- 
    #region DataArrays
    /// <summary>
   /// An array structure for provider integer data.  Each cell corresponds to a different provider.  Used to set and retrieve ModelParameter data.  See ProviderClass for indexes to array 
   /// <see cref="ProviderClass"/> 
   /// </summary>
    public struct ProviderIntArray
    {
        /// <summary> The int values array</summary>
        public int[] Values;
        /// <summary>   The length of the Values array </summary>
        /// <value> The length zero based. </value>
        public int Length
        { get { return Values.Length; } }

        /// <summary>
        ///     Indexer to get or set items within this array using index syntax based on an eProvider vakue.
        /// </summary>
        ///
        /// <value> The indexed item. </value>

        public int this[eProvider ep]
        { 
            get {return Values[(int)ep]; }
            set { Values[(int)ep] = value; }
        }

        /// <summary>
        ///     Indexer to get or set items within this array using an int index (overloaded)
        /// </summary>
        ///
        /// <value> The indexed item. </value>
        public int this[int index]
        {
            get { return Values[index]; }
            set { Values[index] = value; }
        }

        ///// <summary>
        ///// Default Constructor
        ///// </summary>
        //public ProviderIntArray()
        //{
        //    Values = new int[ProviderClass.NumberOfProviders];
            
        //}        
        /// <summary>   Constructor. </summary>
        /// <param name="Default">  The default int value to set each cell at instantiation. </param>
        public ProviderIntArray(int Default)
        {
            Values = new int[ProviderClass.NumberOfProviders];
            if (Default != 0) for (int i = 0; i < ProviderClass.NumberOfProviders; i++) Values[i] = Default;
        }
    }
   /// <summary>
   /// An array structure for provider integer data.  Each cell corresponds to a different provider.  Used to set and retrieve ModelParameter data.  See ProviderClass for indexes to array 
   /// <see cref="ProviderClass"/> 
   /// </summary>
    public struct ProviderDoubleArray
    {
        /// <summary> The Double values array</summary>
        public double[] Values;
        /// <summary>   The length of the Values array </summary>
        /// <value> The length zero based. </value>
        public int Length
        { get { return Values.Length; } }

        /// <summary>
        ///     Indexer to get or set items within this array using index syntax based on an eProvider vakue.
        /// </summary>
        ///
        /// <value> The indexed item. </value>

        public double this[eProvider ep]
        { 
            get {return Values[(int)ep]; }
            set { Values[(int)ep] = value; }
        }
        /// <summary>
        ///     Indexer to get or set items within this array using an int index (overloaded)
        /// </summary>
        ///
        /// <value> The indexed item. </value>
        public double this[int index]
        {
            get { return Values[index]; }
            set { Values[index] = value; }
        }

        ///// <summary>   Default Constructor. </summary>
        ///// <param name="Default">  The default int value to set each cell at instantiation. </param>
        //public ProviderDoubleArray()
        //{
        //    Values = new double[ProviderClass.NumberOfProviders];
        //}
  
        /// <summary>   Constructor. </summary>
        /// <param name="Default">  The default int value to set each cell at instantiation. </param>
        public ProviderDoubleArray(double Default)
        {
            Values = new double[ProviderClass.NumberOfProviders];
            if (Default != 0.0) for (int i = 0; i < ProviderClass.NumberOfProviders; i++) Values[i] = Default;
        }
    }
    //--------------------------------------------------------------------------------- 
   /// <summary>
   /// An array structure for provider strings.  Each cell corresponds to a different provider.  Used primarily for labeling
   /// </summary>
    public struct ProviderStringArray
    {

        /// <summary> The string values </summary>
        public string[] Values;

        /// <summary>   The length of the Values array </summary>
        /// <value> The length zero based. </value>

        public int Length
        { get { return Values.Length; } }

        /// <summary>
        ///     Indexer to get or set items within this array using index syntax based on an eProvider vakue.
        /// </summary>
        ///
        /// <value> The indexed item. </value>

        public string this[eProvider ep]
        {
            get { return Values[(int)ep]; }
            set { Values[(int)ep] = value; }
        }
        /// <summary>
        ///     Indexer to get or set items within this array using an int index (overloaded)
        /// </summary>
        ///
        /// <value> The indexed item. </value>
        public string this[int index]
        {
            get { return Values[index]; }
            set { Values[index] = value; }
        }
        /// <summary>   Constructor. </summary>
        /// <param name="Default">  The default string value to set each string to on instantiation.. </param>

        public ProviderStringArray(string Default)
        {
            Values = new string[ProviderClass.NumberOfProviders];
            if (Default != "") for (int i = 0; i < ProviderClass.NumberOfProviders; i++) Values[i] = Default;
        }
    }
    //----------------------------------------------------------------
   /// <summary>
   /// An array structure for ModelParamter integer data for Base parameters.
    /// <remarks>Each cell corresponds to a different ModelParmeter value. See ParameterManager for indexes to array.</remarks>  
   /// </summary>
    public struct ModelParameterBaseArray
    {
        /// <summary> The values </summary>
        public int[] Values;

        /// <summary>   The length of the Values array </summary>
        /// <value> The length zero based. </value>

        public int Length
        { get { return Values.Length; } }

        /// <summary>
        ///     Indexer to get or set items within this array using an int index (overloaded)
        /// </summary>
        ///
        /// <value> The indexed item. </value>

        public int this[int index]
        {
            get { return Values[index]; }
            set { Values[index] = value; }
        }

        /// <summary>   Constructor. </summary>
        /// <param name="size"> The size of the BaseArray. </param>

        public ModelParameterBaseArray(int size)
        {
            Values = new int[size];
        }
    }
    //----------------------------------------------------------------
   /// <summary>
    /// An Array structure for ModelParameters for Provider Data.  
    /// <remarks>Each cell corresponds to a different ModelParmeter value. See ParameterManager for indexes to array.</remarks> 
   /// </summary>
    public struct ModelParameterProviderArray
    {

        /// <summary> The values </summary>
        
        public ProviderIntArray[] Values;

        /// <summary>   The length of the Values array </summary>
        /// <value> The length zero based. </value>

        public int Length
        { get { return Values.Length; } }


        /// <summary>
        ///     Indexer to get or set items within this array using an int index (overloaded)
        /// </summary>
        ///
        /// <value> The indexed item. </value>
        public ProviderIntArray this[int index]
        {
            get { return Values[index]; }
            set { Values[index] = value; }
        }

        /// <summary>   Constructor. </summary>
        /// <param name="size"> The size of the array</param>
        
        public ModelParameterProviderArray(int size)
        {
            Values = new ProviderIntArray[size];
            for (int i = 0; i < size; i++) Values[i] = new ProviderIntArray(0);
        }

    }
    //----------------------------------------------------------------------------
  /// <summary>
  /// Struct to hold base and Provider Inputs
  /// </summary>
  /// <remarks>The ParameterManagerClass provides a method to create this struct without passing sizes to the constructor</remarks>
  /// <seealso cref="ParameterManagerClass.NewSimulationInputs"/>
  /// <example><code>
  ///          WaterSim WSim = new WaterSim("Temp","Data");
  /// 		   SimulationInputs MySI = WSim.ParamManager.NewSimulationInputs();
  ///		   
  /// </code></example>
    public struct SimulationInputs
  {

        /// <summary> Name of the scenario for these inputs</summary>
        public string ScenarioName;
      /// <summary>
      /// Base Input Values
      /// </summary>
      public ModelParameterBaseArray BaseInput;
      /// <summary>
      /// Provider Input Values
      /// </summary>
      public ModelParameterProviderArray ProviderInput;

      /// <summary>   Constructor. </summary>
      /// <param name="BaseInputSize">        Size of the base input array. </param>
      /// <param name="ProviderInputSize">    Size of the provider input array. </param>
      
      public SimulationInputs(int BaseInputSize, int ProviderInputSize)
      {
          ScenarioName = "";
          BaseInput = new ModelParameterBaseArray(BaseInputSize);
          ProviderInput = new ModelParameterProviderArray(ProviderInputSize);
      }
  }
    //----------------------------------------------------------------------
    /// <summary>
    /// Struct to hold base and Provider Outputs
    /// <remarks>The ParameterManagerClass provides a method to create this struct without passing sizes to the constructor</remarks>
    /// <seealso cref="ParameterManagerClass.NewSimulationOutputs"/>
    /// <example><code>
    ///          WaterSim WSim = new WaterSim("Temp","Data");
    /// 		   SimulationOutputs MySO = WSim.ParamManager.NewSimulationOutputs();
    ///		   
    /// </code></example>
    /// </summary>
    public struct SimulationOutputs
    {

        /// <summary> Name of the scenario for these outputs</summary>
        public string ScenarioName;
        /// <summary>
        /// Base Output Values
        /// </summary>
        public ModelParameterBaseArray BaseOutput;
        /// <summary>
        /// Provider Output Values
        /// </summary>
        public ModelParameterProviderArray ProviderOutput;

        /// <summary>   Constructor. </summary>
        /// <param name="BaseOutputSize">        Size of the base output array. </param>
        /// <param name="ProviderOutputSize">    Size of the provider output array. </param>

        public SimulationOutputs(int BaseOutputSize, int ProviderOutputSize)
        {
            ScenarioName = "";
            BaseOutput = new ModelParameterBaseArray(BaseOutputSize);
            ProviderOutput = new ModelParameterProviderArray(ProviderOutputSize);
        }
    }
    //----------------------------------------------------------------------
    
    /// <summary>
    /// A structure for one years worth of Simulation Results
    /// </summary>
    public struct AnnualSimulationResults
    {
        /// <summary>
        ///  Year of Simulation
        /// </summary>
        public int year;
        /// <summary> The base output array, each cell is a base output parameter value based on the eModePAram index </summary>
        public int[] BaseOutput;
        /// <summary> The base input array, each cell is a base input parameter value</summary>
        public int[] BaseInput;
        /// <summary> The provider output array</summary>
        public ModelParameterProviderArray ProviderOutput;

        /// <summary> The provider input array</summary>
        public ModelParameterProviderArray ProviderInput;

        /// <summary>   Constructor. </summary>
        /// <param name="BaseOutputSize">       Size of the base output array. </param>
        /// <param name="ProviderOutputSize">   Size of the provider output array. </param>
        /// <param name="BaseInputSize">        Size of the base input array. </param>
        /// <param name="ProviderInputSize">    Size of the provider input array. </param>

        public AnnualSimulationResults(int BaseOutputSize, int ProviderOutputSize, int BaseInputSize, int ProviderInputSize)
        {
            year = 0;
            BaseOutput = new int[BaseOutputSize];
            BaseInput = new int[BaseInputSize];
            ProviderOutput = new ModelParameterProviderArray(ProviderOutputSize);
            ProviderInput = new ModelParameterProviderArray(ProviderInputSize);
        }
    }

    internal struct AnnualSimulationResults2
    {
        /// <summary>
        ///  Year of Simulation
        /// </summary>
        public int year;
        /// <summary> The base output array, each cell is a base output parameter value based on the eModePAram index </summary>
        public int[] BaseOutput;
        /// <summary> The base input array, each cell is a base input parameter value</summary>
        public int[] BaseInput;
        /// <summary> The provider output </summary>
        public ProviderIntArray[] ProviderOutput;
        public ProviderIntArray[] ProviderInput;

        public AnnualSimulationResults2(int BaseOutputSize, int ProviderOutputSize, int BaseInputSize, int ProviderInputSize)
        {
            year = 0;
            BaseOutput = new int[BaseOutputSize];
            BaseInput = new int[BaseInputSize];
            ProviderOutput = new ProviderIntArray[ProviderOutputSize];
            ProviderInput = new ProviderIntArray[ProviderInputSize];
        }
    }//----------------------------------------------------------------------------

    /// <summary>   Simulation strings for Labels.  </summary>
        public struct SimulationStrings
    {

        /// <summary> The base output </summary>
        public string[] BaseOutput;
            
         /// <summary> The base input </summary>
         public string[] BaseInput;

        /// <summary> The provider output </summary>
        public string[] ProviderOutput;

        /// <summary> The provider input </summary>
        public string[] ProviderInput;

        /// <summary>   Constructor. </summary>
        /// <param name="BaseOutputSize">       Size of the base output. </param>
        /// <param name="ProviderOutputSize">   Size of the provider output. </param>
        /// <param name="BaseInputSize">        Size of the base input. </param>
        /// <param name="ProviderInputSize">    Size of the provider input. </param>

        public SimulationStrings(int BaseOutputSize, int ProviderOutputSize, int BaseInputSize, int ProviderInputSize)
        {
            BaseOutput = new string[BaseOutputSize];
            BaseInput = new string[BaseInputSize];
            ProviderOutput = new string[ProviderOutputSize];
            ProviderInput = new string[ProviderInputSize];
        }
    }


        //----------------------------------------------------------------------------
    #endregion

        #region Utils

        internal static class util
    {
        #region apiconstants
        //====================================================================
        // CONSTANTS  ENUMS STRUCTS
        // THESE CAN NOT BE CHANGED WITH OUT CHANGING DATA INPUTS!
        // There is no error checking for this
        internal const int ModelParamterUseDefault = -1;

        internal const int WaterSimDCDC_Default_Simulation_StartYear = 2006;
        internal const int WaterSimDCDC_Default_Simulation_EndYear = 2085;
        internal const int WaterSimDCDC_Default_Colorado_Historical_Extraction_Start_Year = 2000;
        internal const int WaterSimDCDC_Default_ProviderDemandOption = 0;
        internal const int WaterSimDCDC_Default_SaltVerde_Historical_Data = 2;
        // Fixes
        internal const int WaterSimDCDC_Default_Colorado_User_Adjustment_StartYear = 2006;
        //internal const int WaterSimDCDC_Default_SaltVerde_Historical_Data = 1;
        internal const int WaterSimDCDC_Default_SaltVerde_User_Adjustment_Start_Year = 2006;
        internal const int WaterSimDCDC_Provider_Demand_Option = 3;
        internal const int DefaultReduceGPCDValue = 1;
        // Fix for provider inputs
        internal const int WaterSimDCDC_Default_Use_GPCD = 150;
        internal const int WaterSimDCDC_Default_Surface_to_Vadose_Time_Lag_in_Years = 25;
        internal const int WaterSimDCDC_Default_SurfaceWater_to_WaterBank = 0;

        //=================================================================================
        internal const int WaterSimDCDC_Default_Colorado_Historical_Data_Source = (int)ColoradoRiverRecord.eBureauRecord;

        //public const int    ProviderClass.NumberOfProviders  = 33;

        public const int UseDefaultInput = 0;
        #endregion

        #region  utilroutines
        /************************************************
         * ROUTINES
         * 
         * 
         * *********************************************/
        //----------------------------------------------------------------
        public static void fillzero(int[] values)
        {
            int cnt = values.Length;
            for (int i = 0; i < cnt; i++)
            { values[i] = 0; }
        }
        //----------------------------------------------------------------
        public static void fillvalues(int[] values, int value)
        {
            int cnt = values.Length;
            for (int i = 0; i < cnt; i++)
            { values[i] = value; }
        }
        //----------------------------------------------------------------
        public static void fillDefault(int[] values)
        {
            int cnt = values.Length;
            for (int i = 0; i < cnt; i++)
            { values[i] = UseDefaultInput; }
        }
        #endregion
    }
        #endregion
        /****************************************************************
     * Parameter Classes, Routines, Contants and enums
     * ************************************************************/
    #region Params

    /// <summary>   Values that represent ModelParameter.  </summary>
    /// <remarks> Unique identifier for each ModelParameter</remarks>
    public enum eModelParam
    {
        //Output Parameters
        epColorado_River_Flow,
        epPowell_Storage,
        epMead_Storage,
        epSaltVerde_River_Flow,
        epSaltVerde_Storage,
        epEffluent_To_Agriculture,
        epGroundwater_Pumped_Municipal,
        epGroundwater_Balance,
        epSaltVerde_Annual_Deliveries_SRP,
        epSaltVerde_Class_BC_Designations,
        epColorado_Annual_Deliveries,
        epGroundwater_Bank_Used,
        epGroundwater_Bank_Balance,
        epReclaimed_Water_Used,
        epReclaimed_Water_To_Vadose,
        epReclaimed_Water_Discharged,
        epReclaimed_Water_to_DirectInject,
        epRO_Reclaimed_Water_Used,
        epRO_Reclaimed_Water_to_DirectInject,
        epEffluent_Created,
        epEffluent_To_Vadose,
        epEffluent_To_PowerPlant,
        epEffluent_Discharged,
        epDemand_Deficit,
        epTotal_Demand,
        epGPCD_Used,
        epPopulation_Used,

        //Input Parameters

        epSimulation_Start_Year,
        epSimulation_End_Year,
        epColorado_Historical_Extraction_Start_Year,
        epColorado_Historical_Data_Source,
        epColorado_Climate_Adjustment_Percent,
        epColorado_User_Adjustment_Percent,
        epColorado_User_Adjustment_StartYear,
        epColorado_User_Adjustment_Stop_Year,
        epSaltVerde_Historical_Extraction_Start_Year,
        epSaltVerde_Historical_Data,
        epSaltVerde_Climate_Adjustment_Percent,
        epSaltVerde_User_Adjustment_Percent,
        epSaltVerde_User_Adjustment_Start_Year,
        epSaltVerde_User_Adjustment_Stop_Year,
        epProvider_Demand_Option,
        epPCT_Reduce_GPCD,   // added 8 11 2011
        //            epUse_Water_Demand,
        //            epUse_Population,
        epUse_GPCD,
        epPCT_WasteWater_to_Reclaimed,
        epPCT_Wastewater_to_Effluent,
        epPCT_Reclaimed_to_RO,
        epPCT_RO_to_Water_Supply,
        epPCT_Reclaimed_to_DirectInject,
        epPCT_Reclaimed_to_Water_Supply,
        epPCT_Reclaimed_to_Vadose,
        epPCT_Effluent_to_Vadose,
        epPCT_Effluent_to_PowerPlant,
        epSurfaceWater__to_Vadose,  // edited 7 23 11  epPCT_SurfaceWater__to_Vadose, 
        epSurface_to_Vadose_Time_Lag,
        epWaterBank_Source_Option,
        epPCT_SurfaceWater_to_WaterBank,
        epUse_SurfaceWater_to_WaterBank,
        epPCT_WaterSupply_to_Residential,
        epPCT_WaterSupply_to_ComInd,
        //          epPCT_WaterSupply_to_DirectInject,  // deleted 7 22 11  Deleted from Model
        epUse_WaterSupply_to_DirectInject,
        epPCT_Outdoor_WaterUse,
        epPCT_Groundwater_Treated,
        epPCT_Reclaimed_Outdoor_Use,
        epPCT_Growth_Rate_Adjustment,
        epPCT_Max_Demand_Reclaim,
        // LENGTH
        epLENGTH
    };
    // --------------------------------------------------------------

    //----------------------------------------------------------------

    /// <summary>   Model parameter class.  </summary>
    public static class ModelParamClass
    {
        //LIST of parameters  
        static int FNumberOfModelParameters = (int)eModelParam.epLENGTH;
        //--------------------------------------------

        /// <summary>   Valids. </summary>
        /// <param name="p">    a EModelParam  </param>
        ///
        /// <returns>   true if p is a valid ModelParameter. </returns>

        public static bool valid(eModelParam p)
        {
            if (((int)p >= 0) & (FNumberOfModelParameters > (int)p)) return true;
            else return false;
        }
        //------------------------------------------

        /// <summary>   Gets the number of valid ModelParameters. </summary>
        /// <returns>   the length of the ModelParameter list </returns>

        public static int Length()
        { return FNumberOfModelParameters; }
        //------------------------------------------
        internal static string Name(eModelParam p) { return p.ToString(); }
        //-------------------------------------------------------------------
        internal static void SeteModelParamLength(int Size)
        {
            FNumberOfModelParameters = Size;
         }
        // 
    }
    //---------------------------------------------------------

    /// <summary>   Values that represent modelParamtype.  </summary>
    public enum modelParamtype { 
        

        /// <summary> Unknown so ignore.  </summary>
        mptUnknown, 

        /// <summary> a base output paramter.  </summary>
        mptOutputBase, 

        /// <summary> a provider output paramter.  </summary>
        mptOutputProvider, 

        /// <summary> a base input parameter.  </summary>
        mptInputBase, 

        /// <summary> a provider input parameter. </summary>
        mptInputProvider 
        };

    //---------------------------------------------------------

    /// <summary>   Values that represent rangeChecktype.  </summary>
    public enum rangeChecktype {

        /// <summary>  Unknown, no range check done.  </summary>
        rctUnknown, 

        /// <summary> parameter needs no range check.  </summary>
        rctNoRangeCheck,

        /// <summary> parameter has a valid low and high range.  </summary>
        rctCheckRange,

        /// <summary> parameter has a valid low and high range. which is affected by or affects the range of other model parameters.  </summary>
        rctCheckRangeSpecial,

        /// <summary> parameter value must be positive.  </summary>
        rctCheckPositive }

    //---------------------------------------------------------
    /// <summary>   Delegate for handling Reload events. </summary>
    ///
    /// <remarks>   Ray, 8/5/2011. </remarks>
    ///
    /// <param name="source">   Source for the. </param>
    /// <param name="e">        Event information. </param>
    public delegate void ReloadEventHandler(object source, System.EventArgs e);

    /// <summary>   Additional information for model parameter events.  </summary>
    public class ModelParameterEventArgs : System.EventArgs
    {
        /// <summary> The ModelParameter evoking the event</summary>
        public ModelParameterClass modelParameter;

        /// <summary>   Constructor. </summary>
        /// <param name="MP">   The ModelParameter. </param>

        public ModelParameterEventArgs(ModelParameterClass MP)
        {
            modelParameter = MP;
        }
    }
    //---------------------------------------------------------

    /// <summary>   Model parameter class.  </summary>
    /// <remarks>   This is the heart of the ModelParameter management system.  Each parameter for the Fortran model exposed in the API has a unique ModelParameter object.. </remarks>

    public class ModelParameterClass
    {
        //----------------------------------------
        //Fields

        /// <summary> The model parameter </summary>
        protected eModelParam FModelParam;

        /// <summary> The label </summary>
        protected string FLabel = "";

        /// <summary> The fieldname </summary>
        protected string FFieldname = "";

        /// <summary> The ftype </summary>
        protected modelParamtype Ftype = modelParamtype.mptUnknown;

        /// <summary> Type of the range check </summary>
        protected rangeChecktype FRangeCheckType = rangeChecktype.rctUnknown;

        /// <summary> The low range </summary>
        protected int FLowRange = 0;

        /// <summary> The high range </summary>
        protected int FHighRange = 0;

        /// <summary> true if need to notify for relaod on value change </summary>
        protected bool FReload = false;

        // method references
        DcheckBase FSpecialBaseCheck;
        DcheckProvider FSpecialProviderCheck;
        Dget Fget;
        Dset Fset;
        Dgetarray Fgetarray;
        Dsetarray Fsetarray;

        providerArrayProperty FProviderProperty;


        /// <summary> Event queue for all listeners interested in Reload events. </summary>
        public event ReloadEventHandler ReloadEvent;

        /// <summary>   Constructor. </summary>
        ///
        /// <remarks>   For now this constructor is hidden </remarks>
        ///
        /// <param name="aModelParam">                  a model parameter. </param>
        /// <param name="aLabel">                       a label. </param>
        /// <param name="aFieldname">                   a fieldname. </param>
        /// <param name="aParamType">                   Type of a parameter. </param>
        /// <param name="aRangeCheckType">              Type of a range check. </param>
        /// <param name="aLowRange">                    a low range. </param>
        /// <param name="aHighRange">                   a high range. </param>
        /// <param name="agetint">                      The agetint. </param>
        /// <param name="agetintarray">                 The agetintarray. </param>
        /// <param name="asetint">                      The asetint. </param>
        /// <param name="asetintarray">                 The asetintarray. </param>
        /// <param name="specialBaseRangeCheck">        The special base range check. </param>
        /// <param name="specialProviderRangeCheck">    The special provider range check. </param>
        /// <param name="Providerproperty">             The providerproperty. </param>

        internal ModelParameterClass(eModelParam aModelParam, string aLabel, string aFieldname, modelParamtype aParamType,
            rangeChecktype aRangeCheckType, int aLowRange, int aHighRange, Dget agetint, Dgetarray agetintarray, Dset asetint, Dsetarray asetintarray,
            DcheckBase specialBaseRangeCheck, DcheckProvider specialProviderRangeCheck, providerArrayProperty Providerproperty)
            : base()
        {
            FModelParam = aModelParam;
            FLabel = aLabel;
            FFieldname = aFieldname;
            Ftype = aParamType;
            FRangeCheckType = aRangeCheckType;
            FLowRange = aLowRange;
            FHighRange = aHighRange;
            Fget = agetint;
            Fgetarray = agetintarray;
            Fset = asetint;
            Fsetarray = asetintarray;
            FSpecialBaseCheck = specialBaseRangeCheck;
            FSpecialProviderCheck = specialProviderRangeCheck;
            FProviderProperty = Providerproperty;
            // FReload = ReloadValuesOnSet;
        }
        //----------------------------------------

        /// <summary>   Gets the model parameter emumerator. </summary>
        ///
        /// <value> The eModelParam enumerator. </value>

        public eModelParam ModelParam
        { get { return FModelParam; } }
        //----------------------------------------

        /// <summary>   Gets or sets the label. </summary>
        ///
        /// <value> The label. </value>

        public string Label
        // Text Lable for the model parameter for use text label
        { get { return FLabel; } set { FLabel = value; } }
        //----------------------------------------

        /// <summary>   Gets or sets the fieldname. </summary>
        ///
        /// <value> The fieldname. that is used in data tables as a column name</value>

        public string Fieldname
        { get { return FFieldname; } set { FFieldname = value; } }
        //----------------------------------------

        /// <summary>   Gets or sets the type of the parameter. </summary>
        ///
        /// <value> The type of the parameter. </value>

        public modelParamtype ParamType
        { get { return Ftype; } set { Ftype = value; } }


        //----------------------------------------

        /// <summary>   Gets a value indicating whether this object is input parameter. </summary>
        ///
        /// <value> true if this object is input parameter, false if not. </value>

        public bool isInputParam
        { get { return ((Ftype == modelParamtype.mptInputBase) | (Ftype == modelParamtype.mptInputProvider)); } }
        //----------------------------------------

        /// <summary>   Gets a value indicating whether this object is output parameter. </summary>
        ///
        /// <value> true if this object is output parameter, false if not. </value>

        public bool isOutputParam
        { get { return ((Ftype == modelParamtype.mptOutputBase) | (Ftype == modelParamtype.mptOutputProvider)); } }
        //----------------------------------------

        /// <summary>   Gets a value indicating whether this object is provider parameter. </summary>
        ///
        /// <value> true if this object is provider parameter, false if not. </value>

        public bool isProviderParam
        { get { return ((Ftype == modelParamtype.mptOutputProvider) | (Ftype == modelParamtype.mptInputProvider)); } }
        //----------------------------------------

        /// <summary>   Gets a value indicating whether this object is base parameter. </summary>
        ///
        /// <value> true if this object is base parameter, false if not. </value>

        public bool isBaseParam
        { get { return ((Ftype == modelParamtype.mptOutputBase) | (Ftype == modelParamtype.mptInputBase)); } }
        //----------------------------------------

        /// <summary>   Gets or sets the type of the range check. </summary>
        ///
        /// <value> The type of the range check. </value>

        public rangeChecktype RangeCheckType
        { get { return FRangeCheckType; } set { FRangeCheckType = value; } }
        //----------------------------------------

        /// <summary>   Gets or sets the low range for conducting range checks. </summary>
        ///
        /// <value> The low range. </value>

        public int LowRange
        { get { return FLowRange; } set { FLowRange = value; } }
        //----------------------------------------

        /// <summary>   Gets or sets the high range for conducting range checks. </summary>
        ///
        /// <value> The high range. </value>

        public int HighRange
        { get { return FHighRange; } set { FHighRange = value; } }
        //------------------------------------------

        /// <summary>   Gets the providerArrayproperty for the ModelParameter. </summary>
        /// <remarks> This provides access to the provider model parameter</remarks>
        /// <value> The provider property. </value>
        /// <exception cref="WaterSim_Exception">if value or values are out of range for the parameter</exception>

        public providerArrayProperty ProviderProperty
        {
            get 
            { 
                if (!isProviderParam) throw new WaterSim_Exception(WS_Strings.wsMustBeProviderParameter);
                return FProviderProperty;
            }
        }
        //------------------------------------------

        /// <summary>   Gets or sets the ModelParameter value. </summary>
        /// <remarks> This is the basic access to base model parameter</remarks>
        /// <value> The value. </value>
        /// <exception cref="WaterSim_Exception">if value or values are out of range for the parameter</exception>

        public int Value
        {
            get 
            {
                if (!isBaseParam) throw new WaterSim_Exception(WS_Strings.wsMustBeBaseParameter);
                return GetInt(); 
            }
            set 
            {
                if ((!isBaseParam)|(!isInputParam))  throw new WaterSim_Exception(WS_Strings.wsModelParamMUstBeInputBase);
                Fset(value); 
            }
        }
        //------------------------------------------
        //------------------------------------------
        //------------------------------------------
        //public bool Reload { get { return FReload; } }
        //internal bool Reload { set { FReload = value; } }
        //----------------------------------------
        //public int HighRange
        internal Dget GetInt { get { return Fget; } }
        internal Dset SetInt { get { return Fset; } }
        internal Dgetarray GetIntArray { get { return Fgetarray; } }
        private Dsetarray SetIntArray { get { return Fsetarray; } }
        //------------------------------------------------
        //------------------------------------------------
        internal bool TestReadWrite()
        // VERY HAZARDOUS Methiod.  Model Should ALways be Reset after using this method before trying to run the model. 
        {
            const int badNumber = -99999;
            const int oddnumber = 23700;
            int testValue1 = 0;
            int testValue2 = 0;
            int backvalue = badNumber;
            int[] testarray = new int[ProviderClass.NumberOfProviders];
            int[] backarray = new int[ProviderClass.NumberOfProviders];
            bool test = true;
            if (FRangeCheckType == rangeChecktype.rctNoRangeCheck)
            {
                testValue1 = oddnumber;
                testValue2 = oddnumber / 2;
            }
            else
                if (FRangeCheckType == rangeChecktype.rctCheckRange)
                {
                    testValue1 = HighRange;
                    testValue2 = LowRange;
                }
                else
                   if ((FModelParam == eModelParam.epColorado_Historical_Extraction_Start_Year) | (FModelParam == eModelParam.epSaltVerde_Historical_Extraction_Start_Year))
                    {
                        testValue1 = 1960;
                        testValue2 = 1950;
                    }
                    else
                    {
                        testValue1 = 10;
                        testValue2 = 20;
                    }

            switch (ParamType)
            {
                case modelParamtype.mptOutputBase:
                    backvalue = badNumber;
                    backvalue = GetInt();
                    test = (backvalue != badNumber);
                    break;
                case modelParamtype.mptOutputProvider:
                    for (int i = 0; i < ProviderClass.NumberOfProviders; i++)
                        backarray[i] = badNumber;
                    backarray = GetIntArray(); ;
                    for (int i = 0; i < ProviderClass.NumberOfProviders; i++)
                    {
                        test = (backarray[i] != badNumber);
                        if (!test) break;
                    }
                    break;
                case modelParamtype.mptInputBase:
                    SetInt(testValue1);
                    backvalue = GetInt();
                    test = (testValue1 == backvalue);
                    if (test)
                    {
                        SetInt(testValue2);
                        backvalue = GetInt();
                        test = (testValue2 == backvalue);
                    }
                    break;
                case modelParamtype.mptInputProvider:
                    {
                        testValue1 = 1960;
                        testValue2 = 1950;
                    }


                    for (int i = 0; i < ProviderClass.NumberOfProviders; i++)
                        testarray[i] = testValue1;
                    SetIntArray(testarray);
                    backarray = GetIntArray(); ;
                    for (int i = 0; i < ProviderClass.NumberOfProviders; i++)
                    {
                        test = (testarray[i] == backarray[i]);
                        if (!test) break;
                    }

                    for (int i = 0; i < ProviderClass.NumberOfProviders; i++)
                        testarray[i] = testValue2;
                    SetIntArray(testarray);
                    backarray = GetIntArray(); ;
                    for (int i = 0; i < ProviderClass.NumberOfProviders; i++)
                    {
                        test = (testarray[i] == backarray[i]);
                        if (!test) break;
                    }
                    break;

            }
            return test;
        }
        //----------------------------------------

        /// <summary>   Queries if we should check range./ </summary>
        /// <returns>   true if should check range, otherwise false</returns>

        public bool ShouldCheckRange()
        {
            return ((FRangeCheckType == rangeChecktype.rctCheckRange) |
                    (FRangeCheckType == rangeChecktype.rctNoRangeCheck) |
                    (FRangeCheckType == rangeChecktype.rctCheckPositive) |
                    (FRangeCheckType == rangeChecktype.rctCheckRangeSpecial));
        }
        //----------------------------------------

        /// <summary>   Query if this object is special range check. </summary>
        /// <returns>   true if special range check, false if not. </returns>

        public bool isSpecialRangeCheck()
        { return (FRangeCheckType == rangeChecktype.rctCheckRangeSpecial); }
        

        //----------------------------------------

        /// <summary>   Formats a Range check error string. </summary>
        /// <param name="value">    The value. </param>
        ///
        /// <returns> formatted string  . </returns>

        string rangeCheckErrString(int value)
        {
            return "Range Error: " + Label + "  = " + value.ToString() + " is not in the range of " + LowRange.ToString() + " to " + HighRange.ToString();
        }
        //----------------------------------------

        /// <summary>   formats a Positive range check error string. </summary>
        /// <param name="value">  formatted string. </param>
        ///
        /// <returns>   . </returns>

        string positiveCheckErrString(int value)
        {
            return "Range Error: " + Label + "  = " + value.ToString() + " must be positive.";
        }
        //----------------------------------------

        /// <summary>   Query if 'value' is in base ModelPArameter range. </summary>
        /// <param name="value">    The value to be checked. </param>
        ///
        /// <returns>   true if base value in range, false if not. </returns>

        public bool isBaseValueInRange(int value)
        {
            string junk = "";
            return isBaseValueInRange(value, ref junk);
        }
        //----------------------------------------

        /// <summary>   Query if 'value' is in base ModelParameter range. </summary>
        /// <exception cref="WaterSim_Exception">   Thrown when ModelParameter is not a modelParamtype.mptInputBase Parameter </exception>
        ///
        /// <param name="value">        The value to be checked. </param>
        /// <param name="errMessage">   [in,out] Message describing the error. </param>
        ///
        /// <returns>   true if base value in range, false if not. </returns>

        public bool isBaseValueInRange(int value, ref string errMessage)
        {
            bool test = true;
                if ((isBaseParam) & (isInputParam))
                {
                    switch (FRangeCheckType)
                    {
                        case rangeChecktype.rctCheckRange:
                            test = ((value >= FLowRange) & (value <= FHighRange));
                            if (!test) errMessage += rangeCheckErrString(value);
                            break;
                        //--------------------
                        case rangeChecktype.rctCheckRangeSpecial:
                            string temp = "";
                            test = FSpecialBaseCheck(value, ref temp, this);
                            if (!test) errMessage += temp;
                            break;
                        //--------------------
                        case rangeChecktype.rctCheckPositive:
                            test = (value >= 0);
                            if (!test) errMessage += positiveCheckErrString(value);
                            break;
                        //--------------------
                        case rangeChecktype.rctUnknown:
                            break;
                        //--------------------
                        case rangeChecktype.rctNoRangeCheck:
                            break;
                        //--------------------
                        default:
                            break;
                    }
                }
                else
                    throw new WaterSim_Exception(WS_Strings.wsModelParamMUstBeInputBase);
            return test;
        }
        //-------------------------------------------------------------------------------

        /// <summary>   Query if 'value' is in provider ModelPArameter range. </summary>
        /// <param name="value">    The value to be checked. </param>
        /// <param name="provider"> The provider. </param>
        ///
        /// <returns>   true if value in range for this provider, false if not. </returns>

        public bool isProviderValueInRange(int value, eProvider provider)
        {
            string junk = "";
            return isProviderValueInRange(value, provider, ref junk);
        }
        //-------------------------------------------------------------------------------

        /// <summary>   Query if 'value' is in provider ModelPArameter range. </summary>
        /// <exception cref="WaterSim_Exception">   Thrown when ModelParameter is not a modelParamtype.mptInputProvider parameter. </exception>
        ///
        /// <param name="value">        The value to be checked. </param>
        /// <param name="provider">     The provider. </param>
        /// <param name="errMessage">   [in,out] Message describing the error. </param>
        ///
        /// <returns>   true if value in range for this provider, false if not. </returns>

        public bool isProviderValueInRange(int value, eProvider provider, ref string errMessage)
        {
            bool test = true;
                if ((isProviderParam) & (isInputParam))
                {
                    switch (FRangeCheckType)
                    {
                        case rangeChecktype.rctCheckRange:
                            test = ((value >= FLowRange) & (value <= FHighRange));
                            if (!test) errMessage += rangeCheckErrString(value);
                            break;
                        //--------------------
                        case rangeChecktype.rctCheckRangeSpecial:
                            string temp = "";
                            test = FSpecialProviderCheck(value, provider, ref temp, this);
                            if (!test) errMessage += temp;
                            break;
                        //--------------------
                        case rangeChecktype.rctCheckPositive:
                            test = (value >= 0);
                            if (!test) errMessage += positiveCheckErrString(value);
                            break;
                        //--------------------
                        case rangeChecktype.rctUnknown:
                            break;
                        //--------------------
                        case rangeChecktype.rctNoRangeCheck:
                            break;
                        //--------------------
                        default:
                            break;
                    }
                }
                else
                    throw new WaterSim_Exception(WS_Strings.wsModelParamMUstBeInputProvider);
            return test;
        }
        //------------------- 

        /// <summary>   Raises the reload event. </summary>
        /// <param name="e">    Event information to send to registered event handlers. </param>

        protected virtual void OnReloadEvent(ModelParameterEventArgs e)
        {
            if (ReloadEvent != null) ReloadEvent(this, e);
        }
        //---------------------------------

        /// <summary>   Evokes reload event for any object listening to this ModelParameter </summary>
        /// <seealso cref="ModelParameterEventArgs"/>
        /// <seealso cref="ReloadEvent"/>
        public void EvokeReloadEvent()
        {
            ModelParameterEventArgs MP = new ModelParameterEventArgs(this);
            OnReloadEvent(MP);
        }
    }

    //============================================================

    //============================================================

    //===========================================================

    /// <summary>   Parameter manager class. / </summary>
    /// <remarks> This class is used by the WaterSim class to manage the model parameters.  It can be access from a WaterSim object through the ParamManager property<see cref="WaterSim.ParamManager"/></remarks>
    /// The constructor is not exposed becuase only the WaterSim constructor should instantiate an object of this class.   The AddPArameter method is also not exposed at this time.  The intent is to evntually allow user define parameters to be added.

    public class ParameterManagerClass
    {
        private ModelParameterClass[] FModelParameters;
 
        internal ParameterManagerClass() : base()
        {
            FModelParameters = new ModelParameterClass[ModelParamClass.Length()];
        }
        //------------------------------------------------------------------------
        internal bool AddParameter(ModelParameterClass parameter)
        {
            bool mresult = false;
            if (parameter != null)
            {
                if (ModelParamClass.valid(parameter.ModelParam))
                {
                    FModelParameters[(int)parameter.ModelParam] = parameter;
                }
            }
            return mresult;
        }
        //---------------------------------------------------

        /// <summary>   Gets the number of parameters being managed. </summary>
        /// <returns>   The total number of parameters. </returns>

        public int NumberOfParameters()
        {
            return ModelParamClass.Length();
        }

        //---------------------------------------------------------

        /// <summary>   Queries the number of parameters being managed that are mpt modelParamtype. </summary>
        /// <param name="mpt">  The modelParamtype to be queried. </param>
        ///
        /// <returns>   The total number of parameters. </returns>

        public int NumberOfParameters(modelParamtype mpt)
        {
            int cnt = 0;
            foreach (eModelParam mp in ModelParameters())
            {
                if (Model_Parameter(mp).ParamType == mpt) cnt++;
            }
            return cnt;
        }

        //---------------------------------------------------------

        /// <summary>   Enumerates all model parameters managed in this collection of ModelPArameters./ </summary>
        /// <returns>
        ///     An enumerator that allows foreach to be used to process model parameters in this
        ///     collection.
        /// </returns>

        public IEnumerable<eModelParam> ModelParameters()
        {
            int length = NumberOfParameters();
            for (int i = 0; i < length; i++)
            {
                yield return (eModelParam)i;
            }
        }
        //---------------------------------------------------------
 

        /// <summary>   Enumerates base inputs in this collection./ </summary>
        /// <returns>
        ///     An enumerator that allows foreach to be used to process base inputs in this collection.
        /// </returns>
        /// <example><code>
        ///        bool SetOrFail(int [] values)
        ///       {
        ///            bool test = true;
        ///            int index = 0;
        ///             foreach (ModelParameterClass MP in BaseInputs())
        ///             {
        ///             if (MP.isBaseValueInRange(values[index])) MP.Value = values[index];
        ///                    else
        ///                     {
        ///                        test = false;
        ///                         break;
        ///                     }
        ///             index++;
        ///             }
        ///             return test;
        ///        }
        ///        </code>
        ///        </example>
        public IEnumerable<ModelParameterClass> BaseInputs()
        {
            int length = NumberOfParameters();

            for (int i = 0; i < length; i++)
            {
                if (Model_Parameter((eModelParam)i).ParamType == modelParamtype.mptInputBase)
                    yield return Model_Parameter((eModelParam)i);
            }
        }
        //---------------------------------------------------------

        /// <summary>   Enumerates base outputs in this collection./ </summary>
        /// <returns>
        ///     An enumerator that allows foreach to be used to process base outputs in this collection.
        /// </returns>
        /// <example><code>
        ///     ModelParameterProviderArray MyData = NewOutputBaseArray();
        ///     int index = 0;
        ///     foreach (ModelParameterClass MP in BaseOutputs())
        ///     {
        ///     MyData.Values[index].Values = MP.ProviderProperty.getvalues();
        ///     index++;
        ///     }
        ///     </code>
        /// </example>

        public IEnumerable<ModelParameterClass> BaseOutputs()
        {
            int length = NumberOfParameters();

            for (int i = 0; i < length; i++)
            {
                if (Model_Parameter((eModelParam)i).ParamType == modelParamtype.mptOutputBase)
                    yield return Model_Parameter((eModelParam)i);
            }
        }
        //---------------------------------------------------------

        /// <summary>   Enumerates provider inputs in this collection./ </summary>
        /// <returns>
        ///     An enumerator that allows foreach to be used to process provider inputs in this
        ///     collection.
        /// </returns>
        /// <example><code>
        ///        bool SetOrFail(ProviderIntArray values)
        ///       {
        ///            bool test = true;
        ///             foreach (eProvider ep in ProviderClass.providers())
        ///                 foreach (ModelParameterClass MP in ProviderOutputs())
        ///                     if (MP.isProviderValueInRange(values[ep], ep)) MP.Value = values[ep];
        ///                    else
        ///                     {
        ///                        test = false;
        ///                         break;
        ///                     }
        ///             return test;
        ///        }
        ///        </code>
        ///        </example>

        public IEnumerable<ModelParameterClass> ProviderInputs()
        {
            int length = NumberOfParameters();

            for (int i = 0; i < length; i++)
            {
                if (Model_Parameter((eModelParam)i).ParamType == modelParamtype.mptInputProvider)
                    yield return Model_Parameter((eModelParam)i);
            }
        }
        //---------------------------------------------------------

        /// <summary>   Enumerates provider outputs in this collection. </summary>
        /// <returns>
        ///     An enumerator that allows foreach to be used to process provider outputs in this
        ///     collection.
        /// </returns>
        ///
        /// <example><code>
        ///     ModelParameterProviderArray MyData = NewOutputProviderArray();
        ///     int index = 0;
        ///     foreach (ModelParameterClass MP in ProviderOutputs())
        ///     {
        ///     MyData.Values[index].Values = MP.ProviderProperty.getvalues();
        ///     index++;
        ///     }
        ///     </code>
        /// </example>

        public IEnumerable<ModelParameterClass> ProviderOutputs()
        {
            int length = NumberOfParameters();

            for (int i = 0; i < length; i++)
            {
                if (Model_Parameter((eModelParam)i).ParamType == modelParamtype.mptOutputProvider)
                    yield return Model_Parameter((eModelParam)i);
            }
        }
        //------------------------------------------------------

        /// <summary>   Creates a new ModelParameterBaseArray for modePAramType.mptInputBase ModelParameters.  <see cref="ModelParameterProviderArray"/> </summary>
        /// <returns> ModelParameterBaseArray  . </returns>
        ///<seealso cref="ModelParameterProviderArray"/>

        public ModelParameterBaseArray NewInputBaseArray()
        {
            return new ModelParameterBaseArray(NumberOfParameters(modelParamtype.mptInputBase));
        }
        //------------------------------------------------------

        /// <summary>   Creates a new ModelParameterBaseArray for modePAramType.mptOutputBase ModelParameters.  /// <see cref="ModelParameterProviderArray"/></summary>
        /// <returns> ModelParameterBaseArray  . </returns>
        /// <seealso cref="ModelParameterProviderArray"/>

        public ModelParameterBaseArray NewOutputBaseArray()
        {
            return new ModelParameterBaseArray(NumberOfParameters(modelParamtype.mptOutputBase));
        }
        //------------------------------------------------------
        /// <summary>   Creates a new ModelParameterProviderArray for modeParamtype.mptInputProvider ModelParameters.  /// <see cref="ModelParameterProviderArray"/></summary>
        /// <returns> ModelParameterProviderArray  . </returns>
        ///  /// <seealso cref="ModelParameterProviderArray"/>

        public ModelParameterProviderArray NewInputProviderArray()
        {
            return new ModelParameterProviderArray(NumberOfParameters(modelParamtype.mptInputProvider));
        }
        //------------------------------------------------------
        /// <summary>   Creates a new ModelParameterProviderArray for modeParamtype.mptOutputProvider ModelParameters. <see cref="ModelParameterProviderArray"/></summary>
        /// <returns> ModelParameterProviderArray  . </returns>
        /// <seealso cref="ModelParameterProviderArray"/>
        public ModelParameterProviderArray NewOutputProviderArray()
        {
            return new ModelParameterProviderArray(NumberOfParameters(modelParamtype.mptOutputProvider));
        }
        //-------------------------------------------------------
        /// <summary>
        /// Creates a new SimulationInputs struct
        /// </summary>
        /// <returns>a new SimulationInputs struct</returns>
        public SimulationInputs NewSimulationInputs()
        {
            return new SimulationInputs(NumberOfParameters(modelParamtype.mptInputBase
                ), NumberOfParameters(modelParamtype.mptInputProvider));
        }
        //-------------------------------------------------------
        /// <summary>
        /// Creates a new SimulationOutputs struct
        /// </summary>
        /// <returns>a new SimulationOuput struct</returns>
        public SimulationInputs NewSimulationOutputs()
        {
            return new SimulationInputs(NumberOfParameters(modelParamtype.mptOutputBase
                ), NumberOfParameters(modelParamtype.mptOutputProvider));
        }
        //-------------------------------------------------------
        /// <summary>
        /// Gets All Simulation Input Values from the model and returns them in a SimulationInputs struct
        /// </summary>
        /// <remarks>Some foresight should be used with this method.  This reads the input parameters regardless of the state of the model.  
        /// 		 Though BaseInputs can not be changed once a Simulation has started, the provider inputs can, so the values of these can change 
        /// 		 from one annula run to another.   </remarks>
        /// <returns> all current input parmeter values in a SimulationInputs struct</returns>
        public SimulationInputs GetSimulationInputs()
        {
            SimulationInputs si = NewSimulationInputs();
            int index = 0;
            foreach (ModelParameterClass mp in BaseInputs())
            {
                si.BaseInput[index] = mp.GetInt();
                index++;
            }
            index = 0;
            ProviderIntArray pdata;
            foreach (ModelParameterClass mp in ProviderInputs())
            {
                pdata = si.ProviderInput[index];
                pdata.Values =mp.GetIntArray();
                index++;
            }
            return si;
        }
        //-------------------------------------------------------
        /// <summary>
        /// Sets all model parameter values using values in a SimulationInputs struct
        /// </summary>
        /// <remarks>Some caution should be used with this method, a true return value only means thet values were in range, not that the
        /// 		 ModelParameters were actually set.  Write access to baseinputs is blocked once a Simulation starts.  Attempting to set them at this Simulation state 
        /// 		 will not succeed, but will return a true value and evoke no exception.  </remarks>
        /// <param name="si">values to set in model</param>
        /// <returns>true if no errors, including range errors, otherwise ModelParamters will throw an exception</returns>
        /// <exception cref="WaterSim_Exception">if values in si do not pass range checks</exception>
        public bool SetSimulationInputs(SimulationInputs si)
        {
            bool noerrors = true;
            try
            {
                int index = 0;
                foreach (ModelParameterClass mp in BaseInputs())
                {
                    mp.SetInt(si.BaseInput[index]);
                    index++;
                }
                index = 0;
                //ProviderIntArray pdata;
                foreach (ModelParameterClass mp in ProviderInputs())
                {
                    //pdata = si.ProviderInput[index];
                    mp.ProviderProperty.setvalues( si.ProviderInput.Values[index]);
                    index++;
                }
            }
            catch
            {
                noerrors = false;
            }
            return noerrors;
        }
        //-------------------------------------------------------
        //-------------------------------------------------------

        /// <summary>   Retreives the ModelParameter based on eModelParam value. </summary>
        /// <exception cref="WaterSim_Exception">   Thrown when invlaid eModelParam is used for p. </exception>
        ///<remarks>This is the primary way to access specific ModelParameters.  Each ModelParameter has a unique eModelParam value that can be used to identify and access each ModelParameter.</remarks>
        /// <param name="p">    The eModelParam value. </param>
        /// <see cref="eModelParam"/>
        /// <returns>  a ModelParameterClass based on unique p value. </returns>

        public ModelParameterClass Model_Parameter(eModelParam p)
        {
            if (ModelParamClass.valid(p))
            {
                return FModelParameters[(int)p];
            }
            else
            {
                throw new WaterSim_Exception(WS_Strings.Get(WS_Strings.wsInvalid_EModelPAram) + " " + ((int)p).ToString());
            }
        }
        //---------------------------------------------------------------------------
        //Parameter Groups  

        /// <summary>  List of Reclaimed ModelParameters <remarks>used in range check to check to see if their values add up to more than 100</remarks> </summary>
        
        public eModelParam[] reclaimedchecklist = new[] {
                eModelParam.epPCT_Reclaimed_to_RO,
                eModelParam.epPCT_Reclaimed_to_DirectInject,
                eModelParam.epPCT_Reclaimed_to_Water_Supply,
                eModelParam.epPCT_Reclaimed_to_Vadose };

        /// <summary>  List of Effluent ModelParameters <remarks>used in range check to check to see if their values add up to more than 100</remarks> </summary>
        
        public eModelParam[] effluentchecklist = new[] {
 //               eModelParam.epPCT_Effluent_Reclaimed,  // opps this is mislabled, should be waterWater to reclaimed, not effluent
                eModelParam.epPCT_Effluent_to_PowerPlant,
                eModelParam.epPCT_Effluent_to_Vadose };

        /// <summary>  List of Water Use ModelParameters <remarks>used in range check to check to see if their values add up to more than 100</remarks> </summary>

        public eModelParam[] wateruserchecklist = new[] {
                eModelParam.epPCT_WaterSupply_to_ComInd,
                eModelParam.epPCT_WaterSupply_to_Residential };

        //------------------------------------------------------------------------

        /// <summary>   Parameter group total. </summary>
        ///
        /// <remarks>   Used to check if the list of ModelParamters adds up to more than 100</remarks>
        ///
        /// <param name="paramlist">    The paramlist. </param>
        /// <param name="provider">     The provider to whose values are checked. </param>
        ///
        /// <returns>   total of provider values for paramlist . </returns>

        public int eParamGroupTotal(eModelParam[] paramlist, eProvider provider)
        {
            int total = 0;
            ProviderIntArray pvalues = new ProviderIntArray(0);
            foreach (eModelParam ep in paramlist)
            {
                pvalues = Model_Parameter(ep).ProviderProperty.getvalues();
                total += pvalues[provider];
            }
            return total;
        }
        //------------------------------------------------------------------------

        /// <summary>   Parameter group message. </summary>
        /// <remarks>Primarily used to create a range error message if values of paramlist will exceed 100</remarks>
        /// <param name="paramlist">    The paramlist. </param>
        /// <param name="provider">     The provider to whose values are checked. </param>
        ///
        /// <returns>   a string that provides information on value for each ModelParameter in paramlist. </returns>

        public string eParamGroupMessage(eModelParam[] paramlist, eProvider provider)
        {
            string pmessage = "For Provider" + ProviderClass.Label(provider);
            ProviderIntArray pvalues = new ProviderIntArray(0);
            foreach (eModelParam ep in paramlist)
            {
                pvalues = Model_Parameter(ep).ProviderProperty.getvalues();
                pmessage += "[ " + Model_Parameter(ep).Label + " = " + pvalues[provider].ToString() + " ] ";
            }
            return pmessage;
        }
        ////------------------------------------------------------------------------
        /// <summary>
        ///  Range Check for Base Input Parameters
        /// </summary>
        /// <remarks>Does throw exception on failed range check</remarks>
        /// <param name="eparam">ModelParameter eModelParam identifier</param>
        /// <param name="value">value to be checked</param>
        /// <returns>true if in range, otherwise throws exception</returns>
        /// <exception cref="WaterSim_Exception"> if range check fails</exception>
        public bool CheckBaseValueRange(eModelParam eparam, int value)
        {
            string junk = "";
            bool test = CheckBaseValueRange(eparam,value,true, ref junk);
            return test;
        }
        ////------------------------------------------------------------------------
        /// <summary>
        ///  Range Check for Base Input Parameters
        /// </summary>
        /// <remarks>Does NOT throw exception on failed range check</remarks>
        /// <param name="eparam">ModelParameter eModelParam identifier</param>
        /// <param name="value">value to be checked</param>
        /// <param name="errMessage">a string that contains an error message if rnage check returns a false vakue</param>
        /// <returns>true if in range, otherwise fakse</returns>

        public bool CheckBaseValueRange(eModelParam eparam, int value, ref string errMessage)
        {
            bool test = CheckBaseValueRange(eparam, value, false, ref errMessage);
            return test;
        }
        ////------------------------------------------------------------------------
        /// <summary>
        ///  Protected Base Input Range Check
        /// </summary>
        /// <param name="eparam">ModelParameter eModelParam identifier</param>
        /// <param name="value">value to be checked</param>
        /// <param name="throwexception">if true throws and exception if range check fails, other wise returns false if range check fails</param>>
        /// <param name="errMessage">a string that contains an error message if rnage check returns a false vakue</param>
        /// <returns>true if in range, otherwise fakse</returns>
        /// <exception cref="WaterSim_Exception"> if throwexception = true and range check fails</exception>
        protected bool CheckBaseValueRange(eModelParam eparam, int value, bool throwexception, ref string errMessage)
        {
            string temp = errMessage;
            ModelParameterClass MP;
            bool test = true;
            // Do Safety Checks
            if (!ModelParamClass.valid(eparam)) throw new  WaterSim_Exception (WS_Strings.wsInvalid_EModelPAram);
            MP = Model_Parameter(eparam);
            if (!(MP.ParamType == modelParamtype.mptInputBase)) throw new WaterSim_Exception(WS_Strings.wsModelParamMUstBeInputBase);
            if (MP.ShouldCheckRange())
                 test = MP.isBaseValueInRange(value, ref temp);
            if (throwexception &(!test))  throw new WaterSim_Exception(temp);
            if (!test) errMessage = temp;
            return test;
        }
        ////------------------------------------------------------------------------
        /// <summary>
        ///  Range Check for Provider Input Parameters
        /// </summary>
        /// <remarks>Does throw exception on failed range check</remarks>
        /// <param name="eparam">ModelParameter eModelParam identifier</param>
        /// <param name="value">value to be checked</param>
        /// <param name="provider">eProvider identifier for provider</param>
        /// <returns>true if in range, otherwise throws exception</returns>
        /// <exception cref="WaterSim_Exception"> if range check fails</exception>
        /// 
        public bool CheckProviderValueRange(eModelParam eparam, int value, eProvider provider)
        {
            string junk = "";
            bool test = CheckProviderValueRange(eparam, value, provider, true, ref junk);
            return test;
        }
        ////------------------------------------------------------------------------
        /// <summary>
        ///  Range Check for Provider Input Parameters
        /// </summary>
        /// <remarks>Does NOT throw exception on failed range check</remarks>
        /// <param name="eparam">ModelParameter eModelParam identifier</param>
        /// <param name="value">value to be checked</param>
        /// <param name="provider">eProvider identifier for provider</param>
        /// <param name="errMessage">a string that contains an error message if rnage check returns a false vakue</param>
        /// <returns>true if in range, otherwise false</returns>
        public bool CheckProviderValueRange(eModelParam eparam, int value, eProvider provider, ref string errMessage)
        {
            bool test = CheckProviderValueRange(eparam, value, provider, false, ref errMessage);
            return test;
        }

        ////------------------------------------------------------------------------
        /// <summary>
        ///  protected Provider Input Range Check
        /// </summary>
        /// <param name="eparam">ModelParameter eModelParam identifier</param>
        /// <param name="value">value to be checked</param>
        /// <param name="provider">eProvider identifier for provider</param>
        /// <param name="throwexception">if true throws and exception if range check fails, other wise returns false if range check fails</param>>
        /// <param name="errMessage">a string that contains an error message if rnage check returns a false vakue</param>
        /// <returns>true if in range, otherwise fakse</returns>
        /// <exception cref="WaterSim_Exception"> if throwexception = true and range check fails</exception>
        protected bool CheckProviderValueRange(eModelParam eparam, int value, eProvider provider, bool throwexception, ref string errMessage)
        {
            string temp = errMessage;
            ModelParameterClass MP;
            bool test = false;
            // Do Safety Checks
            if (!ModelParamClass.valid(eparam)) throw new  WaterSim_Exception (WS_Strings.wsInvalid_EModelPAram);
            MP = Model_Parameter(eparam);
            if (!(MP.ParamType == modelParamtype.mptInputProvider)) throw new WaterSim_Exception(WS_Strings.wsModelParamMUstBeInputProvider);
            if (MP.ShouldCheckRange())
                 test = MP.isProviderValueInRange(value, provider, ref temp);
            if (throwexception &(!test))  throw new WaterSim_Exception(temp); 
            if (!test) errMessage = temp;
            return test;
        }
        //--------------------------------------------------

        /// <summary>   Check range for all provider values for specified ModelParameter. </summary>
        /// <param name="eparam">   ModelParameter eModelParam identifier. </param>
        /// <param name="values">   The values to check. </param>
        ///
        /// <returns>   true if all pass range check, throws an exception if any one falis. </returns>
        /// <exception cref="WaterSim_Exception"> if throwexception = true and range check fails</exception>

        public bool CheckProviderValuesRange(eModelParam eparam, ProviderIntArray values)
        {
            bool test = true;
            string junk = "";
            test = CheckProviderValuesRange(eparam, values, true, ref junk);
            return test;
        }

        //--------------------------------------------------

        /// <summary>   Protected Check range for all provider values for specified ModelParameter. </summary>
        ///
        /// <remarks>   Ray, 8/5/2011. </remarks>
        ///
        /// <exception cref="WaterSim_Exception">   Thrown if throwexception true and range check fails </exception>
        ///
        /// <param name="eparam">           ModelParameter eModelParam identifier. </param>
        /// <param name="values">           The values to check. </param>
        /// <param name="throwexception">   if throwexception = true and range check fails, other wise
        ///                                 returns false if range check fails. </param>
        /// <param name="errMessage">       [in,out] a string that contains an error message if range
        ///                                 check returns a false vakue. </param>
        ///
        /// <returns>   true if all pass range check, if range check fails, false if throwexception false, throws an exception if throwexception true. </returns>

        protected bool CheckProviderValuesRange(eModelParam eparam, ProviderIntArray values, bool throwexception, ref string errMessage)
        {
            bool isValid = true;
            bool test = true;
            string bigerrMessage = "";
            string temp = "";
            foreach (eProvider ep in ProviderClass.providers())
            {
                test = CheckProviderValueRange(eparam, values[ProviderClass.index(ep)], ep, false, ref temp);
                if (!test)
                {
                    isValid = false;
                    bigerrMessage += "[" + ProviderClass.Label(ep) + " " + temp + "]";
                }
            }
            if ((!test) & throwexception) throw new WaterSim_Exception("Invalid Value Found: " + bigerrMessage);
            return isValid;
        }
        //--------------------------------------------------
        internal bool testModelParameters(ref string iferrMessage, bool displayerrors)
        // Test to make sure all the Modelparameters are connected.
        {
            string errors = "";
            string temp = "";
            bool test = true;
           // ModelParameterClass MP;
            for (int i = 0; i < ModelParamClass.Length(); i++)
            {
                if (FModelParameters[i] == null)
                {
                    temp = ((eModelParam)i).ToString() + " is not implemented!";
                    errors += "[" + temp + "]";
                    test = false;
                }
            }
            if (!test)
            {
                iferrMessage = temp;
                if (displayerrors) System.Windows.Forms.MessageBox.Show(temp);
            }
            else
                iferrMessage = "";
            
            return test;
        }
    }
    //----------------------------------------
    // End of PramaterManagerClass
    #endregion
    //====================================================================================
    #region Report
    //----------------------------------------------------

    /// <summary>   Report header.  </summary>
    
    public class ReportHeader
    {

        /// <summary> The lines of the header, 1, 2 or 3</summary>
        public string[] Lines;

        /// <summary>   Default constructor. </summary>
  
        public ReportHeader()
        {
            Lines = new string[3];
        }
        // indexer

        /// <summary>
        ///     Indexer to get or set items within this collection using array index syntax.
        /// </summary>
        ///
        /// <value> The indexed item. </value>

        public string this[int index]
        {
            get { return Lines[index]; }
            set { Lines[index] = value; }
        }

        /// <summary>   Gets the length of the lines array 1, 2 or 3. </summary>
        ///
        /// <value> The length. </value>

        public int Length
        { get { return Lines.Length; } }
    }
    //---------------------------------------------------------

    /// <summary>   Report class.  </summary>
    /// <remarks> static class with methods for creating and processing reports and report headers</remarks>
    public static class ReportClass
    {
        const int top = 0;
        const int bottom = 1;
        //---------------------------------------------------------
        static string[] ParseLabel(string value, int size)
        {
            string[] templines = new string[2];
            if (value.Length > size)
            {
                int Len = value.Length;
                int start = (Len - size);
                int pos = 0;
                int i = start;
                do
                {
                    if (value[i] == ' ')
                        pos = i;
                    i++;
                } while ((i < Len) & (pos == 0));
                if (pos == 0)
                {
                    templines[top] = value.Substring(0, size);
                    templines[bottom] = value.Substring(size, start);
                }
                else
                {
                    templines[top] = value.Substring(0, pos);
                    pos++;
                    templines[bottom] = value.Substring(pos, (Len - pos));
                }
            }
            else
            {
                templines[bottom] = value;
                templines[top] = "";
            }

            return templines;
        }

        /// <summary> The parameter input label used for the report header, labels each paramter as input</summary>
        public static string ParamInputLabel = "Input";

        /// <summary> The parameter output label used for the report header, labels each paramter as output</summary>
        public static string ParamOutputLabel = "Output";
        //---------------------------------------------------------

        /// <summary>   Parameter header. </summary>
        ///
        /// <remarks>   Ray, 8/5/2011. </remarks>
        ///
        /// <param name="_header">      [in,out] The header. </param>
        /// <param name="ClearHeader">  true to clear header otherwise adds to what ever is passed with _header. </param>
        /// <param name="WSim">         The WaterSim class object being used for Simulation. </param>
        /// <param name="columnSize">   width of each column. </param>
        /// <param name="UseTabs">      true to use tabs otherwise pads with spaces. </param>
        /// <param name="Use2ndLine">   true to use 2nd line for label wrapping into column size, otherwise truncates label to fit column size. </param>
        /// <param name="Parms">        The parms enumerator to use. </param>

        public static void ParamHeader(ref ReportHeader _header, bool ClearHeader, WaterSim WSim, int columnSize, bool UseTabs, bool Use2ndLine, IEnumerable<ModelParameterClass> Parms)
        {
            string[] tempLines = new string[_header.Length];
            string[] parsedLines = new string[2];
            string temp = "";
            int LineNum = _header.Length;
            if (ClearHeader)   // set header to ""
                for (int i = 0; i < _header.Length; i++)
                    _header[i] = "";
            // get the input Base values
            foreach (ModelParameterClass Mp in Parms) //WSim.ParamManager.BaseInputs())
            {
                temp = Mp.Label;
                if (Mp.isInputParam) tempLines[2] = ParamInputLabel;
                else tempLines[2] = ParamOutputLabel; ;


                if ((columnSize > 0) & (temp.Length > (columnSize - 1)))
                {
                    if (Use2ndLine)
                    {
                        parsedLines = ParseLabel(temp, columnSize - 1);
                        tempLines[0] = parsedLines[0];
                        tempLines[1] = parsedLines[1];

                    }
                    else
                    {
                        tempLines[bottom] = temp.Substring(0, columnSize - 1);
                    }
                }
                else
                {
                    if (Use2ndLine)
                    {
                        tempLines[bottom] = temp;
                        tempLines[top] = "";
                    }
                    else
                    {
                        tempLines[0] = temp;
                        tempLines[1] = "";
                    }
                }

                if (!UseTabs)
                {
                    if (columnSize > 0)
                    {
                        for (int i = 0; i < LineNum; i++)
                        {
                            if (tempLines[i].Length > columnSize)
                                tempLines[i] = tempLines[i].Substring(0, columnSize - 1);
                            _header.Lines[i] += tempLines[i].PadLeft(columnSize, ' ');
                        }
                    }
                    else
                    {
                        for (int i = 0; i < LineNum; i++)
                            _header.Lines[i] += " " + tempLines[i];
                    }
                }
                else
                {
                    for (int i = 0; i < LineNum; i++)
                        _header.Lines[i] += " \t" + tempLines[i];
                }

            }
        }
        ////---------------------------------------------------------
        
        //-------------------------------------------------------------------------------------

        /// <summary>   Full header. </summary>
        ///
        /// <remarks>   Creates a full reportheader, all parameters plus year, scenario name, and provider label and proivider fieldname </remarks>
        ///
        /// <param name="WSim">         The WaterSim class object being used for Simulation. </param>
        /// <param name="columnSize">   width of each column. </param>
        /// <param name="UseTabs">      true to use tabs otherwise pads with spaces. </param>
        /// <param name="Use2ndLine">   true to use 2nd line for label wrapping into column size,
        ///                             otherwise truncates label to fit column size. </param>
        ///
        /// <returns> a ReportHeader <see cref="ReportHeader"/> </returns>

        public static ReportHeader FullHeader(WaterSim WSim, int columnSize, bool UseTabs, bool Use2ndLine)
        {
            ReportHeader temp = new ReportHeader();
            ReportHeader header = new ReportHeader();

            string pTemp = "Provider";
            string yTemp = "Year";
            string sTemp = "Scenario";
            string blank = "";

            if (columnSize != 0)
            {
                if (pTemp.Length > columnSize) pTemp = pTemp.Substring(0, columnSize - 1);
                if (yTemp.Length > columnSize) yTemp = yTemp.Substring(0, columnSize - 1);
                if (sTemp.Length > columnSize) sTemp = sTemp.Substring(0, columnSize - 1);
            }
            if (UseTabs)
            {
                pTemp = " \t" + pTemp;
                yTemp = " \t" + yTemp;
                sTemp = " \t" + sTemp;
                blank = " \t" + blank; ;
            }
            else
            {
                if (columnSize == 0)
                {
                    pTemp = " " + pTemp;
                    yTemp = " " + yTemp;
                    sTemp = " " + sTemp;
                    blank = " " + blank;
                }
                else
                {
                    pTemp = pTemp.PadLeft(columnSize, ' ');
                    yTemp = yTemp.PadLeft(columnSize, ' ');
                    sTemp = sTemp.PadLeft(columnSize, ' ');
                    blank = blank.PadLeft(columnSize, ' ');
                }

            }
            if (Use2ndLine)
            {

                header[1] = sTemp + yTemp + pTemp;
                header[0] = blank + blank + blank;
            }
            else
                header[0] = sTemp + yTemp + pTemp;


            header[2] = blank + blank + blank;
            ParamHeader(ref header, false, WSim, columnSize, UseTabs, Use2ndLine, WSim.ParamManager.BaseOutputs());
            //header[0] += temp[0];
            //header[1] += temp[1];
            ParamHeader(ref header, false, WSim, columnSize, UseTabs, Use2ndLine, WSim.ParamManager.ProviderOutputs());
            //header[0] += temp[0];
            //header[1] += temp[1];
            ParamHeader(ref header, false, WSim, columnSize, UseTabs, Use2ndLine, WSim.ParamManager.BaseInputs());
            //header[0] += temp[0];
            //header[1] += temp[1];
            ParamHeader(ref header, false, WSim, columnSize, UseTabs, Use2ndLine, WSim.ParamManager.ProviderInputs());
            //header[0] += temp[0];
            //header[1] += temp[1];

            return header;
        }

        //--------------------------------------------------------------------------------------

        /// <summary>   Annual full data. </summary>
        ///
        /// <remarks>   returns an array of strings, one string for each provider with the value of all parameters in order to match report header. Essentiall the report header strings plus these strings is all the input and output data for one Simulation year.  Reads the current data from the paramters for the WaterSim class object passed.</remarks>
        ///
        /// <param name="WSim">         The WaterSim class object being used for Simulation. </param>
        /// <param name="ScenarioName"> Name of the scenario. </param>
        /// <param name="year">         The year. </param>
        /// <param name="columnSize">   width of each column. </param>
        /// <param name="UseTabs">      true to use tabs otherwise pads with spaces. </param>
        ///
        /// <returns>   . </returns>

        public static string[] AnnualFullData(WaterSim WSim, string ScenarioName, int year, int columnSize, bool UseTabs)
        {
            ModelParameterClass Mp;
            int parmvalue = 0;
            int index = 0;
            string temp = "";
            string close = "";
            string InputBaseLine = "";
            string OutputBaseLine = "";
            string OutputProviderLine = "";
            string InputProviderLine = "";
            int ProviderSize = ProviderClass.NumberOfProviders;
            ProviderStringArray AllStrings = new ProviderStringArray("");


            if (UseTabs)
            {
                close = System.Environment.NewLine;
            }

            // get the output values

            foreach (eModelParam emp in WSim.ParamManager.ModelParameters())
            {
                Mp = WSim.ParamManager.Model_Parameter(emp);
                if (Mp.ParamType == modelParamtype.mptOutputBase)
                {
                    parmvalue = Mp.GetInt();
                    if (!UseTabs) temp = parmvalue.ToString().PadLeft(columnSize, ' ');
                    else temp = " \t" + parmvalue.ToString();
                    OutputBaseLine += temp;
                }
            }
            // get the output values
            foreach (eModelParam emp in WSim.ParamManager.ModelParameters())
            {
                Mp = WSim.ParamManager.Model_Parameter(emp);
                if (Mp.ParamType == modelParamtype.mptInputBase)
                {
                    parmvalue = Mp.GetInt();
                    if (!UseTabs) temp = parmvalue.ToString().PadLeft(columnSize, ' ');
                    else temp = " \t" + parmvalue.ToString();
                    InputBaseLine += temp;
                }
            }
            // Now Get providers Data
            ProviderIntArray ProviderArray = new ProviderIntArray(0);

            //Provider Output Data
            int OutputProviderCnt = WSim.ParamManager.NumberOfParameters(modelParamtype.mptOutputProvider);
            int[,] ProviderOutputData = new int[OutputProviderCnt, ProviderSize];//ProviderClass.NumberOfProviders);

            index = 0;
            foreach (eModelParam emp in WSim.ParamManager.ModelParameters())
            {
                Mp = WSim.ParamManager.Model_Parameter(emp);
                if (Mp.ParamType == modelParamtype.mptOutputProvider)
                {
                    ProviderArray.Values = Mp.GetIntArray();
                    foreach (eProvider p in ProviderClass.providers())
                        ProviderOutputData[index, ProviderClass.index(p)] = ProviderArray.Values[ProviderClass.index(p)];
                    index++;
                }
            }
            //Provider Input Data
            int InputProviderCnt = WSim.ParamManager.NumberOfParameters(modelParamtype.mptInputProvider);
            int[,] ProviderInputData = new int[InputProviderCnt, ProviderSize];//ProviderClass.NumberOfProviders);

            index = 0;
            foreach (eModelParam emp in WSim.ParamManager.ModelParameters())
            {
                Mp = WSim.ParamManager.Model_Parameter(emp);
                if (Mp.ParamType == modelParamtype.mptInputProvider)
                {
                    ProviderArray.Values = Mp.GetIntArray();
                    foreach (eProvider p in ProviderClass.providers())
                        ProviderInputData[index, ProviderClass.index(p)] = ProviderArray.Values[ProviderClass.index(p)];
                    index++;
                }
            }

            foreach (eProvider p in ProviderClass.providers())
            {
                // Output Line
                OutputProviderLine = "";
                for (int i = 0; i < OutputProviderCnt; i++)
                {
                    if (UseTabs) OutputProviderLine += " \t" + ProviderOutputData[i, ProviderClass.index(p)].ToString();
                    else OutputProviderLine += ProviderOutputData[i, ProviderClass.index(p)].ToString().PadLeft(columnSize, ' ');
                }
                // Input Line
                InputProviderLine = "";
                for (int i = 0; i < InputProviderCnt; i++)
                {
                    if (UseTabs) InputProviderLine += " \t" + ProviderInputData[i, ProviderClass.index(p)].ToString();
                    else InputProviderLine += ProviderInputData[i, ProviderClass.index(p)].ToString().PadLeft(columnSize, ' ');
                }
                // Get the year and provider 
                temp = ProviderClass.Label(p);
                if (temp.Length > (columnSize - 1)) temp = temp.Substring(0, columnSize - 1);
                if (UseTabs) { temp = "\t" + ScenarioName + "\t" + year.ToString() + " \t" + temp; }
                else
                {
                    temp = ScenarioName.PadLeft(columnSize, ' ') + year.ToString().PadLeft(columnSize, ' ') + temp.PadLeft(columnSize, ' ');
                }

                AllStrings.Values[ProviderClass.index(p)] = temp + OutputBaseLine + OutputProviderLine + InputBaseLine + InputProviderLine + close;
            }
            return AllStrings.Values;
        }  // FullAnnualData

    }  // Report Class
    #endregion
    //=================================================================
    #region WaterSimException
    /// <summary>   Exception for signalling water simulation errors.  </summary>

    public class WaterSim_Exception : Exception
    {

        /// <summary> Identifier string for this exception</summary>
        protected const string Pre = "WaterSim Error: ";
        internal WaterSim_Exception(string message) : base(Pre + message) { }
        internal WaterSim_Exception(int index) : base(Pre + WS_Strings.Get(index)) { }
    }

     #endregion

#region ProcessManager

    //===============================================
    // Post and Pre Process delegates, fields, and properties
    //=============================================================
    /// <summary>
    /// Enumeration of Process Type, Pre or Post <see cref="ProcessClass"/>
    /// </summary>
    /// <remarks>ProcessClass objects, or ProcessMethod delegate methods can either be evoke before the model begins executing an 
    /// annual Simulation, or after an annual Simulation is complete</remarks> 
    /// <seealso cref="ProcessClass"/>
    
    public enum ProcessType {

        /// <summary> Undefined so do not call this.  </summary>
        ptUnknown,
        /// <summary>
        /// Process runs before annual model run
        /// </summary>
        ptPre,
        /// <summary>
        /// Process runs after annual model run
        /// </summary>
        ptPost
    }

    /// <summary>
    /// Delegate for method to be called by WaterSim before and after each year is run
    /// </summary>
    /// <remarks> This method is called either before (ptPre) or after (ptPost) running each year of </remarks>
    /// <param name="year">Simulation year process evoked</param>
    /// <param name="WSim">WaterSim Object evoking process</param>
    /// <param name="pt">The process type, ProcessType.ptPre or ProcessType.ptPost</param>
    /// <returns>True if done no eros, false if not done or errors</returns>
    /// <seealso cref="ProcessType"/>
    public delegate bool ProcessMethod(int year, WaterSim WSim, ProcessType pt );
        
    //-----------------------------------------------------------------------------------
    /// <summary>
    /// A process class managed by the ProcessManager
    /// </summary>
    /// <seealso cref="ProcessManagerClass"/>
    public class ProcessClass : IDisposable
    {
        /// <summary> The process </summary>
        protected ProcessMethod FProcess = null;

        /// <summary> Type of the process </summary>
        protected ProcessType FProcessType = ProcessType.ptUnknown;

        /// <summary> The name </summary>
        protected string FName = "";

        /// <summary> Message describing the error </summary>
        protected string FErrorMessage = "";

        internal ProcessManagerClass _processManager = null;
        /// <summary>
        /// Default Constructor
        /// </summary>
        public ProcessClass() :base ()
        {
        
        }
        /// <summary>
        /// Constructor for Process Manager
        /// </summary>
        /// <param name="pm">the process method to be called<see cref="ProcessMethod"/></param>
        /// <param name="pt">the process type, pre or post <see cref="ProcessType"/></param>
        /// <param name="name">a name for the process</param>
        public ProcessClass(ProcessMethod pm, ProcessType pt, string name): base ()
        {
            FProcess = pm;
            FProcessType = pt;
            FName = name;
        }

        //-----------------------------------------
       /// <summary>    Finaliser. </summary>

       ~ProcessClass()
        {
            Dispose(false);
        }
        //------------------------------------------------------------------------------------------
        /// <summary>
        /// Called by Dispose() and ~ProcessCLass() Finalizer  Prevents Finalizer from closing model files.
        /// </summary>
        /// <param name="disposing">Finalizer will call with a false</param>
       protected virtual void Dispose(bool disposing)
       {
           if (disposing)
           {
               // check if this is being managed, if so then delete from ProcessManager
               if (_processManager != null)
               {
                   if (_processManager.inList(this)) _processManager.DeleteProcess(this);
               }
               // Ok this as well for both entry paths
               FProcess = null;
               _processManager = null;
           }
       }
       //-----------------------------------------

        /// <summary>
        ///     Called by Dispose() and ~ProcessCLass() Finalizer  .
        /// </summary>
        /// <remarks>User created ProcessClass objects should always call Dispose() before the lose scope.  </remarks>

        public void Dispose()
       {
           Dispose(true);
       }
        //-----------------------------------------
        /// <summary>
        /// Returns the process method to be called
        /// </summary>
        /// <value> a ProcessMethod delegate method</value>
        public ProcessMethod ProcessHandler
        { get { return FProcess;  } }
        
        //----------------------------------------
        /// <summary>
        /// Returns the type of process
        /// </summary>
        /// <value>either ProcessType.ptPre or ProcessType.ptPost</value>
        public ProcessType TypeOfProcess
        { get { return FProcessType;  } }
        
        //------------------------------------
        /// <summary>
        /// The name of the process
        /// </summary>
        /// <value> The name of the ProcessClass object</value>
        public string ProcessName
        { get { return FName; } }
        
        //------------------------------------
        /// <summary>
        /// Error message if process method generated an exception
        /// </summary>
        /// <value> if an exception was thrown by the process, this is the exception message</value>
        public virtual string ErrorMessage
        { get { return FErrorMessage; }
            set { FErrorMessage = value; }
        }
    }

    /// <summary>
    /// ProcessManager Class
    /// </summary>
    /// <remarks> Manages Pre and Post Processes for WaterSim Class</remarks>
   
    public class ProcessManagerClass
    {
        WaterSim FWSim;
        List<ProcessClass> FProcessList = new List<ProcessClass>();
        List<ProcessClass> _ProcessErrorList = new List<ProcessClass>();
        int _ProcessErrorCount = 0;

        //--------------------------------------------------
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="WSim"></param>
        internal ProcessManagerClass(WaterSim WSim)
        {
            FWSim = WSim;
        }
        //--------------------------------------------------
        /// <summary>
        /// Creates a ProcessClass object and adds it to the process que
        /// </summary>
        /// <param name="pm">method to call</param>
        /// <param name="pt">pre or post process type</param>
        /// <param name="name">name for process</param>
        /// <exception cref="WaterSim_Exception">if a process method of type pt is already in the process que</exception>
        
        public void AddProcess(ProcessMethod pm, ProcessType pt, string name)
        {
            if (!inList(pm, pt))
            {
                ProcessClass PC = new ProcessClass(pm, pt, name);
                PC._processManager = this;
                FProcessList.Add(PC);

            }
            else throw new WaterSim_Exception(WS_Strings.wsProcessInList + name+" "+pt.ToString());
        }
        //--------------------------------------------------
        /// <summary>
        /// Adds the ProcessClass pc to the process pool
        /// </summary>
        /// <param name="pc">ProcessClass to add</param>
        /// <exception cref="WaterSim_Exception">if a ProcessClass pc already in the process que or pc is null</exception>

        public void AddProcess(ProcessClass pc)
        {
            if ((!inList(pc))&(pc!=null))
            {
                pc._processManager = this;
                FProcessList.Add(pc);

            }
            else throw new WaterSim_Exception(WS_Strings.wsProcessInList + pc.ProcessName + " " + pc.TypeOfProcess.ToString());
        }
        ////--------------------------------------------------
        //public bool inList(ProcessMethod pm)
        //{
        //    bool found = false;
        //    found = FProcessList.Exists(delegate(ProcessClass pc) { return (pc.ProcessHandler == pm); });
        //    return found;
        //}
        //--------------------------------------------------
       
        /// <summary>
        /// Searches process que for the pm method and pt process type.
        /// </summary>
        /// <param name="pm">he process method to find</param>
        /// <param name="pt">the type of process pre or post</param>
        /// <returns>true of the process method or type pt is found</returns>
        
        public bool inList(ProcessMethod pm, ProcessType pt)
        {
            bool found = false;
            found = FProcessList.Exists(delegate(ProcessClass pc) { return ((pc.ProcessHandler == pm)&(pc.TypeOfProcess== pt)); });
            return found;
        }
        //--------------------------------------------------
 
        /// <summary>
        /// Searches process que for the pc ProcessClass.
        /// </summary>
        /// <param name="pc">he ProcessClass object to find</param>
        /// <returns>true if pc is found</returns>
        
        public bool inList(ProcessClass pc)
        {
            bool found = false;
            found = FProcessList.Exists(delegate(ProcessClass lpc) { return (lpc==pc); });
            return found;
        }
        ////--------------------------------------------------
        //public ProcessClass GetProcess(ProcessClass pc)
        //{
        //    ProcessClass found;
        //    found = FProcessList.Find(delegate(ProcessClass lpc) { return (lpc == pc); });
        //    return found;
        //}
        //--------------------------------------------------
       
        /// <summary>
        /// Gets the ProcessClass using pm and process type pt
        /// </summary>
        /// <param name="pm"></param>
        /// <param name="pt"></param>
        /// <returns>ProcessClass using pm as pt process, null if not found</returns>
        public ProcessClass GetProcess(ProcessMethod pm, ProcessType pt)
        {
            ProcessClass found = null;
            found = FProcessList.Find(delegate(ProcessClass pc) { return ((pc.ProcessHandler == pm) & (pc.TypeOfProcess == pt)); });
            return found;
        }
        //--------------------------------------------------
        /// <summary>
        /// Deletes a ProcessClass pc from the list.
        /// </summary>
        /// <param name="pc">to delete</param>
        /// <returns>true if deleted, false if not</returns>
        public bool DeleteProcess(ProcessClass pc)
        {
            return FProcessList.Remove(pc);
        }

        //--------------------------------------------------

        /// <summary>
        ///  Finds ProcessClass using pm and pt and if found deletes it
        /// </summary>
        /// <param name="pm">process method to delete</param>
        /// <param name="pt">type of process being deleted, pre or post</param>
        /// <returns>true if deleted, false if not</returns>
        public bool DeleteProcess(ProcessMethod pm, ProcessType pt)
        {
            ProcessClass pc = GetProcess(pm, pt);
            return FProcessList.Remove(pc);
        }

        //--------------------------------------------------
        /// <summary>
        /// Evokes all processs of ProcessType pt
        /// </summary>
        /// <param name="pt">the process type, pre or post</param>
        /// <param name="year">the model year for which the process is evoked</param>
        /// <returns>true if all processes returned true (no error), otherwise false <see cref="isProcessError"/><see cref="ProcessErrorList"/></returns>
        /// <remarks>If a false is returned, then one of the processes returned false (indicates error) or threw and exception , isProcessError indicates if such has occurs, ProcessErrorCount returns then number of such errors, and ProcessErrorList provides a list of the classes that generated these errors.</remarks>

        internal bool EvokeProcesses(ProcessType pt, int year)
        {
            bool executeOK = true;
            bool test = true;
            foreach (ProcessClass pc in FProcessList)
              if (pc.TypeOfProcess == pt)
              {
                  try
                  {
                      test = pc.ProcessHandler(year, FWSim, pt);
                      if (!test)
                      {
                          SetProcessError(pc, "");
                          executeOK = false;
                      }
                  }
                  catch (Exception e)
                  {
                          SetProcessError(pc, e.Message);
                          executeOK = false;
                  }
              }
            
            return executeOK;
        }
        //--------------------------------------------------
       /// <summary>
       /// Indicates if an error has occured
       /// </summary>
       /// <value>True if process error has occurred</value>

        public bool isProcessError
        { get { return _ProcessErrorCount > 0; } }
        //--------------------------------------------------
        /// <summary>
        /// Number of Process Errors that occurred
        /// </summary>
        /// <value>count of process errors, length of process error list</value>
        public int ProcessErrorCount
        { get { return _ProcessErrorCount; } }
        //--------------------------------------------------

        void SetProcessError(ProcessClass pc, string exceptionMessage)
        {
            _ProcessErrorCount++;
            pc.ErrorMessage = exceptionMessage;
            _ProcessErrorList.Add(pc);
        }
        
        //--------------------------------------------------
        /// <summary>
        /// Clear Error list
        /// </summary>
        internal void ClearProcessErrors()
        {
            _ProcessErrorList.Clear();
            _ProcessErrorCount = 0;
        }
        //--------------------------------------------------
        /// <summary>
        /// A list of processes (ProcessClass) that returned false on execution (ie indicated an error) or threw an exception
        /// </summary>
        /// <param name="index">zero based index to list</param>
        /// <returns>ProcessClass that indicated an error (false return, or threw and exeception) on execution</returns>
        /// <exception cref="WaterSim_Exception">if invalid index</exception>
        public ProcessClass ProcessErrorList(int index)
        {

            if ((index > -1) & (index < _ProcessErrorCount))
                return _ProcessErrorList[index];
            else
                throw new WaterSim_Exception("Invlaid index for Process Error list");
        }
        //--------------------------------------------------
    }
    //----------------------------------------------
    /// <summary>
    /// A ProcessClass that keeps track of ProviderDeficits;
    /// </summary>
    public class TrackProviderDeficits : ProcessClass
    {
       /// <summary>
       /// A provider array with number of years of Deficit for each provider
       /// </summary>
        public ProviderIntArray CountList = new ProviderIntArray(0);
        /// <summary>
        /// a provider array with the total AF of deficit for each provide across all years
        /// </summary>
        public ProviderDoubleArray TotalList = new ProviderDoubleArray(0.0);
        //--------------------------------------------------------------
       /// <summary>
       /// Default Constructor
       /// </summary>
        public TrackProviderDeficits()
            : base()
        {
            FName = "Count Provider Deficits";
            FProcess = CountProcessMethod;
            FProcessType = ProcessType.ptPost;
 
        }
        //----------------------------------------------------
        /// <summary>
        /// This is the ProcessMethod that keeps track of Deficits by provider
        /// </summary>
        /// <param name="year">Simulation Year (not used)</param>
        /// <param name="WSim">The WaterSim object (used to get data)</param>
        /// <param name="pt">the process stage, pre post (will be post), used to check is post</param>
        /// <returns></returns>
        public bool CountProcessMethod(int year, WaterSim WSim, ProcessType pt)
        {
            if (pt == ProcessType.ptPost)
            {
                ProviderIntArray Deficits = new ProviderIntArray();
                // get the deficit data
                Deficits = WSim.Demand_Deficit.getvalues();

                foreach (eProvider ep in ProviderClass.providers())
                {
                    if (Deficits[ep] > 0)
                    {
                        CountList[ep]++;
                        TotalList[ep] += Deficits[ep];
                    }
                }
                return true;
            }
            else
                return false;
        }
        //----------------------------------------------------

        /// <summary>   Clears the deficit counts and totals. </summary>
        /// <remarks> Called by SimulationInitialize</remarks>

        public void ClearDeficitCounts()
        {
            for (int i = 0; i < TotalList.Length; i++)
                TotalList[i] = 0.0;
            for (int i = 0; i < CountList.Length; i++)
                CountList[i] = 0;
        }
        //----------------------------------------------------

    }
#endregion
}


//================================================================================================
       
