﻿//      WaterSimDCDC Regional Water Demand and Supply Model Version 5.0

//       This is the C# API wrapper for the C# interface to the Fortran Model.

//       Copyright (C) 2011 , The Arizona Board of Regents
//              on behalf of Arizona State University

//       All rights reserved.

//       Developed by the Decision Center for a Desert City
//       Lead Model Development - David A. Sampson <david.a.sampson@asu.edu>

//       This program is free software: you can redistribute it and/or modify
//       it under the terms of the GNU General Public License version 3 as published by
//       the Free Software Foundation.

//       This program is distributed in the hope that it will be useful,
//       but WITHOUT ANY WARRANTY; without even the implied warranty of
//       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//       GNU General Public License for more details.

//       You should have received a copy of the GNU General Public License
//       along with this program.  If not, please see <http://www.gnu.org/licenses/>.
//
//====================================================================================
using System;
using System.Collections.Generic;
using System.Text;


/**************************************************
 * WaterSimDCDC_API
 * 
 * Version 9
 * 
 * 7/27/11
 * 
 * Keeper Ray Quay
 * 
 * ************************************************/
namespace WaterSimDCDC
{
    
    /// <summary>
    ///  This is the root class for interface with the WaterSim model.
    /// </summary>
    public class WaterSim : IDisposable 
    {
        // private fields
        string _TempDirectoryName;
        string _DataDirectoryName;
        
        int _StartYear = 0;
        int _EndYear = 0;                   // used to keep track of start and end year for run routines Available for new classes should not be exposed as public
        int _CurrentYear = 0;               // used to keep track of years inside run routines
        bool _inRun = false;                // used to keep track if runs have started
        bool _simulationStarted = false;    // used to keep track if Simulation has been initialized  False means not yet or stopped
        bool FModelLocked = false;                   // used to keep track if model is locked
        const string _APiVersion = "1.5";
        string _ModelBuild = "";
        static bool _FortranOutput = false;           // used to set Fortran debug mode
        static bool _isWaterSimInstatiated = false;  // used to keep track if a WAterSim object has been constructed

        // KLUDGE FOR USE GPCD
        int[] GPCDDefaults = new int[ProviderClass.NumberOfProviders];
        int[] GPCDBuffer = new int[ProviderClass.NumberOfProviders];
        internal ParameterManagerClass _pm;
        internal ProcessManagerClass _ProcessManager;
        internal WaterSimU _ws;  // this is the interface to fortran model

        //----------------------------------------------------------------------------------------------------
        // Constructor
        /// <summary>
        ///  Constructor - Loads DLL (if not yet loaded) and initializes the model, only one instance of this class can be created.  Will throw an exception if more than one is created 
        /// </summary>
        /// <param name="TempDirectoryName">This is where the model places its output files, Creates a diectoy (TempDirectoryName) if it does not exist</param>
        /// <param name="DataDirectoryName">Location of data</param>
        /// <exception cref="WaterSim_Exception"> If an object has already been instantiated</exception>
        /// <exception cref="WaterSim_Exception">if the data directory is invalid</exception>
        /// <exception cref="WaterSim_Exception">Internal Error - If Model Parameters as not initialized properly </exception>
        
        public WaterSim(string TempDirectoryName, string DataDirectoryName)
        {
            // Some Basic Tests
            if (_isWaterSimInstatiated)  throw new WaterSim_Exception(WS_Strings.wsOnlyOneObject); 
            _isWaterSimInstatiated = true;
            
            if (!TestDataDirectory(DataDirectoryName)) throw new WaterSim_Exception(WS_Strings.wsBadDataDirectory);


            // Ok Get started with create
            CreateDirectory(TempDirectoryName);
            _TempDirectoryName = TempDirectoryName;
            _DataDirectoryName = DataDirectoryName;
            // Instantiate WaterSimU
            _ws = new WaterSimU(_DataDirectoryName, _TempDirectoryName, _FortranOutput);
            // Get version
            _ModelBuild = _ws.Build;
            // Reset Model
            _ws.ResetAll();

            // Setup Parameter Manager and Model Paramters
            // Create a ParameterManager
            _pm = new ParameterManagerClass();
            // Initialize all of the Model Parameters
            initialize_ModelParameters();
            // now test that they are all there
            if (!testParmeters()) throw new WaterSim_Exception(WS_Strings.wsModelParameterMissing);
            // Ok create Process Manager
            _ProcessManager = new ProcessManagerClass(this);
            // Add any parameter Processes
            
            // Last thing before setup  Call the GPCD Fix  
            GPCDFIX();

            InitDefaultVariables();  // Sets Variables to Default Values
    
         }

        //----------------------------------------------------------------------------------------------------
        // finalizer
        // I am not sure we need this

        /// <summary>   Finaliser. </summary>
        /// <remarks>Do not want finalizer disposing of orphaned WaterSim_model.dll stuff</remarks>
        ~WaterSim()
        {
            Dispose(false);
        }
        //------------------------------------------------------------------------------------------
        /// <summary>
        /// Called by Dispose() and ~WaterSim() Finalizer  Prevents Finalizer from closing model files.
        /// </summary>
        /// <param name="disposing">Finalizer will call with a false</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                WaterSim._isWaterSimInstatiated = false;
                _ws.CloseFiles();
                

            }
            // Get rid of unmanged stuff, like dbConnections
        }
        //------------------------------------------------------------------------------------------
       /// <summary>
       /// Must be called before WaterSim object loses scope or is lost (reassigned).
       /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        /*****************************************************
         * General Properties
         * 
      
         * ***************************************************/
        /// <summary>
        /// Exposes to inherited classes for read only the private StartYear field (int) that tracks the Start Year during Simulation Runs
        /// </summary>
        protected int Sim_StartYear
        { get { return _StartYear; } }
        
        // ------------------------------------------- 
        /// <summary>
        /// Exposes to inherited class for read only the private EndYear field (int) that tracks the End Year during Simulation Runs
        /// </summary>
        protected int Sim_EndYear
        { get { return _EndYear; } }
        
        // ------------------------------------------- 
        /// <summary>
        /// Exposes to inherited class for read only the private CurrentYear field (int) that tracks the Current Year during Simulation Runs
        /// </summary>
        protected int Sim_CurrentYear
        { get { return _CurrentYear; } }
        
        // ------------------------------------------- 
        /// <summary>
        /// Provides access to the Parameter Manager
        /// </summary>
        /// <remarks> This object is used to manage the model parameters, <see cref="ParameterManagerClass"/></remarks>
        /// <seealso cref="ParameterManagerClass"/>
        /// <seealso cref="ModelParameterClass"/>
      
        public ParameterManagerClass ParamManager { get { return _pm; } }
        // ------------------------------------------- 
        /// <summary>
        /// Provides access to the Process Manager
        /// </summary>
        /// <remarks>The process manager manages the pre and post processes of yearly Simulations <see cref="ProcessManagerClass"/></remarks>
        /// <seealso cref="ProcessManagerClass"/>
        /// <seealso cref="ProcessMethod"/>
        /// <seealso cref="ProcessType"/>

        public ProcessManagerClass ProcessManager { get { return _ProcessManager; } }
        // ------------------------------------------- 
        /// <summary>
        /// Version of the API interface
        /// </summary>
        public string APiVersion { get { return _APiVersion; } }
        // ------------------------------------------- 
        /// <summary>
        /// Verson of the Fortran Model
        /// </summary>
        public string ModelBuild { get { return _ModelBuild; } }
        // ------------------------------------------- 
        /// <summary>
        /// Tells FORTRAN model to write debug files.  Must be set TRUE before WaterSim constructor is called.
        /// </summary>
        public static bool CreateModelOutputFiles
        { get { return _FortranOutput; }  set {_FortranOutput = value; } }
        // ------------------------------------------- 


        #region ClassInitialization
        /*************************************************************
         * Class Initialzation routines
         * 
         * ************************************************************/
        //----------------------------------------------------------------------------------------------------
        private void initialize_ModelParameters()
        {
            // Initialize Model ProviderArray Input properties

            Use_GPCD = new providerArrayProperty(_pm, eModelParam.epUse_GPCD, get_Use_GPCD, set_Use_GPCD);

            PCT_Wastewater_Reclaimed = new providerArrayProperty(_pm, eModelParam.epPCT_WasteWater_to_Reclaimed, get_PCT_Wastewater_Reclaimed, set_PCT_Wastewater_Reclaimed);

            PCT_Wastewater_to_Effluent = new providerArrayProperty(_pm, eModelParam.epPCT_Wastewater_to_Effluent, get_PCT_Wastewater_to_Effluent, set_PCT_Wastewater_to_Effluent);
            
            PCT_Reclaimed_to_RO = new providerArrayProperty(_pm, eModelParam.epPCT_Reclaimed_to_RO, get_PCT_Reclaimed_to_RO, set_PCT_Reclaimed_to_RO);

            PCT_RO_to_Water_Supply = new providerArrayProperty(_pm, eModelParam.epPCT_RO_to_Water_Supply, get_PCT_RO_to_Water_Supply, set_PCT_RO_to_Water_Supply);
            
            PCT_Reclaimed_to_DirectInject = new providerArrayProperty(_pm, eModelParam.epPCT_Reclaimed_to_DirectInject, get_PCT_Reclaimed_to_DirectInject, set_PCT_Reclaimed_to_DirectInject);
            
            PCT_Reclaimed_to_Water_Supply = new providerArrayProperty(_pm, eModelParam.epPCT_Reclaimed_to_Water_Supply, get_PCT_Reclaimed_to_Water_Supply, set_PCT_Reclaimed_to_Water_Supply);
            
            PCT_Reclaimed_to_Vadose = new providerArrayProperty(_pm, eModelParam.epPCT_Reclaimed_to_Vadose, get_PCT_Reclaimed_to_Vadose, set_PCT_Reclaimed_to_Vadose);
            
            PCT_Effluent_to_Vadose = new providerArrayProperty(_pm, eModelParam.epPCT_Effluent_to_Vadose, get_PCT_Effluent_to_Vadose, set_PCT_Effluent_to_Vadose);
            
            PCT_Effluent_to_PowerPlant = new providerArrayProperty(_pm, eModelParam.epPCT_Effluent_to_PowerPlant, get_PCT_Effluent_to_PowerPlant, set_PCT_Effluent_to_PowerPlant);
            
            SurfaceWater__to_Vadose = new providerArrayProperty(_pm, eModelParam.epSurfaceWater__to_Vadose, get_SurfaceWater__to_Vadose, set_SurfaceWater__to_Vadose);
            
            Surface_to_Vadose_Time_Lag = new providerArrayProperty(_pm, eModelParam.epSurface_to_Vadose_Time_Lag, get_Surface_to_Vadose_Time_Lag, set_Surface_to_Vadose_Time_Lag);

            WaterBank_Source_Option = new providerArrayProperty(_pm, eModelParam.epWaterBank_Source_Option, get_WaterBank_Source_Option, set_WaterBank_Source_Option);
            
            PCT_SurfaceWater_to_WaterBank = new providerArrayProperty(_pm, eModelParam.epPCT_SurfaceWater_to_WaterBank, get_PCT_SurfaceWater_to_WaterBank, set_PCT_SurfaceWater_to_WaterBank);
            
            Use_SurfaceWater_to_WaterBank = new providerArrayProperty(_pm, eModelParam.epUse_SurfaceWater_to_WaterBank, get_Use_SurfaceWater_to_WaterBank, set_Use_SurfaceWater_to_WaterBank);
            
            PCT_WaterSupply_to_Residential = new providerArrayProperty(_pm, eModelParam.epPCT_WaterSupply_to_Residential, get_PCT_WaterSupply_to_Residential, set_PCT_WaterSupply_to_Residential);
            
            PCT_WaterSupply_to_ComInd = new providerArrayProperty(_pm, eModelParam.epPCT_WaterSupply_to_ComInd, get_PCT_WaterSupply_to_ComInd, set_PCT_WaterSupply_to_ComInd);

            Use_WaterSupply_to_DirectInject = new providerArrayProperty(_pm, eModelParam.epUse_WaterSupply_to_DirectInject, get_Use_WaterSupply_to_DirectInject, set_Use_WaterSupply_to_DirectInject);
            
            PCT_Outdoor_WaterUse = new providerArrayProperty(_pm, eModelParam.epPCT_Outdoor_WaterUse, get_PCT_Outdoor_WaterUse, set_PCT_Outdoor_WaterUse);
            
            PCT_Groundwater_Treated = new providerArrayProperty(_pm, eModelParam.epPCT_Groundwater_Treated, get_PCT_Groundwater_Treated, set_PCT_Groundwater_Treated);
            
            PCT_Reclaimed_Outdoor_Use = new providerArrayProperty(_pm, eModelParam.epPCT_Reclaimed_Outdoor_Use, get_PCT_Reclaimed_Outdoor_Use, set_PCT_Reclaimed_Outdoor_Use);

            PCT_Growth_Rate_Adjustment = new providerArrayProperty(_pm, eModelParam.epPCT_Growth_Rate_Adjustment, get_PCT_Growth_Rate_Adjustment, set_PCT_Growth_Rate_Adjustment);

            PCT_Max_Demand_Reclaim = new providerArrayProperty(_pm, eModelParam.epPCT_Max_Demand_Reclaim, get_PCT_Max_Demand_Reclaim, set_PCT_Max_Demand_Reclaim);



            // Initialize Model ProviderArray Output properties
            Groundwater_Pumped_Municipal = new providerArrayProperty_readonly(get_Groundwater_Pumped_Municipal);
            Groundwater_Balance = new providerArrayProperty_readonly(get_Groundwater_Balance);
            SaltVerde_Annual_Deliveries_SRP = new providerArrayProperty_readonly(get_SaltVerde_Annual_Deliveries_SRP);
            SaltVerde_Class_BC_Designations = new providerArrayProperty_readonly(get_SaltVerde_Class_BC_Designations);
            Colorado_Annual_Deliveries = new providerArrayProperty_readonly(get_Colorado_Annual_Deliveries);
            Groundwater_Bank_Used = new providerArrayProperty_readonly(get_Groundwater_Bank_Used);
            Groundwater_Bank_Balance = new providerArrayProperty_readonly(get_Groundwater_Bank_Balance);
            Reclaimed_Water_Used = new providerArrayProperty_readonly(get_Reclaimed_Water_Used);
            Reclaimed_Water_To_Vadose = new providerArrayProperty_readonly(get_Reclaimed_Water_To_Vadose);
            Reclaimed_Water_Discharged = new providerArrayProperty_readonly(get_Reclaimed_Water_Discharged);
            Reclaimed_Water_to_DirectInject = new providerArrayProperty_readonly(get_Reclaimed_Water_to_DirectInject);
            RO_Reclaimed_Water_Used = new providerArrayProperty_readonly(get_RO_Reclaimed_Water_Used);
            RO_Reclaimed_Water_to_DirectInject = new providerArrayProperty_readonly(get_RO_Reclaimed_Water_to_DirectInject);
            Effluent_Created = new providerArrayProperty_readonly(get_Effluent_Created);
            Effluent_To_Vadose = new providerArrayProperty_readonly(get_Effluent_To_Vadose);
            Effluent_To_PowerPlant = new providerArrayProperty_readonly(get_Effluent_To_PowerPlant);
            Effluent_Discharged = new providerArrayProperty_readonly(get_Effluent_Discharged);
            Demand_Deficit = new providerArrayProperty_readonly(get_Demand_Deficit);
            Total_Demand = new providerArrayProperty_readonly(get_Total_Demand);
            GPCD_Used = new providerArrayProperty_readonly(get_GPCD_Used);
            Population_Used = new providerArrayProperty_readonly(get_Population_Used);


            /*************************************************************************
             * Setup all Model Parameters in the Prameter Manager
             * 
             * 
             *************************************************************************/
            // Base Inputs
          
            //// Provider Inpputs
          
            //// Provider Outputs
          
            //// Provider Outputs
          
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSimulation_Start_Year, "Simulation Start Year", "STARTYR", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 2000, 2006, geti_Simulation_Start_Year, null, seti_Simulation_Start_Year, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSimulation_End_Year, "Simulation End Year", "STOPYR", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 2012, 2085, geti_Simulation_End_Year, null, seti_Simulation_End_Year, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_Historical_Extraction_Start_Year, "Colorado Historical Extraction Start Year", "COEXTSTYR", modelParamtype.mptInputBase, rangeChecktype.rctCheckRangeSpecial, 762, 1982, geti_Colorado_Historical_Extraction_Start_Year, null, seti_Colorado_Historical_Extraction_Start_Year, null,ColoradoYearRangeCheck,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_Historical_Data_Source, "Colorado Historical Data Source", "COSRC", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 1, 3, geti_Colorado_Historical_Data_Source, null, seti_Colorado_Historical_Data_Source, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_Climate_Adjustment_Percent, "Colorado Climate Adjustment %", "COCLMADJ", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 0, 300, geti_Colorado_Climate_Adjustment_Percent, null, seti_Colorado_Climate_Adjustment_Percent, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_User_Adjustment_Percent, "Colorado User Adjustment %", "COUSRADJ", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 0, 300, geti_Colorado_User_Adjustment_Percent, null, seti_Colorado_User_Adjustment_Percent, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_User_Adjustment_StartYear, "Colorado User Adjustment Start Year", "COUSRSTR", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 2006, 2081, geti_Colorado_User_Adjustment_StartYear, null, seti_Colorado_User_Adjustment_StartYear, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_User_Adjustment_Stop_Year, "Colorado User Adjustment Stop Year", "COUSRSTP", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 2006, 2081, geti_Colorado_User_Adjustment_Stop_Year, null, seti_Colorado_User_Adjustment_Stop_Year, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_Historical_Extraction_Start_Year, "SaltVerde Historical Extraction Start Year", "SVEXTSTYR", modelParamtype.mptInputBase, rangeChecktype.rctCheckRangeSpecial, 1330, 1982, geti_SaltVerde_Historical_Extraction_Start_Year, null, seti_SaltVerde_Historical_Extraction_Start_Year, null,SaltVerdeYearRangeCheck,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_Historical_Data, "SaltVerde Historical Data Source", "SVSRC", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 1, 3, geti_SaltVerde_Historical_Data, null, seti_SaltVerde_Historical_Data, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_Climate_Adjustment_Percent, "SaltVerde Climate Adjustment %", "SVCLMADJ", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 0, 300, geti_SaltVerde_Climate_Adjustment_Percent, null, seti_SaltVerde_Climate_Adjustment_Percent, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_User_Adjustment_Percent, "SaltVerde User Adjustment %", "SVUSRADJ", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 0, 300, geti_SaltVerde_User_Adjustment_Percent, null, seti_SaltVerde_User_Adjustment_Percent, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_User_Adjustment_Start_Year, "SaltVerde User Adjustment Start Year", "SVUSRSTR", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 2006, 2081, geti_SaltVerde_User_Adjustment_Start_Year, null, seti_SaltVerde_User_Adjustment_Start_Year, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_User_Adjustment_Stop_Year, "SaltVerde User Adjustment Stop Year", "SVUSRSTP", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 2006, 2081, geti_SaltVerde_User_Adjustment_Stop_Year, null, seti_SaltVerde_User_Adjustment_Stop_Year, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epProvider_Demand_Option, "Provider Demand Option", "DMOPT", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 1, 4, geti_Provider_Demand_Option, null, seti_Provider_Demand_Option, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Reduce_GPCD, "Annual % Reduction in GPCD", "PCRDGPCD", modelParamtype.mptInputBase, rangeChecktype.rctCheckRange, 1, 75, geti_PCT_Reduce_GPCD, null, seti_PCT_Reduce_GPCD, null, null, null, null));

            
            _pm.AddParameter(new ModelParameterClass(eModelParam.epUse_GPCD, "Use GPCD", "USEGPCD", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 30, 500, null, get_Use_GPCD, null, set_Use_GPCD,null,null,Use_GPCD));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_WasteWater_to_Reclaimed, "% Effluent to Reclaimed Plant", "PCEFFREC", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_Wastewater_Reclaimed, null, set_PCT_Wastewater_Reclaimed, null, null, PCT_Wastewater_Reclaimed));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Wastewater_to_Effluent, "% Total Wastewater is Usable Effluent", "PCWWEFF", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_Wastewater_to_Effluent, null, set_PCT_Wastewater_to_Effluent,null,null,PCT_Wastewater_to_Effluent));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Reclaimed_to_RO, "% Reclaimed to RO", "PCRECRO", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRangeSpecial, 0, 100, null, get_PCT_Reclaimed_to_RO, null, set_PCT_Reclaimed_to_RO,null,PCTReclaimedRangeCheck,PCT_Reclaimed_to_RO));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_RO_to_Water_Supply, "% RO to Water Supply", "PCROWS", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_RO_to_Water_Supply, null, set_PCT_RO_to_Water_Supply,null,null,PCT_RO_to_Water_Supply));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Reclaimed_to_DirectInject, "% Reclaimed to DirectInject", "PCRECDI", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRangeSpecial, 0, 100, null, get_PCT_Reclaimed_to_DirectInject, null, set_PCT_Reclaimed_to_DirectInject,null,PCTReclaimedRangeCheck,PCT_Reclaimed_to_DirectInject));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Reclaimed_to_Water_Supply, "% Reclaimed to Water Supply", "PCERECWS", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRangeSpecial, 0, 100, null, get_PCT_Reclaimed_to_Water_Supply, null, set_PCT_Reclaimed_to_Water_Supply,null,PCTReclaimedRangeCheck,PCT_Reclaimed_to_Water_Supply));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Reclaimed_to_Vadose, "% Reclaimed to Vadose", "PCRECVAD", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRangeSpecial, 0, 100, null, get_PCT_Reclaimed_to_Vadose, null, set_PCT_Reclaimed_to_Vadose,null,PCTReclaimedRangeCheck,PCT_Reclaimed_to_Vadose));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Effluent_to_Vadose, "% Effluent to Vadose", "PCEFFVAD", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRangeSpecial, 0, 100, null, get_PCT_Effluent_to_Vadose, null, set_PCT_Effluent_to_Vadose,null,PCTEffluentRangeCheck,PCT_Effluent_to_Vadose));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Effluent_to_PowerPlant, "% Effluent to PowerPlant", "PCEFFPP", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRangeSpecial, 0, 100, null, get_PCT_Effluent_to_PowerPlant, null, set_PCT_Effluent_to_PowerPlant,null,PCTEffluentRangeCheck,PCT_Effluent_to_PowerPlant));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSurfaceWater__to_Vadose, "Amount of SurfaceWater  to Vadose", "PCSWVAD", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100000, null, get_SurfaceWater__to_Vadose, null, set_SurfaceWater__to_Vadose,null,null,SurfaceWater__to_Vadose));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSurface_to_Vadose_Time_Lag, "Surface to Vadose Time Lag in Years", "VADLAG", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 50, null, get_Surface_to_Vadose_Time_Lag, null, set_Surface_to_Vadose_Time_Lag,null,null,Surface_to_Vadose_Time_Lag));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epWaterBank_Source_Option, "WaterBank Source Option", "WBOPT", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 1, 2, null, get_WaterBank_Source_Option, null, set_WaterBank_Source_Option,null,null,WaterBank_Source_Option));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_SurfaceWater_to_WaterBank, "% SurfaceWater to WaterBank", "PCSWWB", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_SurfaceWater_to_WaterBank, null, set_PCT_SurfaceWater_to_WaterBank,null,null,PCT_SurfaceWater_to_WaterBank));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epUse_SurfaceWater_to_WaterBank, "Amount of SurfaceWater to WaterBank", "SWWB", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100000, null, get_Use_SurfaceWater_to_WaterBank, null, set_Use_SurfaceWater_to_WaterBank,null,null,Use_SurfaceWater_to_WaterBank));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_WaterSupply_to_Residential, "% WaterSupply to Residential", "PCWSRES", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRangeSpecial, 0, 100, null, get_PCT_WaterSupply_to_Residential, null, set_PCT_WaterSupply_to_Residential,null,ResComPCTRangeCheck,PCT_WaterSupply_to_Residential));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_WaterSupply_to_ComInd, "% WaterSupply to ComInd", "PCWSCOM", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_WaterSupply_to_ComInd, null, set_PCT_WaterSupply_to_ComInd,null,ResComPCTRangeCheck,PCT_WaterSupply_to_ComInd));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epUse_WaterSupply_to_DirectInject, "Amount of  Water Supply for DirectInject", "USEWSDI", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100000, null, get_Use_WaterSupply_to_DirectInject, null, set_Use_WaterSupply_to_DirectInject,null,null,Use_WaterSupply_to_DirectInject));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Outdoor_WaterUse, "% Outdoor Water Use", "PCOUTUSE", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_Outdoor_WaterUse, null, set_PCT_Outdoor_WaterUse,null,null,PCT_Outdoor_WaterUse));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Groundwater_Treated, "% Groundwater Treated", "PCGWTRT", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_Groundwater_Treated, null, set_PCT_Groundwater_Treated,null,null,PCT_Groundwater_Treated));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Reclaimed_Outdoor_Use, "% Outdoor Can Use Reclaimed", "PCRECOUT", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 100, null, get_PCT_Reclaimed_Outdoor_Use, null, set_PCT_Reclaimed_Outdoor_Use,null,null,PCT_Reclaimed_Outdoor_Use));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Growth_Rate_Adjustment, "% Growth Rate Adjustment", "PCGRWRTE", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 300, null, get_PCT_Growth_Rate_Adjustment, null, set_PCT_Growth_Rate_Adjustment,null,null,PCT_Growth_Rate_Adjustment));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPCT_Max_Demand_Reclaim, "Max % Demand Reclaimed", "PCDEMREC", modelParamtype.mptInputProvider, rangeChecktype.rctCheckRange, 0, 70, null, get_PCT_Max_Demand_Reclaim, null, set_PCT_Max_Demand_Reclaim, null, null, PCT_Max_Demand_Reclaim));


            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_River_Flow, "Colorado River Flow", "CORFLOW", modelParamtype.mptOutputBase, rangeChecktype.rctNoRangeCheck, 0, 0, geti_Colorado_River_Flow,null,null,null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epPowell_Storage, "Powell Storage", "POWSTORE", modelParamtype.mptOutputBase, rangeChecktype.rctNoRangeCheck, 0, 0, geti_Powell_Storage,null,null,null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epMead_Storage, "Mead Storage", "MEDSTORE", modelParamtype.mptOutputBase, rangeChecktype.rctNoRangeCheck, 0, 0, geti_Mead_Storage,null,null,null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_River_Flow, "SaltVerde River Flow", "SVRFLOW", modelParamtype.mptOutputBase, rangeChecktype.rctNoRangeCheck, 0, 0, geti_SaltVerde_River_Flow,null,null,null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_Storage, "SaltVerde Storage", "SVRSTORE", modelParamtype.mptOutputBase, rangeChecktype.rctNoRangeCheck, 0, 0, geti_SaltVerde_Storage,null,null,null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epEffluent_To_Agriculture, "Effluent To Agriculture", "EFLAG", modelParamtype.mptOutputBase, rangeChecktype.rctNoRangeCheck, 0, 0, geti_Effluent_To_Agriculture,null,null,null,null,null,null));


            _pm.AddParameter(new ModelParameterClass(eModelParam.epGroundwater_Pumped_Municipal, "Provider Groundwater Pumped Municipal", "MGWPUMP", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Groundwater_Pumped_Municipal, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epGroundwater_Balance, "Provider Groundwater", "GWAVAIL", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Groundwater_Balance, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_Annual_Deliveries_SRP, "SaltVerde Annual Deliveries SRP", "SRPDELIV", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_SaltVerde_Annual_Deliveries_SRP, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epSaltVerde_Class_BC_Designations, "SaltVerde Class BC Designations", "SRPBCDES", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_SaltVerde_Class_BC_Designations, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epColorado_Annual_Deliveries, "Colorado Annual Deliveries", "COLDELIV", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Colorado_Annual_Deliveries, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epGroundwater_Bank_Used, "Groundwater Bank Used", "BNKUSED", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Groundwater_Bank_Used, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epGroundwater_Bank_Balance, "Groundwater Bank Balance", "BNKAVAIL", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Groundwater_Bank_Balance, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epReclaimed_Water_Used, "Reclaimed Water Used", "RECTOT", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Reclaimed_Water_Used, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epReclaimed_Water_To_Vadose, "Reclaimed Water To Vadose", "RECVADOS", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Reclaimed_Water_To_Vadose, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epReclaimed_Water_Discharged, "Reclaimed Water Discharged", "RECDISC", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Reclaimed_Water_Discharged, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epReclaimed_Water_to_DirectInject, "Reclaimed Water to DirectInject", "RECINJEC", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Reclaimed_Water_to_DirectInject, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epRO_Reclaimed_Water_Used, "RO Reclaimed Water Used", "ROTOT", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_RO_Reclaimed_Water_Used, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epRO_Reclaimed_Water_to_DirectInject, "RO Reclaimed Water to DirectInject", "ROINJEC", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_RO_Reclaimed_Water_to_DirectInject, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epEffluent_Created, "Effluent Created", "EFLTOT", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Effluent_Created, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epEffluent_To_Vadose, "Effluent To Vadose", "EFLVADOS", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Effluent_To_Vadose, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epEffluent_To_PowerPlant, "Effluent To PowerPlant", "EFLPP", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Effluent_To_PowerPlant, null, null,null,null,null));
            _pm.AddParameter(new ModelParameterClass(eModelParam.epEffluent_Discharged, "Effluent Discharged", "EFLDISC", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Effluent_Discharged, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epDemand_Deficit, "Demand Deficit", "DEMDEF", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Demand_Deficit, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epTotal_Demand, "Total Demand", "TOTDEM", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Total_Demand, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epGPCD_Used, "GPCD Used", "GPCDUSED", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_GPCD_Used, null, null,null,null,null));

            _pm.AddParameter(new ModelParameterClass(eModelParam.epPopulation_Used, "Population", "POPUSED", modelParamtype.mptOutputProvider, rangeChecktype.rctNoRangeCheck, 0, 0, null, get_Population_Used, null, null,null,null,null));
 
        
        }

        

        //--------------------------------------------------------------     
        private void InitDefaultVariables()
        {
            //
            _ws.Startyear = _StartYear = _ws.SetSimulationStart = util.WaterSimDCDC_Default_Simulation_StartYear;
            _ws.Endyear = _EndYear = _ws.SetSimulationEnd = util.WaterSimDCDC_Default_Simulation_EndYear;

         
            // Call API.PARMS to Set up Default array
            API.parms(_ws);

            // GPCD FIX


            //TEMP FIX  CHANGE IN FUTURE IF FIXED
            Colorado_User_Adjustment_StartYear= util.WaterSimDCDC_Default_Colorado_User_Adjustment_StartYear;
            SaltVerde_Historical_Data_Source = util.WaterSimDCDC_Default_SaltVerde_Historical_Data;
            SaltVerde_User_Adjustment_Start_Year = util.WaterSimDCDC_Default_SaltVerde_User_Adjustment_Start_Year;
            Provider_Demand_Option = util.WaterSimDCDC_Provider_Demand_Option;
            // GPCDFIX
            GPCDBuffer = GPCDDefaults;

        }
        internal void GPCDFIX()
        {
            // OK This is not exactly elegant but is in place to handle odd nature of Use_GPCD and GPCD_Used.
            // There is only on parameter in the model for provider GPCD.  It can be set to a GPCD for Demand Option 4, for other options
            //  it is calculated and can not be set.  The Get always reports the GPCD_Used in the last run, which for Option 4 is what it was set to, but for all other options 
            // is a calculated value changing each year.
            // To mimic all other values two parameters were created, one input one ouput, that use the same model function, the set for the input, and the get for output.
            // The get for the input paramter uses two dummy arrays, one for option 4 which is changed on a set, and one for all other options which returns 
            // the intitial default value used by the model in calculating GPCD.  These defaults can only be changed in the provider input file.
            // A special exception is added if USE_GPCD is set when option 4 is not selected since that does not work
            // This sets up these arrays.  In order to get the default values, the model has to be run for one year.  
            // fetch the defaults
            int OldStart = Simulation_Start_Year;
            int OldEnd = Simulation_End_Year;
            Simulation_Initialize();
            Simulation_Start_Year = 2006;
            Simulation_End_Year = 2012;
            Simulation_NextYear();
            //// OK now fetch the GCPD defaults
            GPCDBuffer = GPCDDefaults = _ws.get_ProviderGPCD;
            Simulation_Stop();
            // reset the defaults
            Simulation_Start_Year = OldStart;
            Simulation_End_Year = OldEnd;
            
            // OK will use these inthe provider property routines
            // NOTE, this is one of the exceptions where a get call and a run control is made into the model outside the ModelParamter interface.

        }
        #endregion
        //------------------------------------------------------------------------
        /// <summary>
        /// Creates a Directory of directoryName if it does not exist.  Relative references is from the directory the program is executing from
        /// </summary>
        /// <param name="directoryName">Name of new or existing directory</param>
        public void CreateDirectory(string directoryName)
        {
            try
            {
                System.IO.DirectoryInfo dir = new System.IO.DirectoryInfo(directoryName);
                if (!dir.Exists)
                {
                    dir.Create();
                }
            }
            catch
            {
            }
        }
        //---------------------------------------------------
        /// <summary>
        /// Tests to make sure the specified data directoy (path) does contain input files the model needs.
        /// </summary>
        /// <param name="path">path to data directory</param>
        /// <returns>True if data files found, false if not.</returns>
        protected bool TestDataDirectory(string path)
        {
            string datafile = path + "App_Data\\Data\\Initial_storage.txt";
            return System.IO.File.Exists(datafile);
        }

        /*************************************************************
         * Simulation Control Delegates, Fields, Properties and Methods
         * 
         * **************************************************************/

        //----------------------------------------------------------------
        internal ProcessMethod _PreProcess = null;
        internal ProcessMethod _PostProcess = null;
        //----------------------------------------------------------------
        /// <summary>
        /// Called before each year run,
        /// Write Only
        /// </summary>
        public ProcessMethod PreProcess
        { set { _PreProcess = value; } }
        //----------------------------------------------------------------
        /// <summary>
        /// Called after each year run
        /// Write Only
        /// </summary>
        public ProcessMethod PostProcess
        { set { _PostProcess = value; } }
        //----------------------------------------------------------------
       
        /**************************************
          * SimulationLock , LockSimulation(), UnLockSimulation()
          *  These can be used to lock down the simultaion so no input parameters can be set
          *  In event driven environments this is advised
          * ************************************/
        //----------------------------------------------------------------------------------------------------
        /// <summary>
        ///  Property Locks and unlocks (true false) the simulation so no parameters can be set until unlocked.  
        /// </summary>
        /// <remarks>I started with this and then moved to LockSimulation() and UnlockSimulation(), may remove this at some point.</remarks>
        /// <seealso cref="UnLockSimulation"/>
        /// <seealso cref="LockSimulation"/>

        public bool SimulationLock
        // Sets Model Locl
        {
            get { return FModelLocked; }
            set { FModelLocked = value; }
        }
        //----------------------------------------------------------------------------------------------------
        /// <summary>
        /// Locks the simulation so Base Input parameters can be set until unlocked. 
        /// </summary>
        /// <seealso cref="UnLockSimulation"/>
         public void LockSimulation()   {  FModelLocked = true;  }
        //----------------------------------------------------------------------------------------------------

         /// <summary>   Unlock simulation so Base Input parameters can be set . </summary>
        /// <seealso cref="LockSimulation"/>
        public  void UnLockSimulation()  { FModelLocked = false; }
        //----------------------------------------------------------------------------------------------------
        internal bool isLocked() { return FModelLocked; }
        //----------------------------------------------------------------------------------------------------
       // bool _inRunAllYears = false;
 
       //internal virtual void RunAllYears()
       // {
       //     if (!_inRunAllYears)
       //     {
       //          LockSimulation();
       //         _inRunAllYears = true;
       //         _ws.ResetAll();
       //         _ws.StartSimulation = true;
       //         foreach (int year in simulationYears())
       //         {
       //             if (_PreProcess != null) _PreProcess(year);
       //             runYear(year);
       //             if (_PostProcess != null) _PostProcess(year);
       //         }
       //         stopSimulation();
       //         _inRunAllYears = false;
       //     }
       // }
        //----------------------------------------------------------------------------------------------------
        /// <summary>
        /// Must be called to setup a Simulation . Simulation can be run in two ways, 
        ///      calling SimulationNextYear() for each year to be run, call SimulationAllYears() which runs all the years
        ///   All simulations should be stopped with StopSimulation(), which will make sure all files are closed
        ///  Will reset SimulationLock to false;
        /// </summary>
        public virtual void Simulation_Initialize()
        /* -------------------------------------------
         * Initializes a Simulation.  Simulation can be run in two ways, 
         *     calling SimulationNextYear() for each year to be run
         *     call SimulationAllYears() which runs all the years
         *     call fails if Simulation is locked
         * All simulations should be stopped with StopSimulation(), whihc will make sure all files are closed
         * --------------------------------------------*/
        {
                SimulationLock = false;
                _inRun = false;
                _CurrentYear = 0;

                if (_simulationStarted) { Simulation_Stop(); }
                // call model reset
                _ws.ResetAll();
                // Initiallize defaults
                InitDefaultVariables();  // Sets Variables to Default Values
                // clear process errors
                _ProcessManager.ClearProcessErrors();
                // tell model and API we are ready tio start
                _ws.StartSimulation = true;
                _simulationStarted = true;
                // Ready but not yet running
        }
        //----------------------------------------------------------------------------------------------------
        bool _inSimulationNextYear = false;
        /// <summary>
        /// Runs one year of simulation, no pre or post process is evoked.
        /// </summary>
        /// <returns>The year of the simulation run</returns>
        public virtual int Simulation_NextYear()
        /* --------------------------------------------------------
         * Returns the year run, if has already run last year, returns 0;
         * Calls close files afer running last year
        ----------------------------------------------------------- */
        {
            int runyear = 0;
            if ((_simulationStarted)&(!_inSimulationNextYear))
            {
                // Keep from rentry
                _inSimulationNextYear = true;
                // Save Lock State and lock this down
                bool locked = isLocked();
                LockSimulation();
                // Check if this is the first year
                if (!_inRun)
                {
                    // OK We are starting, set the STart and End years
                    _StartYear = _CurrentYear = Simulation_Start_Year;
                    _EndYear = Simulation_End_Year;
                    _inRun = true;
                }
                else
                // OK, Not the first yeat, inc the year;
                    { _CurrentYear++; }
                // OK we should never get to this code, but to be safe, check of we have already called the last year, if so, lets shut it down
                if (_CurrentYear > _EndYear)
                {

                    _simulationStarted = false;
                    runyear = 0;
                }
                else
                {
                    //OK run a year
                    // do PreProcess (but not for first year of run
                    if (_StartYear!=_CurrentYear)
                        _ProcessManager.EvokeProcesses(ProcessType.ptPre, _CurrentYear);
                    runYear(_CurrentYear);
                    // Set STartSimulatio to false
                    _ws.StartSimulation = false;
                    // do PostProcess
                    _ProcessManager.EvokeProcesses(ProcessType.ptPost, _CurrentYear);
                    // Return the year just done
                    runyear = _CurrentYear;
                }
                SimulationLock = locked;
            }
            _inSimulationNextYear = false;
            return runyear;
        }
        //----------------------------------------------------------------

        /// <summary>   Executes the year operation. </summary>
        /// <remarks>This has gone back and forth from protected to internal, might change in future so best to keep it internal for now</remarks>
        /// <param name="year"> The year. </param>

        internal virtual void runYear(int year)
        // Iternal routine Runs one more year, call groundwater
        // Fast as possible, no error checking, no reentry block
        {
            // call model for one year
            _ws.RunOneYear();
            // clean up
            _ws.GroundWater(year);
        }
        //----------------------------------------------------------------------------------
        bool _inSimulationAllYears = false;
        /// <summary>
        /// Runs all years of the specified simulation
        /// </summary>
        /// <remarks>Runs each year.  Before each year, except the first, it calls a preprocess processes.  After each year, including the last it calls all postprocess processes.  No data is retained.  If a post process does not collect data fater each year, the only output data available will be that for the last year run.</remarks>
        public virtual void Simulation_AllYears()
        {
            
            if ((_simulationStarted)&(!_inSimulationAllYears))
            {
                _inSimulationAllYears = true;
                _StartYear = _CurrentYear = Simulation_Start_Year;
                _EndYear = Simulation_End_Year;
                _inRun = true;
                foreach (int year in simulationYears())
                {    // do PreProcess (but not for first year of run
                    if (_StartYear != _CurrentYear)
                        _ProcessManager.EvokeProcesses(ProcessType.ptPre, _CurrentYear);
                    runYear(year);
                    // do PostProcess
                    _ProcessManager.EvokeProcesses(ProcessType.ptPost, _CurrentYear);
                }
                Simulation_Stop();
                _inSimulationAllYears = false;
            }
        }
        //----------------------------------------------------------------------------------
        bool _instopSimulation = false;

        /// <summary>   Stops simulation stop </summary>
        /// <remarks> Unlocks simulation, closes model files, unlocks base input parameters</remarks>
        public virtual void Simulation_Stop()
        // Stops RunByYear() closes out simulation.
        {
            if (!_instopSimulation)
            {
                _instopSimulation = true;
                // if all years not run, run out the model
                while (_CurrentYear < Simulation_End_Year)
                {
                    Simulation_NextYear();
                }
                // Close the model files
                _ws.CloseFiles();
                // Unset lock, recurrence, and in simulation flags
                FModelLocked = false;
                _inRun = false;
                _simulationStarted = false;
                _instopSimulation = false;
            }
        }
        //----------------------------------------------------------------------------------
        // TESTING
        internal bool testParmeters(ref string Results, bool debug)
        {
            return _pm.testModelParameters(ref Results,debug);
        }
        //----------------------------------------------------------------------------------
        internal bool testParmeters()
        {
            string junk = " ";
            return _pm.testModelParameters(ref junk,false);
        }
  
        
        
        // OutPut Properties for Model Variables  OutPut after model year run

        // Regional  Water Supply Data =========================
        //---------------------------------------
        //ColoradoRiverFlow

        /// <summary>   Gets the colorado river flow. </summary>
        ///<remarks>The total annual flow in the Colorado River above Lake Powell. Units AF</remarks>
        /// <value> The colorado river flow. </value>

        public int Colorado_River_Flow
        { get { return _ws.ColoradoRiverFlow; } }

        //---------------------------------------
        //PowellStorage

        /// <summary>   Gets the powell storage. </summary>
        /// <value> The powell storage. </value>
        ///<remarks>The total water storage in Lake Powell. Units maf</remarks>

        public int Powell_Storage
        { get { return _ws.PowellStorage; } }

        //---------------------------------------
        //MeadStorage

        /// <summary>   Gets the mead storage. </summary>
        /// <value> The mead storage. </value>
        ///<remarks>The total annual water storage in Lake Mead Units maf</remarks>

        public int Mead_Storage
        { get { return _ws.MeadStorage; } }

        //---------------------------------------
        //SaltVerdeRiverFlow

        /// <summary>   Gets the salt verde river flow. </summary>
        /// <value> The salt verde river flow. </value>
        ///<remarks>The total annula flow of the Salt and Verde Rivers. Units AF</remarks>

        public int SaltVerde_River_Flow
        { get { return _ws.SaltVerdeRiverFlow; } }

        //---------------------------------------
        //SaltVerdeStorage

        /// <summary>   Gets the salt verde storage. </summary>
        /// <value> The salt verde storage. </value>
        ///<remarks>The total annual storage in the Salt River Project reservoirs. Units maf</remarks>

        public int SaltVerde_Storage
        { get { return _ws.SaltVerdeStorage; } }

        //---------------------------------------
        //EffluentToAgriculture

        /// <summary>   Gets the effluent to agriculture. </summary>
        /// <value> The effluent to agriculture. </value>
        ///<remarks>The total amount of wastewater effluent delivered to agriculural users. Units AF</remarks>

        public int Effluent_To_Agriculture
        { get { return _ws.EffluentToAgriculture; } }


        // Provider data ===========================

        //---------------------------------------
        // ProviderGWPumpedMunicipal
        private int[] get_Groundwater_Pumped_Municipal()
        { return _ws.ProviderGWPumpedMunicipal; }  

        /// <summary> The groundwater pumped municipal </summary>
        ///<remarks>The total amount of annual groundwater pumped. Units AF</remarks>

        public providerArrayProperty_readonly Groundwater_Pumped_Municipal;

        //---------------------------------------
        //ProviderGroundwater
        private int[] get_Groundwater_Balance()
        { return _ws.ProviderGroundwater; }

        /// <summary> The groundwater balance </summary>
        ///<remarks>The total groundwater supply available at end of year. Units AF</remarks>

        public providerArrayProperty_readonly Groundwater_Balance;
        
        //---------------------------------------
        //SVTAnnualDeliveriesSRP
        private int[] get_SaltVerde_Annual_Deliveries_SRP()
        { return _ws.SVTAnnualDeliveriesSRP; }

        /// <summary> The salt verde annual deliveries srp </summary>
        ///<remarks>The total annual surface water and pumped groundwater delivered by SRP. Units AF</remarks>
        /// <seealso cref="Groundwater_Bank_Used"/>
        /// 
        public providerArrayProperty_readonly SaltVerde_Annual_Deliveries_SRP;

        //---------------------------------------
        //SaltVerdeClassBCDesignations
        private int[] get_SaltVerde_Class_BC_Designations()
        { return _ws.SaltVerdeClassBCDesignations; }

        /// <summary> The salt verde class bc designations </summary>
        ///<remarks>The total annual B / C designated surface water delivered by SRP. Units AF</remarks>
 
        public providerArrayProperty_readonly SaltVerde_Class_BC_Designations;

        //---------------------------------------
        //ColoradoAnnualDeliveries
        private int[] get_Colorado_Annual_Deliveries()
        { return _ws.ColoradoAnnualDeliveries; }

        /// <summary> The colorado annual deliveries </summary>
        ///<remarks>The total annual surface water deliveries by CAP, does not included banked water. Units AF</remarks>
        /// <seealso cref="Groundwater_Bank_Used"/>
        public providerArrayProperty_readonly Colorado_Annual_Deliveries;

        //---------------------------------------
        //GroundwaterBankUsed
        private int[] get_Groundwater_Bank_Used()
        { return _ws.GroundwaterBankUsed; }

        /// <summary> The groundwater bank used </summary>
        ///<remarks>The total annual amount of water delivered from water banking facilities.  These ground water facilities are assumed to be physically  outside the providers groundwater assets and would be delivered from these remote facilities to the provider.  Some of this may be delivered through the CAP or SRP canals but is not icnluded in SRP or CAP totals. Units AF</remarks>
        /// <seealso cref="SaltVerde_Annual_Deliveries_SRP"/>
        /// <seealso cref="Colorado_Annual_Deliveries"/>
        public providerArrayProperty_readonly Groundwater_Bank_Used;

        //---------------------------------------
        //GroundwaterBankBalance
        private int[] get_Groundwater_Bank_Balance()
        { return _ws.GroundwaterBankBalance; }

        /// <summary> The groundwater bank balance </summary>
        ///<remarks>The total banked water supply available at end of year. Units AF</remarks>

        public providerArrayProperty_readonly Groundwater_Bank_Balance;

        //---------------------------------------
        //ReclaimedWaterCreated
        private int[] get_Reclaimed_Water_Created()
        { return _ws.ReclaimedWaterCreated; }

        /// <summary> The reclaimed water created </summary>
        ///<remarks>The total annual amount of reclaimed water created, used and discharged. Units AF</remarks>

        public providerArrayProperty_readonly Reclaimed_Water_Created;

        //---------------------------------------
        //ReclaimedWaterUsed
        private int[] get_Reclaimed_Water_Used()
        { return _ws.ReclaimedWaterUsed; }

        /// <summary> The reclaimed water used </summary>
        ///<remarks>The total annual amount of reclaimed water used. Units AF</remarks>
        ///<see cref="Reclaimed_Water_Created"/>

        public providerArrayProperty_readonly Reclaimed_Water_Used;

        //---------------------------------------
        //ReclaimedWaterToVadose
        private int[] get_Reclaimed_Water_To_Vadose()
        { return _ws.ReclaimedWaterToVadose; }

        /// <summary> The reclaimed water to vadose </summary>
        ///<remarks>The annual amount of reclimed water used for vadose zone recharge. Units AF</remarks>

        public providerArrayProperty_readonly Reclaimed_Water_To_Vadose;

        //---------------------------------------
        //ReclaimedWaterDischarged
        private int[] get_Reclaimed_Water_Discharged()
        { return _ws.ReclaimedWaterDischarged; }

        /// <summary> The reclaimed water discharged </summary>
        public providerArrayProperty_readonly Reclaimed_Water_Discharged;

        //---------------------------------------
        //ReclaimedWaterDirectInject
        private int[] get_Reclaimed_Water_to_DirectInject()
        { return _ws.ReclaimedWaterDirectInject; }

        /// <summary> The reclaimed water to direct inject </summary>
        public providerArrayProperty_readonly Reclaimed_Water_to_DirectInject;

        //---------------------------------------
        //ROreclaimedWaterCreated
        private int[] get_RO_Reclaimed_Water_Created()
        { return _ws.ROreclaimedWaterCreated; }

        /// <summary> The ro reclaimed water created </summary>
        public providerArrayProperty_readonly RO_Reclaimed_Water_Created;

        //---------------------------------------
        //ROreclaimedWaterUsed
        private int[] get_RO_Reclaimed_Water_Used()
        { return _ws.ROreclaimedWaterUsed; }

        /// <summary> The ro reclaimed water used </summary>
        public providerArrayProperty_readonly RO_Reclaimed_Water_Used;

        //---------------------------------------
        //ROreclaimedWaterDirectInject
        private int[] get_RO_Reclaimed_Water_to_DirectInject()
        { return _ws.ROreclaimedWaterDirectInject; }

        /// <summary> The ro reclaimed water to direct inject </summary>
        public providerArrayProperty_readonly RO_Reclaimed_Water_to_DirectInject;

        //---------------------------------------
        //EffluentCreated
        private int[] get_Effluent_Created()
        { return _ws.EffluentCreated; }

        /// <summary> The effluent created </summary>
        ///<remarks>The total annual amount of wastewater effluent produced. Units </remarks>

        public providerArrayProperty_readonly Effluent_Created;

        //---------------------------------------
        //EffluentToVadose
        private int[] get_Effluent_To_Vadose()
        { return _ws.EffluentToVadose; }

        /// <summary> The effluent to vadose </summary>
        ///<remarks>The annual amount of reclaimed water used for vadose zone recharge. Units AF</remarks>

        public providerArrayProperty_readonly Effluent_To_Vadose;

        //---------------------------------------
        //EffluentToPowerPlant
        private int[] get_Effluent_To_PowerPlant()
        { return _ws.EffluentToPowerPlant; }

        /// <summary> The effluent to power plant </summary>
        ///<remarks>The annual amount of effluent delivered to power plants. Units AF</remarks>

        public providerArrayProperty_readonly Effluent_To_PowerPlant;

        //---------------------------------------
        //EffluentDischarged
        private int[] get_Effluent_Discharged()
        { return _ws.EffluentDischarged; }

        /// <summary> The effluent discharged </summary>
        ///<remarks>The annual amount of wastewater effluent discharged to a water course (envirionment). Units AF</remarks>

        public providerArrayProperty_readonly Effluent_Discharged;

        //---------------------------------------
        //DemandDeficit
        private int[] get_Demand_Deficit()
        { return _ws.DemandDeficit; }

        /// <summary> The demand deficit </summary>
        ///<remarks>The annual difference between demand and supply (demand - supply), 0 if supply is larger than demand. Units AF.  This is a good candidate as a indicator parameter.  
        ///		   There are a number of management actions that can be taken, and reflected in the model parameters to manage water supplies and demand.  
        ///		   A community's demand exceeding supply is an indicator of either a non-sustainable water resources budget or inadequate management of available water supplies and demand.    
        ///		   Both indicate an unresolved  water resource issue for the community.  Currently there is no feedback in the model to reduce growth (thus demand growth) for a community when a demand deficit occurs, nor a feedback to allocate this growth to other providers.</remarks>
   
        public providerArrayProperty_readonly Demand_Deficit;

        //---------------------------------------
        //ProviderGPCD
        private int[] get_GPCD_Used()
        { return _ws.get_ProviderGPCD; }

        /// <summary> The gpcd used </summary>
        ///<remarks>The GPCD used to estimate demand for the completed simulation year. When Provider_Demand_Option is 1,2, or 3, this is the calculated GPCD used to estimate demand.  
        ///		   if the Provider_Demand_Option =4, this is the GPCD specified by Use_GPCD parameter. Units: gpcd-gallons per capita per daya of water use</remarks>
        /// <seealso cref="Provider_Demand_Option"/>
        /// <seealso cref="Use_GPCD"/>
        public providerArrayProperty_readonly GPCD_Used;

        //---------------------------------------
        //TotalDemands
        private int[] get_Total_Demand()
        { return _ws.WaterDemand; }  //  { return _ws.TotalDemands; }

        /// <summary> The total demand </summary>
        ///<remarks>The total annual demand from all water customers. Units AF.  This calculated based on provider demand</remarks>

        public providerArrayProperty_readonly Total_Demand;

        //---------------------------------------
        //Population 
        private int[] get_Population_Used()
        { return _ws.Populations; }

        /// <summary> The population used </summary>
        ///<remarks>The population used to estimate demand for the completed simulation year. Units people</remarks>

        public providerArrayProperty_readonly Population_Used;

        /*****************************************************************
        * Input Parameters
        * 
        * 
        * 
        * ****************************************************************/
        //===============================================================
        // Model Dummy Input Parameters  (REMEMBER TO PUT NEW ONES IN THE DEFAULT INIIALIZE ROUTINE
        private int _Simulation_Start_Year = util.WaterSimDCDC_Default_Simulation_StartYear;
        private int _Simulation_End_Year = util.WaterSimDCDC_Default_Simulation_EndYear;
  


        //---------------------------------------
        //      Start year of simulation
        // Cannot be set while simulation in progress
        // Using shadow value _Simulation_Start_Year;  no get in WaterSimU
        /// <summary>
        /// Simulation_Start_Year
        /// </summary>
        /// <value>The first year of the Simulation.</value>
        /// <remarks> The first year of the Simulation. Range = 2000 to 2006  Cannot be set after Simulation starts.</remarks>
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        
        public int Simulation_Start_Year
        {
            set
            {
                if ((!_inRun)&(!FModelLocked))                
                {
                    _pm.CheckBaseValueRange(eModelParam.epSimulation_Start_Year, value);
                    _ws.SetSimulationStart = value;
                    _Simulation_Start_Year = value;
                }
            }
            get { return _Simulation_Start_Year; }    // 
        }

        //---------------------------------------
        //End year of simulation	SetSimulationEnd
        // Cannot be set while simulation in progress
        // Using shadow value _Simulation_End_Year;  no get in WaterSimU

        /// <summary>   Gets or sets the simulation end year. </summary>
        /// <value> The simulation end year. </value>
        ///<remarks> The last year of the Simulation. Range = 2012 to 2085  Cannot be set after Simulation starts.</remarks>
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        
        public int Simulation_End_Year
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epSimulation_End_Year, value);
                    _ws.SetSimulationEnd = value + 1;  // Added this to force model to run last year;
                    _Simulation_End_Year = value;
                }
                // ELSE do we throw an exception? No Just document that it is ignored
            }
            get { return _Simulation_End_Year; }
        }

        //---------------------------------------
        //Index Year	ColoradoHistoricalExtractionStartYear
        // Cannot be set while simulation in progress
        /// <summary>   Gets or sets the colorado historical extraction start year. </summary>
        /// <value> The colorado historical extraction start year. </value>
        ///<remarks> The first year of the Colorado River flow record that will be used to create a 25 year trace to simulate river flow conditions. Special Range Check Applies See ColoradoYearRangeCheck  Cannot be set after Simulation starts.
        ///	Valid ranges are defined as follows <code>
        ///	       internal RiverRange ColoradoPaleo = new RiverRange(762, 1982);
        ///        internal RiverRange ColoradoBureau = new RiverRange(1906, 1982);</code>		</remarks>
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="RiverRange"/>
        /// <seealso cref="Colorado_Historical_Data_Source"/>
         
        public int Colorado_Historical_Extraction_Start_Year
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epColorado_Historical_Extraction_Start_Year, value);
                    _ws.ColoradoHistoricalExtractionStartYear = value;
                }
                // ELSE do we throw an exception, A Message? No Just document that it is ignored
            }
            get { return _ws.ColoradoHistoricalExtractionStartYear; }

        }
        //---------------------------------------
        //Replace File	ColoradoHistoricalData    =1, paleo data; =2, Bireau of Rec data;=3, scenario data
        // Cannot be set while simulation in progress
        // using shadow value _Colorado_Historical_Data_Source no get in WaterSimU 

        /// <summary>   Gets or sets the colorado historical data source. </summary>
        /// <value> The colorado historical data source. </value>
        /// <remarks> The source of the Colorado River flow record: Value 1 uses the Bureau of Reclamation recorded record, Value 2 uses the tree ring reconstructed paleo record, Value 3 uses a user supplied river flow trace record Range = 1 to 3  Cannot be set after Simulation starts.</remarks>
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
       
        public int Colorado_Historical_Data_Source
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epColorado_Historical_Data_Source, value);
                    // ok setting but start and stop may not be set, change these and document
                    ColoradoRiverHistoricalReSetStart(value);
                    _ws.ColoradoHistoricalData = value;
                    // _Colorado_Historical_Data_Source = value; changed 7 24 11
                }
            }
            get { return _ws.ColoradoHistoricalData; } // _Colorado_Historical_Data_Source; } changed 7 24 11
        }

        //---------------------------------------
        //Climate Adjustment	ColoradoClimateAdjustmentPercent

        /// <summary>   Gets or sets the colorado climate adjustment percent. </summary>
        /// <value> The colorado climate adjustment percent. </value>
        /// <remarks> The percent (Value=50 is 50%) which is used to modify the Colorado river flow record, simulating impacts of climate change.  Change starts at beginning of Simulation. Range = 0 to 300  Cannot be set after Simulation starts.</remarks>
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
       
        public int Colorado_Climate_Adjustment_Percent
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epColorado_Climate_Adjustment_Percent, value);
                    _ws.ColoradoClimateAdjustmentPercent = value;
                }
            }
            get { return _ws.ColoradoClimateAdjustmentPercent; }
        }
        //---------------------------------------
        //Drought-Adjustment	ColoradoUserAdjustmentPercent

        /// <summary>   Gets or sets the colorado user adjustment percent. </summary>
        /// <value> The colorado user adjustment percent. </value>
        /// <remarks> The percent (Value=50 is 50%) which is used to modify the Colorado River flow record, starting and stopping in the years specified.  This is used to simulate a drought condition.  Range = 0 to 300  Cannot be set after Simulation starts.</remarks>
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="Colorado_User_Adjustment_StartYear"/> 
        /// <seealso cref="Colorado_User_Adjustment_Stop_Year"/> 

        public int Colorado_User_Adjustment_Percent
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epColorado_User_Adjustment_Percent, value);
                    _ws.ColoradoUserAdjustmentPercent = value;
                }
            }
            get { return _ws.ColoradoUserAdjustmentPercent; }
        }
        //---------------------------------------
        //Drought-Start Year	ColoradoUserAdjustmentStartYear

        /// <summary>   Gets or sets the colorado user adjustment start year. </summary>
        /// <value> The colorado user adjustment start year. </value>
       /// <remarks> Determines the year the [Colorado User Adjustment %] will be start being applied.  Range = 2006 to 2081  Cannot be set after Simulation starts.</remarks>  
       /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
       /// <seealso cref="Colorado_User_Adjustment_Percent"/>
        /// <seealso cref="Colorado_User_Adjustment_Stop_Year"/> 

        public int Colorado_User_Adjustment_StartYear
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epColorado_User_Adjustment_StartYear, value);
                    _ws.ColoradoUserAdjustmentStartYear = value;
                }
            }
            get { return _ws.ColoradoUserAdjustmentStartYear; }
        }


        //---------------------------------------
        //Drought_Stop year	ColoradoUserAdjustmentStopYear

        /// <summary>   Gets or sets the colorado user adjustment stop year. </summary>
        /// <value> The colorado user adjustment stop year. </value>
        /// <remarks> Determines the year the [Colorado User Adjustment %] will be stopped being applied. Range = 2006 to 2081  Cannot be set after Simulation starts.</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="Colorado_User_Adjustment_Percent"/>
        /// <seealso cref="Colorado_User_Adjustment_StartYear"/> 
         
        public int Colorado_User_Adjustment_Stop_Year
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epColorado_User_Adjustment_Stop_Year, value);
                    _ws.ColoradoUserAdjustmentStopYear = value;
                }
            }
            get { return _ws.ColoradoUserAdjustmentStopYear; }
        }
        //---------------------------------------
        //Index Year	SaltVerdeHistoricalExtractionStartYear

        /// <summary>   Gets or sets the salt verde historical extraction start year. </summary>
        /// <value> The salt verde historical extraction start year. </value>
        /// <remarks> The first year of the Salt Verde River flow record that will be used to create a 25 year trace to simulate river flow conditions. Special Range Check Applies See SaltVerdeYearRangeCheck  Cannot be set after Simulation starts.  
        ///	Valid ranges are defined as follows <code>
        ///   internal RiverRange SaltVerdePaleo = new RiverRange(1330, 1982);
        ///   internal RiverRange SaltVerdeBureau = new RiverRange(1945, 1982);
        ///   </code></remarks>
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="SaltVerde_Historical_Data_Source"/>
        /// <seealso cref="RiverRange"/>

        public int SaltVerde_Historical_Extraction_Start_Year
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epSaltVerde_Historical_Extraction_Start_Year, value);
                    _ws.SaltVerdeHistoricalExtractionStartYear = value;
                }
            }
            get { return _ws.SaltVerdeHistoricalExtractionStartYear; }
        }
        //---------------------------------------
        //Replace File	SaltVerdeTontoHistoricalData

        /// <summary>   Gets or sets the salt verde historical data source. </summary>
        /// <value> The salt verde historical data source. </value>
        /// <remarks> The source of the Salt Verde Rivers flow record: Value=1 uses the tree ring reconstructed paleo record, Value=2 uses the Bureau of Reclamation recorded record, Value=3 uses a user supplied river flow trace record. Range = 1 to 3  Cannot be set after Simulation starts.</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public int SaltVerde_Historical_Data_Source
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epSaltVerde_Historical_Data, value);
                    // ok setting but start and stop may not be set, change these and document
                    SaltVerdeRiverHistoricalReSetStart(value);
                    _ws.SaltVerdeTontoHistoricalData = value;
                    // _SaltVerde_Historical_Data = value;  // changed 7 24 11
                }
            }
            get { return _ws.SaltVerdeTontoHistoricalData; } // _SaltVerde_Historical_Data; } changed 7 24 11
        }
        //---------------------------------------
        //Climate Adjustment	SaltVerdeClimateAdjustmentPercent

        /// <summary>   Gets or sets the salt verde climate adjustment percent. </summary>
        ///
        /// <value> The salt verde climate adjustment percent. </value>
         /// <remarks> The percent (Value=50 is 50%) which is used to modify the Salt Verde River flow record, simulating impacts of climate change.  Change starts at beginning of Simulation. Range = 0 to 300  Cannot be set after Simulation starts.</remarks>  
          /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public int SaltVerde_Climate_Adjustment_Percent
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epSaltVerde_Climate_Adjustment_Percent, value);
                    _ws.SaltVerdeClimateAdjustmentPercent = value;
                }
            }
            get { return _ws.SaltVerdeClimateAdjustmentPercent; }
        }
        //---------------------------------------
        //Drought-Adjustment	SaltVerdeUserAdjustmentPercent

        /// <summary>   Gets or sets the salt verde user adjustment percent. </summary>
        /// <value> The salt verde user adjustment percent. </value>
        /// <remarks> The percent (Value=50 is 50%) which is used to modify the Salt Verde River flow record, starting and stopping in the years specified.  This is used to simulate a drought condition.  Range = 0 to 300  Cannot be set after Simulation starts.</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="SaltVerde_User_Adjustment_Start_Year"/>
        /// <seealso cref="SaltVerde_User_Adjustment_Stop_Year"/>
        public int SaltVerde_User_Adjustment_Percent
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epSaltVerde_User_Adjustment_Percent, value);
                    _ws.SaltVerdeUserAdjustmentPercent = value;
                }
            }
            get { return _ws.SaltVerdeUserAdjustmentPercent; }
        }
        //---------------------------------------
        //Drought-Start Year	SaltVerdeUserAdjustmentStartYear

        /// <summary>   Gets or sets the salt verde user adjustment start year. </summary>
        /// <value> The salt verde user adjustment start year. </value>
         /// <remarks> Determines the year the [SaltVerde User Adjustment %] will be start being applied.  Range = 2006 to 2081  Cannot be set after Simulation starts.</remarks>  
         /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="SaltVerde_User_Adjustment_Stop_Year"/>
        /// <seealso cref="SaltVerde_User_Adjustment_Percent"/>

        public int SaltVerde_User_Adjustment_Start_Year
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epSaltVerde_User_Adjustment_Start_Year, value);
                    _ws.SaltVerdeUserAdjustmentStartYear = value;
                }
            }
            get { return _ws.SaltVerdeUserAdjustmentStartYear; }
        }
        //---------------------------------------
        //Drought_Stop year	SaltVerdeUserAdjustmentStopYear

        /// <summary>   Gets or sets the salt verde user adjustment stop year. </summary>
        /// <value> The salt verde user adjustment stop year. </value>
        /// <remarks> Determines the year the [SaltVerde User Adjustment %] will be stopped being applied. Range = 2006 to 2081  Cannot be set after Simulation starts.</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>"
        /// <seealso cref="SaltVerde_User_Adjustment_Start_Year"/>
        /// <seealso cref="SaltVerde_User_Adjustment_Percent"/>
        public int SaltVerde_User_Adjustment_Stop_Year
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epSaltVerde_User_Adjustment_Stop_Year, value);
                    _ws.SaltVerdeUserAdjustmentStopYear = value;
                }
            }
            get { return _ws.SaltVerdeUserAdjustmentStopYear; }
        }

        //---------------------------------------
        //Water demand option	ProviderDemandOption

        /// <summary> Demand from file </summary>
        /// <remarks> Use with Provider_Demand_Option <see cref="Provider_Demand_Option"/></remarks> <seealso cref="Provider_Demand_Option"/>
        public const int pdoDemandFromFile = 1;

        /// <summary> Average gpcd and pop </summary>
        /// <remarks> Use with Provider_Demand_Option <see cref="Provider_Demand_Option"/></remarks> <seealso cref="Provider_Demand_Option"/>
        public const int pdoAverageGPCDandPOP = 2;

        /// <summary> Declining gpcd and pop </summary>
        /// <remarks> Use with Provider_Demand_Option <see cref="Provider_Demand_Option"/></remarks> <seealso cref="Provider_Demand_Option"/>
        public const int pdoDecliningGPCDandPOP = 3;

        /// <summary> User gpcd and pop </summary>
        /// <remarks> Use with Provider_Demand_Option <see cref="Provider_Demand_Option"/></remarks> <seealso cref="Provider_Demand_Option"/>
        public const int pdoUserGPCDandPOP = 4;

        //public int Provider_Demand_Option
        //// Cannot set this while simulation running
        //{
        //    set
        //    {
        //        if ((!_inRun) & (!FModelLocked))
        //        {
        //            _pm.CheckBaseValueRange(eModelParam.epProvider_Demand_Option, value);
        //            _ws.ProviderDemandOption = value;
        //            _ProviderDemandOption = value;
        //        }
        //    }
        //    get { return  _ProviderDemandOption; }
        //}
    
        // Changed 7 24 11

        /// <summary>   Gets or sets the provider demand option. </summary>
        ///
        /// <value> The provider demand option. </value>
        /// <remarks> The method that will be used to estimate annual demand  for all providers.  Value=1 reads demand values from an external file, Value=2 calculates demand based on a six year average GPCD and population read from a file, Value=3 estimates demand based on population estimates read from and external file and a smoothing function that slowly declines GPCD, Value=4 uses same method as 3, but allows the user to chnage the GPCD with Use_GPCD Paramtere <see cref="Use_GPCD"/>used for each provider at the beginning or during the simulation. Range = 1 to 4  Cannot be set after Simulation starts.</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="Use_GPCD"/>

        public int Provider_Demand_Option
        {
            set
            {
                    _pm.CheckBaseValueRange(eModelParam.epProvider_Demand_Option, value);
                    _ws.ProviderDemandOption = value;
                }

            get { return _ws.ProviderDemandOption;} 
        }
        //===========================================================

        /// <summary>   Provider demand option label. </summary>
        /// <param name="value"> provider demand option value. </param>
        /// <returns>a string lable for the provider demand option values   </returns>

        public string Provider_Demand_Option_Label(int value)
        {
            _pm.CheckBaseValueRange(eModelParam.epProvider_Demand_Option, value);
            switch (value)
            {
                case pdoDemandFromFile:
                    return "Demand from file";

                case pdoAverageGPCDandPOP:
                    return "6 Year Average GPCD and Population";

                case pdoDecliningGPCDandPOP:
                    return "Declining GPCD and Population";

                case pdoUserGPCDandPOP:
                    return "User Assigned GPCD";

                default:
                    return "Bad Value";

            }
        }
        //---------------------------------------------------------------------------------------------------------------
        //ReduceGPCDpct
        // Fix for no get on GOCD
        int ReduceGPCDValue = util.DefaultReduceGPCDValue;

        /// <summary>   Gets or sets the Percent to Reduce GPCD. </summary>
        /// <value> The percent reduction in gpcd. </value>
        /// <remarks>The amount by which GPCD is expected to decrease by the end  of the simulation (i.e., 2085). 
        /// 		 Use this when Provider Demand Option is = 3 or Provider Demand Option=4.</remarks>
        /// <seealso cref="Provider_Demand_Option"/>
        /// 
        public int PCT_Reduce_GPCD
        {
            set
            {
                if ((!_inRun) & (!FModelLocked))
                {
                    _pm.CheckBaseValueRange(eModelParam.epPCT_Reduce_GPCD, value);
                    _ws.ReduceGPCDpct = value;
                    // Store is this shadow value becuase model does not have one.
                    ReduceGPCDValue = value;
                }
            }
            get { return ReduceGPCDValue;}
        }

        
        //=====================================
        // Provider Inputs
        //=====================================


        //WWtoRWWTP	parmWWtoRWWTPpct
        internal int[] get_PCT_Wastewater_Reclaimed() { return _ws.get_parmWWtoRWWTPpct; }
        internal void set_PCT_Wastewater_Reclaimed(int[] value)
        {  //pm.CheckValueRange(eModelParam.epPCT_Effluent_Reclaimed, value, true);    
            if (!FModelLocked) _ws.get_parmWWtoRWWTPpct = value;
        }

        /// <summary> The percent wastewater reclaimed </summary>
         /// <remarks> The percent of wasterwater that is sent to a Reclaim Plant, Range 0 to 100 </remarks>  
          /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty PCT_Wastewater_Reclaimed;

        //---------------------------------------
        //WWtoEffluent	parmWWtoEffluentPct
        internal int[] get_PCT_Wastewater_to_Effluent() { return _ws.get_parmWWtoEffluentPct; }
        internal void set_PCT_Wastewater_to_Effluent(int[] value) { if (!FModelLocked) _ws.get_parmWWtoEffluentPct = value; }

        /// <summary> The percent wastewater to effluent </summary>
         /// <remarks> The percent of wasterwater effluent that is used and not discharged into a water course. Range = 0 to 100</remarks>  
         /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty PCT_Wastewater_to_Effluent;

        //---------------------------------------
        //RWWtoRO	parmReclaimedWWtoROpct
        internal int[] get_PCT_Reclaimed_to_RO() { return _ws.get_parmReclaimedWWtoROpct; }
        internal void set_PCT_Reclaimed_to_RO(int[] value) { if (!FModelLocked)  _ws.get_parmReclaimedWWtoROpct = value; }

        /// <summary> The percent reclaimed to Reverse Osmosis </summary>
        /// <remarks> The percent of  reclaimed water that is sent to a Reverse Osmosis Plant (becomes potable). Special Range Check Applies See PCTReclaimedRangeCheck</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="PCTReclaimedRangeCheck"/>

        public providerArrayProperty PCT_Reclaimed_to_RO;

        //---------------------------------------
        //ROtoOutput	parmROReclaimedToOutputPct
        internal int[] get_PCT_RO_to_Water_Supply() { return _ws.get_parmROReclaimedToOutputPct; }
        internal void set_PCT_RO_to_Water_Supply(int[] value) { if (!FModelLocked)  _ws.get_parmROReclaimedToOutputPct = value; }

        /// <summary> The percent reverse osmosis water to water supply </summary>
        /// <remarks> The percent of  water from Reverse Osmosis Plant that is used for potable water demand. Range = 0 to 100</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        public providerArrayProperty PCT_RO_to_Water_Supply;

        //---------------------------------------
        //RtoDInjection	parmReclaimedToDirectInjectPct
        internal int[] get_PCT_Reclaimed_to_DirectInject() { return _ws.get_parmReclaimedToDirectInjectPct; }
        internal void set_PCT_Reclaimed_to_DirectInject(int[] value) { if (!FModelLocked)  _ws.get_parmReclaimedToDirectInjectPct = value; }
        

        /// <summary> The percent reclaimed to direct inject </summary>
         /// <remarks> The percent of  reclaimed ater that is used to recharge an aquifer by direct injection into an aquifer. Special Range Check Applies See PCTReclaimedRangeCheck</remarks> 
         /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="PCTReclaimedRangeCheck"/>
        /// <seealso cref="ParameterManagerClass.reclaimedchecklist"/>
        public providerArrayProperty PCT_Reclaimed_to_DirectInject;

        //---------------------------------------
        //RtoOutput	parmReclaimedToOutputPct
        internal int[] get_PCT_Reclaimed_to_Water_Supply() { return _ws.get_parmReclaimedToOutputPct; }
        internal void set_PCT_Reclaimed_to_Water_Supply(int[] value) { if (!FModelLocked)  _ws.get_parmReclaimedToOutputPct = value; }

        /// <summary> The percent reclaimed to water supply </summary>
        /// <remarks> The percent of  recliamed water that is used to meet qualified user demands (non-potable). Special Range Check Applies See PCTReclaimedRangeCheck</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="PCTReclaimedRangeCheck"/>
        /// <seealso cref="ParameterManagerClass.reclaimedchecklist"/>
        public providerArrayProperty PCT_Reclaimed_to_Water_Supply;

        //---------------------------------------
        //RtoVadose	parmReclaimedToVadosePct
        internal int[] get_PCT_Reclaimed_to_Vadose() { return _ws.get_parmReclaimedToVadosePct; }
        internal void set_PCT_Reclaimed_to_Vadose(int[] value) { if (!FModelLocked)  _ws.get_parmReclaimedToVadosePct = value; }

        /// <summary> The percent reclaimed to vadose </summary>
        /// <remarks> The percent of  reclaimed water that is delivered to a vadoze zone recharge basin. Special Range Check Applies See PCTReclaimedRangeCheck</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="PCTReclaimedRangeCheck"/>
        /// <seealso cref="ParameterManagerClass.reclaimedchecklist"/>
        public providerArrayProperty PCT_Reclaimed_to_Vadose;

        //---------------------------------------
        //EffluentToVadose	parmEffluentToVadosePct
        internal int[] get_PCT_Effluent_to_Vadose() { return _ws.get_parmEffluentToVadosePct; }
        internal void set_PCT_Effluent_to_Vadose(int[] value) { if (!FModelLocked)  _ws.get_parmEffluentToVadosePct = value; }
        

        /// <summary> The percent effluent to vadose </summary>
        /// <remarks> The percent of  wastewater effluent delivered to a vadose zone recharge basin. Special Range Check Applies See PCTEffluentRangeCheck</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="PCTEffluentRangeCheck"/>
        /// <seealso cref="ParameterManagerClass.effluentchecklist"/>
        public providerArrayProperty PCT_Effluent_to_Vadose;

        //---------------------------------------
        //EffluentToPP	parmEffluentToPowerPlantPct
        internal int[] get_PCT_Effluent_to_PowerPlant() { return _ws.get_parmEffluentToPowerPlantPct; }
        internal void set_PCT_Effluent_to_PowerPlant(int[] value) { if (!FModelLocked)  _ws.get_parmEffluentToPowerPlantPct = value; }
        

        /// <summary> The percent effluent to power plant </summary>
        /// <remarks> The percent of  wastewater effluent delivered to a power plants for cooling towers. Special Range Check Applies See PCTEffluentRangeCheck</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="PCTEffluentRangeCheck"/>
        /// <seealso cref="ParameterManagerClass.effluentchecklist"/>
        public providerArrayProperty PCT_Effluent_to_PowerPlant;

        //---------------------------------------
        //SWtoVadose	parmSurfaceWaterToVadosePct
        internal int[] get_SurfaceWater__to_Vadose() { return _ws.get_parmSurfaceWaterToVadoseAmt; } // changed 7 22 11 from get_parmSurfaceWaterToVadosePCT
        internal void set_SurfaceWater__to_Vadose(int[] value) { if (!FModelLocked)  _ws.get_parmSurfaceWaterToVadoseAmt = value; } // changed 7 22 11 from get_parmSurfaceWaterToVadosePCT

        /// <summary> Surface water to vadose </summary>
        /// <remarks> The amount of surface water supply delivered to a vadose zone basin. Range = 0 to 100000</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        public providerArrayProperty SurfaceWater__to_Vadose;

        //---------------------------------------
        //SWtoWB	parmSurfaceWaterToWBankPct
        internal int[] get_PCT_SurfaceWater_to_WaterBank() { return _ws.get_parmSurfaceWaterToWbankPct; }  // _ws.get_parmSurfaceWaterToWBankPct chnaged in model 7 20 11
        internal void set_PCT_SurfaceWater_to_WaterBank(int[] value) { if (!FModelLocked)  _ws.get_parmSurfaceWaterToWbankPct = value; }  // _ws.get_parmSurfaceWaterToWBankPct chnaged in model 7 20 11
        

        /// <summary> The percent surface water to water bank </summary>
        /// <remarks> The percent of extra surface water that is sent to a water bank if [WaterBank Source Option] is set to a Value=1. Range = 0 to 100</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        public providerArrayProperty PCT_SurfaceWater_to_WaterBank;

        //---------------------------------------
        //    parmSurfaceWaterToWBankAmt
        internal int[] get_Use_SurfaceWater_to_WaterBank()
        { return _ws.get_parmSurfaceWaterToWBankAmt; }  // changed 7 26 11  API.set_parmSurfaceWaterToWbankAmt; } // Using array of myAPI as Dummy, have to think about this
        internal void set_Use_SurfaceWater_to_WaterBank(int[] value)
        {
            if (!FModelLocked)
            {
                _ws.get_parmSurfaceWaterToWBankAmt = value;
            }
        }
        /// <summary> The use surface water to water bank </summary>
        /// <remarks> The amount of water (source unknown) sent to a water bank if [WaterBank Source Option] is set to a Value=2. Range = 0 to 100000.  This parameter is essentially a way to create a new bucket of water that can be used when other supplies do not meet demand.  This could include fallowing of Colorado River farms, desalination of water in Mexico or California, sale of upper basic water rights etc.</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        public providerArrayProperty Use_SurfaceWater_to_WaterBank;

        //---------------------------------------
        //WStoRes	parmWaterSupplyToResidentialPct
        internal int[] get_PCT_WaterSupply_to_Residential() { return _ws.get_parmWaterSupplyToResidentialPct; }
        internal void set_PCT_WaterSupply_to_Residential(int[] value) { if (!FModelLocked)  _ws.get_parmWaterSupplyToResidentialPct = value; }

        /// <summary> The percent water supply to residential </summary>
         /// <remarks> The percent of total water supply used by residential customers. Special Range Check Applies See ResComPCTRangeCheck</remarks> 
         /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="ResComPCTRangeCheck"/>
        /// <seealso cref="ParameterManagerClass.wateruserchecklist"/>
        public providerArrayProperty PCT_WaterSupply_to_Residential;

        //---------------------------------------
        //WStoCio	parmWaterSupplyToCIOpct
        internal int[] get_PCT_WaterSupply_to_ComInd() { return _ws.get_parmWaterSupplyToCIOpct; }
        internal void set_PCT_WaterSupply_to_ComInd(int[] value) { if (!FModelLocked)  _ws.get_parmWaterSupplyToCIOpct = value; }

        /// <summary> The percent water supply to com ind </summary>
        /// <remarks> The percent of  total water supply used by commercial / industrail customers. Special Range Check Applies See ResComPCTRangeCheck</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
        /// <seealso cref="PCTResComRangeCheck"/>
        /// <seealso cref="ParameterManagerClass.wateruserchecklist"/>
        public providerArrayProperty PCT_WaterSupply_to_ComInd;

        /// Deleted from Model 7 22 11
        //---------------------------------------
        ////WStoDI	parmWaterSupplyToDirectInjectPct
        //internal int[] get_PCT_WaterSupply_to_DirectInject() { return _ws.get_parmWaterSupplyToDirectInjectAmt; } // changed 7 22 11 from get_parmWaterSupplyToDirectInjectPCT
        //internal void set_PCT_WaterSupply_to_DirectInject(int[] value) { if (!FModelLocked)  _ws.get_parmWaterSupplyToDirectInjectAmt = value; } // changed 7 22 11 from get_parmWaterSupplyToDirectInjectPCT
        //public providerArrayProperty PCT_WaterSupply_to_DirectInject;

        //---------------------------------------
        //OutDoorPct	parmOutdoorWaterUsePct
        internal int[] get_PCT_Outdoor_WaterUse() { return _ws.get_parmOutdoorWaterUsePct; }
        internal void set_PCT_Outdoor_WaterUse(int[] value) { if (!FModelLocked)  _ws.get_parmOutdoorWaterUsePct = value; }

        /// <summary> The percent outdoor water use </summary>
        /// <remarks> The percent of  potable water supply used for outdoor wate uses. Range = 0 to 100</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty PCT_Outdoor_WaterUse;

        //---------------------------------------
        //GWtoGWTP	parmGroundwaterToGWTPlantPct
        internal int[] get_PCT_Groundwater_Treated() { return _ws.get_parmGroundwaterToGWTPlantPct; }
        internal void set_PCT_Groundwater_Treated(int[] value) { if (!FModelLocked)  _ws.get_parmGroundwaterToGWTPlantPct = value; }

        /// <summary> The percent groundwater treated </summary>
        /// <remarks> The percent of  groundwater that is treated before it is used for potable water supply. Range = 0 to 100</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty PCT_Groundwater_Treated;

        //---------------------------------------
        //ReclaimedOutdoor	parmReclaimedOutdoorUsePct
        internal int[] get_PCT_Reclaimed_Outdoor_Use() { return _ws.get_parmReclaimedOutdoorUsePct; }
        internal void set_PCT_Reclaimed_Outdoor_Use(int[] value) { if (!FModelLocked)  _ws.get_parmReclaimedOutdoorUsePct = value; }

        /// <summary> The percent reclaimed outdoor use </summary>
        /// <remarks> The percent of  outdoor water use that can be supplied by reclaimed water. Range = 0 to 100</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>
  
        public providerArrayProperty PCT_Reclaimed_Outdoor_Use;

        // New Added to version 3

        //---------------------------------------
        // Use_WaterSupply_to_DirectInject 
        internal int[] get_Use_WaterSupply_to_DirectInject()
        { return _ws.get_parmWaterSupplyToDirectInjectAmt; } // chnaged 7 24 11  API.set_parmWaterSupplyToDirectInjectAmt; } // Using array of myAPI as Dummy, have to think about this
        internal void set_Use_WaterSupply_to_DirectInject(int[] value)
        {
            if (!FModelLocked)
            {
                _ws.get_parmWaterSupplyToDirectInjectAmt = value;
            }
        }

        /// <summary> The amount of water to direct inject </summary>
        /// <remarks> A fixed amount of potable water supply used for aquifer recharge by directly injecting into a potable aquifer. Range = 0 to 100000</remarks> 
       /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty Use_WaterSupply_to_DirectInject;

        //---------------------------------------
        // TimeLagVadoseYears  Surface_to_Vadose_Time_Lag
        internal int[] get_Surface_to_Vadose_Time_Lag()
        { return _ws.TimeLagVadoseYears; } // changed 7 26 11  API.set_TimeLagVadoseYears; } // Using array of myAPI as Dummy, have to think about this
        internal void set_Surface_to_Vadose_Time_Lag(int[] value)
        {
            if (!FModelLocked)
            {
                _ws.TimeLagVadoseYears = value;
            }
        }

        /// <summary> The surface to vadose time lag </summary>
        /// <remarks> The time in years it takes water recharged to the vadose zone to reach the aquifer used for groundwater supply. Range = 0 to 50</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty Surface_to_Vadose_Time_Lag;

        //---------------------------------------
        //ProviderGPCD
        // This is a complicated property depends on how Provider_demand_option is set

        internal int[] get_Use_GPCD()
        {
            if (Provider_Demand_Option == 4)
                return GPCDBuffer ;
            else
                return GPCDDefaults;
        }
       
        internal void set_Use_GPCD(int[] value)
        {
            if (!FModelLocked)
            {
                if (Provider_Demand_Option == 4)
                {
                    _ws.get_ProviderGPCD = value;
                    GPCDBuffer = value;
                }
                else 
                {
                   // OK ignore for now -----  throw new WaterSim_Exception(WS_Strings.wsSetGPCDError);
                }
            }
        }

        /// <summary> The gpcd to use with User Supplied GPCD</summary>
         /// <remarks> The GPCD that will be used if [Provider Demand Option] is set to Value=4. Range = 30 to 500</remarks>  
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check, for now ignoring if Provider_Demand_Option is not set to 4 and values are set</exception>
        /// <seealso cref="Provider_Demand_Option"/>
        public providerArrayProperty Use_GPCD;

        //---------------------------------------
        // get_ProviderPopGrowthRateAdjustPct
        internal int[] get_PCT_Growth_Rate_Adjustment() { return _ws.get_ProviderPopGrowthRateAdjustPct; }
        internal void set_PCT_Growth_Rate_Adjustment(int[] value) { if (!FModelLocked)  _ws.get_ProviderPopGrowthRateAdjustPct = value; }

        /// <summary> The percent growth rate adjustment </summary>
        /// <remarks> A factor (percent) that is used to adjust the rate of population growth.  Used to create faster or slower growth rate scenarios.  100 (100%) means no factor is applied.  A zero means I pack my bags and move to one of the coasts. Range = 0 to 300</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty PCT_Growth_Rate_Adjustment;

        //---------------------------------------
        //_ws.get_WaterBankingOption = set_WaterBankingOption;
        internal int[] get_WaterBank_Source_Option() { return _ws.get_WaterBankingOption; } // API.set_WaterBankingOption; } changed 7 24 11
        internal void set_WaterBank_Source_Option(int[] value)
        {
            if (!FModelLocked)
            {
                _ws.get_WaterBankingOption = value;
            }
        }
 
        /// <summary> The water bank source option </summary>
        /// <remarks> The source of water used for external water banking (outside provider groundwater): Value=1 a percent [% SurfaceWater to WaterBank] of ""unused"" surface water is sent to a water bank, Value= 2 a fixed amount[Amount of SurfaceWater to WaterBank] of an unknown extra source of water is sent to a water bank. Range = 1 to 2</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>

        public providerArrayProperty WaterBank_Source_Option;

        // Added to Model 8 10 2011
        //---------------------------------------
        //get_parmReclaimedToInputPct 
        internal int[] get_PCT_Max_Demand_Reclaim() { return _ws.get_parmReclaimedToInputPct; }
        internal void set_PCT_Max_Demand_Reclaim(int[] value) { if (!FModelLocked)  _ws.get_parmReclaimedToInputPct = value; }

        /// <summary> The percent maximum demand reclaim </summary>
        /// <remarks> The maximum percent of total demand that can be met by reclaimed water Range = 0 to 70</remarks> 
        /// <exception cref="WaterSim_Exception">if setting a value that does not pass the range check</exception>"

        public providerArrayProperty PCT_Max_Demand_Reclaim;

    

        // Gets for Base input and out Properties
        private int geti_Simulation_Start_Year() { return Simulation_Start_Year; }
        private int geti_Simulation_End_Year() { return Simulation_End_Year; }
        private int geti_Colorado_Historical_Extraction_Start_Year() { return Colorado_Historical_Extraction_Start_Year; }
        private int geti_Colorado_Historical_Data_Source() { return Colorado_Historical_Data_Source; }
        private int geti_Colorado_Climate_Adjustment_Percent() { return Colorado_Climate_Adjustment_Percent; }
        private int geti_Colorado_User_Adjustment_Percent() { return Colorado_User_Adjustment_Percent; }
        private int geti_Colorado_User_Adjustment_StartYear() { return Colorado_User_Adjustment_StartYear; }
        private int geti_Colorado_User_Adjustment_Stop_Year() { return Colorado_User_Adjustment_Stop_Year; }
        private int geti_SaltVerde_Historical_Extraction_Start_Year() { return SaltVerde_Historical_Extraction_Start_Year; }
        private int geti_SaltVerde_Historical_Data() { return SaltVerde_Historical_Data_Source; }
        private int geti_SaltVerde_Climate_Adjustment_Percent() { return SaltVerde_Climate_Adjustment_Percent; }
        private int geti_SaltVerde_User_Adjustment_Percent() { return SaltVerde_User_Adjustment_Percent; }
        private int geti_SaltVerde_User_Adjustment_Start_Year() { return SaltVerde_User_Adjustment_Start_Year; }
        private int geti_SaltVerde_User_Adjustment_Stop_Year() { return SaltVerde_User_Adjustment_Stop_Year; }
        private int geti_Provider_Demand_Option() { return Provider_Demand_Option; }
        private int geti_PCT_Reduce_GPCD() { return PCT_Reduce_GPCD; }	
        
        private int geti_Colorado_River_Flow() { return Colorado_River_Flow; }
        private int geti_Powell_Storage() { return Powell_Storage; }
        private int geti_Mead_Storage() { return Mead_Storage; }
        private int geti_SaltVerde_River_Flow() { return SaltVerde_River_Flow; }
        private int geti_SaltVerde_Storage() { return SaltVerde_Storage; }
        private int geti_Effluent_To_Agriculture() { return Effluent_To_Agriculture; }


        // Sets for Base Inputs 
        //private void set_Simulation_Start_Year(int value) { Simulation_Start_Year = value; }
        private void seti_Simulation_Start_Year(int value) { Simulation_Start_Year = value; }
        private void seti_Simulation_End_Year(int value) { Simulation_End_Year = value; }
        private void seti_Colorado_Historical_Extraction_Start_Year(int value) { Colorado_Historical_Extraction_Start_Year = value; }
        private void seti_Colorado_Historical_Data_Source(int value) { Colorado_Historical_Data_Source = value; }
        private void seti_Colorado_Climate_Adjustment_Percent(int value) { Colorado_Climate_Adjustment_Percent = value; }
        private void seti_Colorado_User_Adjustment_Percent(int value) { Colorado_User_Adjustment_Percent = value; }
        private void seti_Colorado_User_Adjustment_StartYear(int value) { Colorado_User_Adjustment_StartYear = value; }
        private void seti_Colorado_User_Adjustment_Stop_Year(int value) { Colorado_User_Adjustment_Stop_Year = value; }
        private void seti_SaltVerde_Historical_Extraction_Start_Year(int value) { SaltVerde_Historical_Extraction_Start_Year = value; }
        private void seti_SaltVerde_Historical_Data(int value) { SaltVerde_Historical_Data_Source = value; }
        private void seti_SaltVerde_Climate_Adjustment_Percent(int value) { SaltVerde_Climate_Adjustment_Percent = value; }
        private void seti_SaltVerde_User_Adjustment_Percent(int value) { SaltVerde_User_Adjustment_Percent = value; }
        private void seti_SaltVerde_User_Adjustment_Start_Year(int value) { SaltVerde_User_Adjustment_Start_Year = value; }
        private void seti_SaltVerde_User_Adjustment_Stop_Year(int value) { SaltVerde_User_Adjustment_Stop_Year = value; }
        private void seti_Provider_Demand_Option(int value) { Provider_Demand_Option = value; }
        private void seti_PCT_Reduce_GPCD(int value) { PCT_Reduce_GPCD = value; }

        // Special 

        /// <summary>   Fast get annual data.</summary>
        /// <param name="AS_Results">   [out] annual results all output parameters of Simulation. </param>

        public void FastGetAnnualData(ref AnnualSimulationResults AS_Results)
        {
            ModelParameterClass MP;
            int ib_index = 0;
            int ob_index = 0;
            int ip_index = 0;
            int op_index = 0;

            foreach (eModelParam emp in _pm.ModelParameters())
            {
                MP = _pm.Model_Parameter(emp);
                switch (MP.ParamType)
                {
                    case modelParamtype.mptInputBase :
                        AS_Results.BaseInput[ib_index] = MP.GetInt();
                        ib_index++;
                        break;
                    case modelParamtype.mptInputProvider:;
                        AS_Results.ProviderInput.Values[ip_index].Values = MP.GetIntArray();
                        ip_index++;
                        break;
                    case modelParamtype.mptOutputBase:
                        AS_Results.BaseOutput[ob_index] = MP.GetInt();
                        ob_index++;
                        break;
                    case modelParamtype.mptOutputProvider:
                        AS_Results.ProviderOutput.Values[op_index].Values = MP.GetIntArray();
                        op_index++;
                        break;

                }
            }
        }
        // SPecial Range Checks
//  Specifically;
//Bureau of Reclamation data:
//Starts: 1906 for CO river and 1945 for Salt-Verde
//Ends: 2007 for both

//Paleo Data:
//Starts: 762 for the Colorado River and 1330 for the Salt-Verde
//Ends: 2005 for both

      /// <summary> River range.  </summary>
      /// <remarks> a struct used to define the range in years of a river's flow record</remarks>
      public  class RiverRange
        {

            /// <summary> The first </summary>
            protected int _First = 0;

            /// <summary> The last </summary>
            protected int _Last = 0;
 
            /// <summary> First year of record </summary>
            /// <value>a year</value>
            public int First 
            { get { return _First; } }

            /// <summary> Last year of record </summary>
            /// <value>a year</value>
            public int Last
            { get { return _Last;} }

            internal RiverRange(int first, int last)
            {
                _First = first;
                _Last = last;
            }
        }

        /// <summary>   User river range.  </summary>
        /// <remarks>   A struct used for defining the range of user supplied river flow records. </remarks>

        public class UserRiverRange : RiverRange
        {
        
            /// <summary> First year of record </summary>
            /// <value>a year</value>
            new public int First 
            { 
                get { return _First; }
                set { _First = value; }
            }

            /// <summary> Last year of record </summary>
            /// <value>a year</value>
            new public int Last
            { 
                get { return _Last;}
                set { _Last = value; }
            }
            internal UserRiverRange(int first, int last) : base(first,last) {}
        }

        /// <summary> The Colorado River Paleo River Range</summary>
        /// <seealso cref="RiverRange"/>
        public RiverRange ColoradoPaleo = new RiverRange(762, 1982);

        /// <summary> The Colorado River Bureau/Historical River Range </summary>
        /// <seealso cref="RiverRange"/>
        public RiverRange ColoradoBureau = new RiverRange(1906, 1982);


        /// <summary> The Salt Verde Rivers Paleo/Tree Ring River Range </summary>
        /// <seealso cref="RiverRange"/>
        public RiverRange SaltVerdePaleo = new RiverRange(1330, 1982);

        /// <summary> The Salt Verde Bureau/Historical River Range </summary>
        /// <seealso cref="RiverRange"/>
        public RiverRange SaltVerdeBureau = new RiverRange(1945, 1982);


        /// <summary> Colorado river range for user supplied data</summary>
        /// <seealso cref="UserRiverRange"/>
        public UserRiverRange ColoradoUser = new UserRiverRange(1906,2005);

        /// <summary> Salt Verde river range for user supplied data</summary>
        /// <seealso cref="UserRiverRange"/>
        public UserRiverRange SaltVerdeUser = new UserRiverRange(1945, 2005);


        /// <summary> Constant for Tree Ring Paleo River Record source </summary>
        /// <value> Riversource </value>
        public const int rsPaleosource = 2;
        /// <summary> Constant for Bureau Historical River Record source </summary>
        /// <value> Riversource </value>
        public const int rsBureausource = 1;
        /// <summary> Constant for User Riversource </summary>
        /// <value> Riversource </value>
        public const int rsUsersource = 3;

        string[] SourceString = new string[] { "Source = 0", "Bureau Records", "Paleo Records", "User Records", "Source = 4", "Source = 5", "Source =6" };

        internal const int rcColorado = 1;
        internal const int rcSaltVerde = 2;
       // string[] RiverString = new string[] {"River = 0","Colorado River","Salt and Verde Rivers"};
        //---------------------------------------------------------------------
        internal void ColoradoRiverHistoricalReSetStart(int source)
        {
            int First = 0; int Last = 0;
            int start = ParamManager.Model_Parameter(eModelParam.epColorado_Historical_Extraction_Start_Year).GetInt();
            if (!RiverHistoricalYearRangeCheck(start,source,rcColorado, ref First, ref Last))
            {
                ParamManager.Model_Parameter(eModelParam.epColorado_Historical_Extraction_Start_Year).Value =First;
                ParamManager.Model_Parameter(eModelParam.epColorado_Historical_Extraction_Start_Year).EvokeReloadEvent();
            }
        }
        //---------------------------------------------------------------------
        internal void SaltVerdeRiverHistoricalReSetStart(int source)
        {
            int First = 0; int Last = 0;
            int start = ParamManager.Model_Parameter(eModelParam.epSaltVerde_Historical_Extraction_Start_Year).GetInt();
            if (!RiverHistoricalYearRangeCheck(start, source, rcSaltVerde , ref First, ref Last))
            {
                ParamManager.Model_Parameter(eModelParam.epSaltVerde_Historical_Extraction_Start_Year).Value = First;
                ParamManager.Model_Parameter(eModelParam.epSaltVerde_Historical_Extraction_Start_Year).EvokeReloadEvent();
            }
        }
        //-----------------------------------------------
        internal string RiverRangeString(int Source, int First, int Last)
        {
            string rrstr = "";
            if ((Source > 0) & (Source < SourceString.Length))
                rrstr = ": " + SourceString[Source] + " (value = " + Source.ToString() + ") start " + First.ToString() + "and end " + Last.ToString();
            return rrstr;
        }
        //-----------------------------------------------
        internal bool RiverHistoricalYearRangeCheck(int Value, int source, int River, ref int FirstYear, ref int LastYear)
        {
            int First = 0;
            int Last = 0;
            switch (River)
            {
                //------------------------------------
                case 1:
                    switch (source)
                    {
                        case rsPaleosource:
                            First = ColoradoPaleo.First;
                            Last = ColoradoPaleo.Last;
                            break;
                        //------------------
                        case rsBureausource:
                            First = ColoradoBureau.First;
                            Last = ColoradoBureau.Last;
                            break;
                        //------------------
                        case rsUsersource:
                            First = ColoradoUser.First;
                            Last = ColoradoUser.Last;
                            break;
                        //------------------
                        default:
                            break;
                        //------------------
                    } // swithc case 1 source
                    break;
                //------------------------------------
                case 2:
                    switch (source)
                    {
                        case rsPaleosource:
                            First = SaltVerdePaleo.First;
                            Last = SaltVerdePaleo.Last;
                            break;
                        //------------------
                        case rsBureausource:
                           First = SaltVerdeBureau.First;
                           Last = SaltVerdeBureau.Last;
                            break;
                        //------------------
                        case rsUsersource:
                           First = SaltVerdeUser.First;
                           Last = SaltVerdeUser.Last;
                            break;
                        //------------------
                        default:  
                            break;
                        //------------------
                    } // switch case 2 source
                    break;
                //------------------------------------
                default:
                    break;
            } // switch source
            FirstYear = First;
            LastYear = Last;
            return  ((Value >= First) & (Value <= Last));
        }
        //-----------------------------------------------

        //-----------------------------------------------
       /// <summary>
        /// ColoradoYearRangeCheck
       /// </summary>
       /// <param name="Value">year</param>
       /// <param name="errMessage"></param>
       /// <param name="aModelParameter"></param>
       /// <returns></returns>
        internal bool ColoradoYearRangeCheck(int Value, ref string errMessage, ModelParameterClass aModelParameter)
        {
            bool valid = false;
            errMessage = "";
            int First = 0;
            int Last = 0; int source = Colorado_Historical_Data_Source;
            valid = RiverHistoricalYearRangeCheck(Value, source, rcColorado, ref First, ref Last);
            if (!valid) errMessage = "Invalid year of " + Value.ToString()+ RiverRangeString(source,First,Last);
            return valid;
        }
        //-----------------------------------------------
       /// <summary>
        /// Colorado_Historical_Extraction_Start_Year_RangeCheck
       /// </summary>
       /// <param name="Year">year to check</param>
       /// <returns>true if year is valid, false otherwise </returns>
        public bool Colorado_Historical_Extraction_Start_Year_RangeCheck(int Year)
        {
            string junk = "";
            return ColoradoYearRangeCheck(Year, ref junk, ParamManager.Model_Parameter(eModelParam.epColorado_Historical_Extraction_Start_Year));
        }
        //-----------------------------------------------
        internal bool SaltVerdeYearRangeCheck(int Value, ref string errMessage, ModelParameterClass aModelParameter)
        {
            bool valid = false;
            errMessage = "";
            int source = SaltVerde_Historical_Data_Source;
            int First = 0;
            int Last = 0;
            valid = RiverHistoricalYearRangeCheck(Value, source, rcSaltVerde, ref First, ref Last);
            if (!valid) errMessage = "Invalid year of " + Value.ToString() + RiverRangeString(source, First, Last);
            return valid;
        }
        //-----------------------------------------------
       /// <summary>
        /// SaltVerde_Historical_Extraction_Start_Year_RangeCheck
       /// </summary>
       /// <param name="Value"></param>
       /// <returns></returns>
        public bool SaltVerde_Historical_Extraction_Start_Year_RangeCheck(int Value)
        {
            string junk = "";
            return SaltVerdeYearRangeCheck(Value, ref junk, ParamManager.Model_Parameter(eModelParam.epSaltVerde_Historical_Extraction_Start_Year));
        }
        //-----------------------------------------------
       //-----------------------------------------------
        /// <summary>
        /// PCTEffluentRangeCheck
        /// </summary>
        /// <param name="Value"></param>
        /// <param name="provider"></param>
        /// <param name="emp"></param>
        /// <returns></returns>
        public bool PCTEffluentRangeCheck(int Value, eProvider provider, eModelParam emp)
        {
            bool test = false;
            string junk = "";
            if (ModelParamClass.valid(emp))
            {
            ModelParameterClass MP = ParamManager.Model_Parameter(emp);
            test = PCTEffluentRangeCheck(Value, provider, ref junk, MP);
            }
            else
                throw new WaterSim_Exception(WS_Strings.wsInvalid_EModelPAram);
            return test;
        }
       //-----------------------------------------------
        internal bool PCTEffluentRangeCheck(int Value, eProvider provider, ref string errMessage, ModelParameterClass MP)
        {
            ProviderIntArray currentvalues = new ProviderIntArray(0);
            currentvalues = MP.ProviderProperty.getvalues();
            int currenttotal =ParamManager.eParamGroupTotal(ParamManager.effluentchecklist, provider);
            int total = currenttotal+Value-currentvalues[(int)provider];
            bool valid = ((Value <= 100) & (Value >= 0) & (total <= 100));
            if (!valid) errMessage = "Total effluent allocated can not exceed 100 (%) " + ParamManager.eParamGroupMessage(ParamManager.effluentchecklist, provider);
            return valid;
        }
        //-----------------------------------------------
       /// <summary>
        /// PCTReclaimedRangeCheck - Check to see of all the PCT reclaimed model paramters do nt eceeed 100% total
       /// </summary>
       /// <param name="Value">int: to be checked if violates over 100% rule</param>
       /// <param name="provider">eProvider: who will set the value</param>
       /// <param name="emp">eModelParam: This is the actual ModelParameter that is doing the check.</param>
       /// <returns>bool True if all values are less than or equal to 100, False of the do not</returns>
        public bool PCTReclaimedRangeCheck(int Value, eProvider provider, eModelParam emp)
        {
            bool test = false;
            string junk = "";
            if (ModelParamClass.valid(emp))
            {
            ModelParameterClass MP = ParamManager.Model_Parameter(emp);
            test = PCTReclaimedRangeCheck(Value, provider, ref junk, MP);
            }
            else
                throw new WaterSim_Exception(WS_Strings.wsInvalid_EModelPAram);
            return test;
        }
        //-----------------------------------------------------------------------
        internal bool PCTReclaimedRangeCheck(int Value, eProvider provider, ref string errMessage, ModelParameterClass MP)
        {
            ProviderIntArray currentvalues = new ProviderIntArray(0);
            currentvalues = MP.ProviderProperty.getvalues();
            int currenttotal = ParamManager.eParamGroupTotal(ParamManager.reclaimedchecklist, provider);
            int total = currenttotal + Value - currentvalues[(int)provider];
            bool valid = ((Value <= 100) & (Value >= 0) & (total <= 100));
            if (!valid) errMessage = "Total reclaimed allocated can not exceed 100 (%) " + ParamManager.eParamGroupMessage(ParamManager.reclaimedchecklist, provider);
            return valid;
        }
       //-----------------------------------------------
       /// <summary>
        /// PCTResComRangeCheck - Checks to see if Residential an Commerical water use PCT does not excede 100.
       /// </summary>
        /// <param name="Value">int: to be checked if violates over 100% rule</param>
        /// <param name="provider">eProvider: who will set the value</param>
        /// <param name="emp">eModelParam: This is the actual ModelParameter that is doing the check.</param>
        /// <returns>bool True if all values are less than or equal to 100, False of the do not</returns>
        public bool PCTResComRangeCheck(int Value, eProvider provider, eModelParam emp)
        {
            bool test = false;
            string junk = "";
            if (ModelParamClass.valid(emp))
            {
            ModelParameterClass MP = ParamManager.Model_Parameter(emp);
            test = ResComPCTRangeCheck(Value, provider, ref junk, MP);
            }
            else
                throw new WaterSim_Exception(WS_Strings.wsInvalid_EModelPAram);
            return test;
        }
       //-----------------------------------------------
        internal bool ResComPCTRangeCheck(int Value, eProvider provider, ref string errMessage, ModelParameterClass MP)
      {
          ProviderIntArray currentvalues = new ProviderIntArray(0);
          currentvalues = MP.ProviderProperty.getvalues();
          int currenttotal = ParamManager.eParamGroupTotal(ParamManager.wateruserchecklist, provider);
          int total = currenttotal + Value - currentvalues[(int)provider];
          bool valid = ((Value<=100)&(Value >= 0) & (total <= 100));
          if (!valid) errMessage = "Total water use can not exceed 100 (%) " + ParamManager.eParamGroupMessage(ParamManager.wateruserchecklist, provider);
           return valid;
         }


        //------------------------------------------------------------------------
        // STATIC MEMBERS AND CONST



        //---------------------------------------------------------
       /// <summary>
        /// IEnumerable for simulationYears()
       /// </summary>
        /// <returns>a year from range of Simulation_Start_Year to Simulation_End_Year</returns>
        public IEnumerable<int> simulationYears()
        {
            for (int i = Simulation_Start_Year; i <= Simulation_End_Year; i++)
            {
                yield return i;
            }
        }
   
        //-------------------------------------------------------------
        
        //============================================================
    }  // end of WaterSim

    
    
 
     
}  // End of WaterSimDCDC
