﻿//      WaterSimDCDC Regional Water Demand and Supply Model Version 5.0

//       A Class the adds Database support to the WaterSimDCDC.WaterSim Class

//       Copyright (C) 2011 , The Arizona Board of Regents
//              on behalf of Arizona State University

//       All rights reserved.

//       Developed by the Decision Center for a Desert City
//       Lead Model Development - David A. Sampson <david.a.sampson@asu.edu>

//       This program is free software: you can redistribute it and/or modify
//       it under the terms of the GNU General Public License version 3 as published by
//       the Free Software Foundation.

//       This program is distributed in the hope that it will be useful,
//       but WITHOUT ANY WARRANTY; without even the implied warranty of
//       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//       GNU General Public License for more details.

//       You should have received a copy of the GNU General Public License
//       along with this program.  If not, please see <http://www.gnu.org/licenses/>.
//
//====================================================================================
using System;
using System.Collections.Generic;

using System.Text;
using System.Data;
using System.Data.OleDb;
using WaterSimDCDC;
/*******************************************
 * WaterSim DB
 * Version 3
 * Keeper Quay
 *   Requires WaterSimDCDC_APIver12, WSAPI_utilities V12
 * ******************************************/
namespace WaterSimDCDC
{
    /// <summary>   WaterSim with database support.  </summary>
    /// <remarks>This class wraps support for database  input and output of Simulation results around the WaterSim simulation methods.</remarks>
    public class WaterSim_DB : WaterSim
    {
        // Class Fields
      

        /// <summary> The database connection </summary>
        protected OleDbConnection FDbConnection;
        /// <summary> Filename of the data base file </summary>
        protected string FDataBaseFilename = "";
        /// <summary> True if database is open </summary>
        protected bool FDbOpen = false;

        bool FisError = false;
        string FErrorMessage = "";

        //AnnualSimulationResults AnnualResults;
        AnnualSimulationResults[] DbRun_AnnualResults;
        SimulationStrings db_fieldnames;
        //--------------------------------------------------------------------
        // Constructor
        /// <summary>
        /// Constructs a WaterSim_DB object.  Creates a new DbConnection and opens the Database if a filename is provided.  Otherwise DBConnection is not created.
        /// </summary>
        /// <param name="TempDirectoryName">Path for temporary directory</param>
        /// <param name="DataDirectoryName">Path to WaterSim data</param>
        /// <param name="DatabaseFilename">Path and filename for the Database to be opened.</param>
        /// <exception cref="WaterSim_DB_Exception"> if unable to open the database.</exception>
        public WaterSim_DB(string TempDirectoryName, string DataDirectoryName, string DatabaseFilename) : base(TempDirectoryName, DataDirectoryName)

        {
            if ((DatabaseFilename!=null)&(DatabaseFilename != "")) OpenDatabase(DatabaseFilename);
        }
        //--------------------------------------------------------------------
        // Dispose

        /// <summary>
        ///     Called by Dispose() and ~WaterSim() Finalizer  Prevents Finalizer from closing model
        ///     files.
        /// </summary>
        /// <param name="disposing">    Finalizer will call with a false. </param>

        protected override void Dispose(bool disposing) 
        {
           base.Dispose(disposing);
           if(disposing) if (FDbConnection != null) FDbConnection.Close();
        } 
        //-----------------------------------------------------------------------------
        // Properties
        /// <summary>
        /// Property 
        /// This is the database connection, can be opened seperately and assigned with this property, must be in the Open state, otherwise and exception is thrown
        /// </summary>
        /// <remarks> OleDbConnection read/write</remarks>
        /// <exception cref="WaterSim_DB_Exception"> if OleDbConnection value passed is not valid or open.</exception>
        /// <exception cref="WaterSim_DB_Exception"> if OleDbConnection value is null.</exception>
        public OleDbConnection DbConnection
        {   get {return FDbConnection; }
            set
            {
                if (value != null)
                {
                    if (value.State == ConnectionState.Open)
                    {
                        FDbConnection = value;
                        FDbOpen = true;
                    }
                    else
                    {
                        throw new WaterSim_DB_Exception(WS_Strings.wsUableOpenDbConnection);
                    }
                }
                else
                {
                    throw new WaterSim_DB_Exception(WS_Strings.wsdbConNull);
                }
            }
        }
        //-----------------------------------------------------------------------------
       /// <summary>
       /// Property
       /// Datbase Filename for DbConnection.  If different than current database (or database not open) closes DbConnection and opens new using this filename.
       /// </summary>
       /// <remarks> string read/write</remarks>
        public string DatabaseFilename
        {   get { return FDataBaseFilename; }
            set 
            {
                OpenDatabase(value);
           }
        }
        //-----------------------------------------------------------------------------
        /// <summary>
        /// Property
        /// Returns true if the Database is open
        /// </summary>
        /// <value> bool read only</value>
        public bool DataBaseOpen
        { get { return FDbOpen; } }
        //-----------------------------------------------------------------------------
//        public List<string> Tablenames
//        { get { return FTablenames; } }
        //-----------------------------------------------------------------------------
       /// <summary>
       /// Property
       /// Returns true if t
       /// </summary>
        public bool isError
        { get { return FisError; } }
        
        #region DBUtils
        //-----------------------------------------------------------------------------

        /// <summary>   Opens a database. </summary>
        /// <exception cref="Exception">if unable to open database </exception>
        ///<remarks>This uses dbConnection to open a database.  Uses the following Connection string
        ///			<code>
       ///         string Connectstr;
       ///         Connectstr = "Provider=Microsoft.JET.OLEDB.4.0; data source=" + Filename + ";User Id=admin;Password=;";
       ///         FDbConnection = new OleDbConnection(Connectstr);
       ///         FDbConnection.Open();
       ///         FDataBaseFilename = DatabaseFilename;
       ///         FDbOpen = true;
        ///			</code></remarks>
        /// <param name="Filename"> Filename of the database. </param>

        protected virtual void OpenDatabase(string Filename)
        {
            try
            {
                string Connectstr;
                Connectstr = "Provider=Microsoft.JET.OLEDB.4.0; data source=" + Filename +
                ";User Id=admin;Password=;";
                FDbConnection = new OleDbConnection(Connectstr);
                FDbConnection.Open();
                FDataBaseFilename = DatabaseFilename;
                FDbOpen = true;
            }
            catch
            {
                FDbOpen = false;
                throw new Exception("WaterSim DB Error: Unable  to open Database " + DatabaseFilename);
            }
        }
        //-----------------------------------------------------------------------------
        internal void SetError(string aMessage)
        {
            FErrorMessage = aMessage;
            FisError = true;
        }
        //-----------------------------------------------------------------------------

        /// <summary>   Gets the error message. </summary>
        /// <returns>   An error message if isError is true, otherwise returns "". </returns>
        /// <value> ErrorMessage</value>
        public string GetErrorMessage()
        {
            string temp = FErrorMessage;
            FisError = false;
            FErrorMessage = "";
            return temp;
        }
        //-----------------------------------------------------------------------------
        // Required Fields

        /// <summary> Number of required fields for WaterSim_DB DataTables</summary>
        protected const int NumberOfRequiredFields = 5;

        /// <summary> Scenario name db fieldname</summary>
        /// <value>a fieldname string const</value>
        /// <remarks> This column will contain the scenario name.  This is one of five fields required to be in a WaterSim_DB valid datatable</remarks>
        /// <seealso cref="Valid_DataTable"/>
        public const string rdbfScenarioName = "SCN_NAME";

        /// <summary> Simulation year db fieldname</summary>
        /// <value>a fieldname string const</value>
        /// <remarks>This column will contain the year of the scenario.  This is one of five fields required to be in a WaterSim_DB valid datatable</remarks>
        /// <seealso cref="Valid_DataTable"/>
        public const string rdbfSimYear = "SIMYEAR";

        /// <summary> Provider label db fieldname</summary>
        /// <value>a fieldname string const</value>
        /// <remarks> This is one of five fields required to be in a WaterSim_DB valid datatable</remarks>
        /// <seealso cref="Valid_DataTable"/>
        public const string rdbfProviderLabel = "PRVDLABEL";

        /// <summary> Provider code db fieldname</summary>
        /// <value>a fieldname string const</value>
        /// <remarks> This is a column for the provider fieldname values.  Each row of each year will have a different provider firldname in this column.  This is one of five fields required to be in a WaterSim_DB valid datatable</remarks>
        /// <seealso cref="Valid_DataTable"/>
        /// <seealso cref="ProviderClass"/>
        /// <seealso cref="ProviderClass.FieldName"/>
        public const string rdbfProviderCode = "PRVDCODE";

        /// <summary> Primary key db fieldname </summary>
        /// <value>a fieldname string const</value>
        /// <remarks>This is the primary key, a unique long integer value. This is one of five fields required to be in a WaterSim_DB valid datatable</remarks>
        /// <seealso cref="Valid_DataTable"/>
        public const string rdbfID = "ID";

        /// <summary> The required fields for a WaterSim_DB DataTable</summary>
        protected string[] RequiredFields = new string[NumberOfRequiredFields] { rdbfID, rdbfScenarioName, rdbfSimYear, rdbfProviderLabel, rdbfProviderCode };

        /// <summary> The required SQL FIeld defines </summary>
        protected string[] RequiredDefines = new string[NumberOfRequiredFields] { "COUNTER PRIMARY KEY", "CHAR(50)", "INT", "CHAR(25)", "CHAR(5)" };
        //-----------------------------------------------------------------------------

        /// <summary>   SQL data define. </summary>
        /// <remarks>Creates SQL syntax for the field list portion of a CREATE TABLE command ie CREATETABLE [TABLENAME] (FIELDLIST)</remarks>
        /// <returns>  SQL syntax string </returns>

        protected string DataDefine()
        {
            string temp = RequiredFields[0]+" "+RequiredDefines[0];
            for (int i=1;i<NumberOfRequiredFields;i++)
               temp += " , "+RequiredFields[i]+" "+RequiredDefines[i];

            ModelParameterClass MP;
            foreach (eModelParam eP in _pm.ModelParameters())
            {
                MP = _pm.Model_Parameter(eP);
                temp +=  ", " + MP.Fieldname + " INT";
            }
            return temp;
        }
        //-----------------------------------------------------------------------------
        /// <summary>
        /// Creates a new table in the open database. Must be a unique name. Creates all possible fields.
        /// </summary>
        /// <param name="NewTablename">tablename</param>
        /// <returns>true if created, otherwise false.  If false check isError and GetErrorMessage()</returns>
        /// <seealso cref="isError"/>
        /// <seealso cref="GetErrorMessage"/>
        public bool CreateNewDataTable(string NewTablename)
        {
            List<string> FTablenames = new List<string>();
            Boolean IsDone = false;

            // Get Table Names from the Schema
            DataTable REGIONSchema = new DataTable();
            REGIONSchema = FDbConnection.GetSchema("TABLES");
            for (int i = 0; i < REGIONSchema.Rows.Count; i++)
            {
                if (REGIONSchema.Rows[i]["TABLE_TYPE"].ToString() == "TABLE")
                {
                    FTablenames.Add(REGIONSchema.Rows[i]["TABLE_NAME"].ToString());
                }
            }
            // check and see if it is in the tablename list
            if (FTablenames.Find(delegate(string value) { if (value == NewTablename) return true; else return false; }) != NewTablename)
           {
               string tc = "CREATE TABLE ";
               string dataDef = DataDefine();
               OleDbCommand cmd = FDbConnection.CreateCommand();

               if (NewTablename.Length > 0)
               {
                    tc +=  '[' + NewTablename + ']' + " (" + dataDef + ")";
                   cmd.CommandText = tc;
                   try
                   {
                       cmd.ExecuteNonQuery();
                       IsDone = true;
                   }
                   catch (Exception ex)
                   {
                       SetError(ex.Message);
                       cmd.Dispose();
                   }
               }
           }
           else
           {
               SetError(NewTablename + " is existing table, must be unique tablename");
           }
            return IsDone;
        }
        //-----------------------------------------------------------------------------
        /// <summary>
        /// Check if Valid WaterSIM_DB datatable
        /// </summary>
        /// <param name="Target">a DataTable to check</param>
        /// <returns>true if valid otherwise false</returns>
        /// <remarks>Checks if the DataTable has the required fields to be a valid WaterSim data table.  Five fields are required and constants are provided for these five fields,
        ///       rdbfID,  rdbfScenarioName,  rdbfSimYear ,  rdbfProviderLabel, and  rdbfProviderCode.  Below is SQL command syntax for creating such a table.
        ///      <code> 
        ///           OleDbConnection dbConnect = new OleDbConnection(MyConnectionstring);
        ///           OleDbCommand cmd = dbConnect.CreateCommand();
        ///            cmd.CommandText = "CREATE TABLE [MYTABLENAME] ("+
        ///                 rdbfID                        + "COUNTER PRIMARY KEY, "+ 
         ///                rdbfScenarioName    + " CHAR(50), "+
         ///                rdbfSimYear              + " INT, "+
         ///                rdbfProviderLabel      + " CHAR(25), "+
         ///                rdbfProviderCode      + " CHAR(5)  )"
         ///            cmd.ExecuteNonQuery();
         ///            </code>
         ///            </remarks>
 
        public bool Valid_DataTable(DataTable Target)
        {
            bool isValid = false;

            for (int i = 0; i < NumberOfRequiredFields; i++)
            {
                isValid = Target.Columns.Contains(RequiredFields[i]);
                if (!isValid) break;
            }

            return isValid;
        }
        //------------------------------------------------------------------------------

        /// <summary>   Loads a scenario from a database. </summary>
        /// <remarks>This method will load Input Model Parameters (BaseInput and ProviderInput) into a SimulationInputs struct.  
        /// 		 The Datatable must be a valid WaterSim_DB datatable (see Valid_DataTable  
        /// 		 However fields for all input parameters do not need to be in the Datatable.  This method will load whatever ModelParameter.Fieldname 
        /// 		 fields it finds.  It does not set values in sInputs struct for Model Parameters it finds.  It is good practice to fill
        /// 		 the sInputs struct with the models current input values before passing it to this method.  See example</remarks>
        /// <param name="ScenarioName"> Name of the scenario, if empty (="") then the first scenario found is used. </param>
        /// <param name="year">         The year. If 0 then the first year found is used</param>
        /// <param name="Source">       Datatable Source. </param>
        /// <param name="sInputs">      [in,out] SimulationInputs. </param>
        ///
        /// <returns>   true if it succeeds, false if it fails.  If false check isError and GetErrorMessage() </returns>
        /// <exception cref="WaterSim_Exception">when Source datatable is not a valid Watersim_DB datatable. </exception>
        /// <example><code>
        ///          try
        ///                {
        ///                    OleDbDataAdapter TableAdapter = new OleDbDataAdapter("Select * from " + '[' + tablename + ']', WSim.DbConnection);
        ///                    TableAdapter.Fill(DT);
        ///                    TableAdapter.Dispose();
        ///                    SimulationInputs SI = WSim.ParamManager.GetSimulationInputs();
        ///                    if (WSim.LoadScenario_DB("", 0, DT, ref SI))
        ///                     {
        ///                        if (WSim.ParamManager.SetSimulationInputs(SI))
        ///                        {
        ///                            WS_InputPanel.Refresh_Inputs();
        ///                        }
        ///                    }
        ///                }
        ///                catch (Exception e)
        ///                {
        ///                    MessageBox.Show("Whoops " + e.Message);
        ///                }
        /// 		 </code></example>
        /// <seealso cref="isError"/>
        /// <seealso cref="GetErrorMessage"/>
        /// <seealso cref="Valid_DataTable"/>
        /// <seealso cref="SimulationInputs"/>
        /// <seealso cref="ParameterManagerClass.GetSimulationInputs"/>
        /// <seealso cref="ParameterManagerClass.SetSimulationInputs"/>
        
        public bool LoadScenario_DB(string ScenarioName, int year, DataTable Source, ref SimulationInputs sInputs)
        {
            bool loadSuccess = false;
            string yearstr = "";
            int tempyr = 0;
            string tempMPvaluestr = "";
            int tempMPvalue = 0;
            string epCodestr = "";
            eProvider ep;
            int index = 0;
            if (!Valid_DataTable(Source)) throw new WaterSim_DB_Exception(WS_Strings.wsInvalidDataTable);

            try
            {
                // OK, first if scenarioname has not been picked then get the first one
                if (ScenarioName == "")
                {
                    DataRow DR = Source.Rows[0];
                    ScenarioName = DR[rdbfScenarioName].ToString();
                }
                // OK, if year has not been picked then get the first one
                if (year == 0)
                {
                    DataRow DR = Source.Rows[0];
                    yearstr  = DR[rdbfSimYear].ToString();
                    year = Convert.ToInt16(yearstr);
                }
                // set this much
                sInputs.ScenarioName = ScenarioName;
                // uppercase scenarioname
                ScenarioName = ScenarioName.ToUpper();
                // only need to do base inputs once, so can use the first row found.  This is set true once first row is found
                bool FirstRowFound = false;
                // go thru table row by row looking for scenario name and year
                foreach (DataRow DR in Source.Rows)
                {
                    yearstr = DR[rdbfSimYear].ToString();
                    tempyr = Convert.ToInt16(yearstr);
                    if ((DR[rdbfScenarioName].ToString().ToUpper() == ScenarioName) & (tempyr == year))
                    {
                        // found a row, get the provider code
                        epCodestr = DR[rdbfProviderCode].ToString();
                        // test it
                        if (ProviderClass.valid(epCodestr))
                        {
                            // Get Provider
                            ep = ProviderClass.provider(epCodestr);
                            // Now get the base dataif first row otherwise skip
                            if (!FirstRowFound)
                            {
                                index = 0;
                                foreach (ModelParameterClass MP in ParamManager.BaseInputs())
                                {
                                    if (Source.Columns.Contains(MP.Fieldname))
                                    {
                                        // found ModelPAram in Table
                                        tempMPvaluestr = DR[MP.Fieldname].ToString();
                                        tempMPvalue = Convert.ToInt16(tempMPvaluestr);
                                        sInputs.BaseInput[index] = tempMPvalue;
                                    }
                                    index++;
                                }
                                FirstRowFound = true;
                            }
                            index = 0;
                            foreach (ModelParameterClass MP in ParamManager.ProviderInputs())
                            {
                                if (Source.Columns.Contains(MP.Fieldname))
                                {
                                    // found ModelPAram in Table
                                    tempMPvaluestr = DR[MP.Fieldname].ToString();
                                    tempMPvalue = Convert.ToInt16(tempMPvaluestr);
                                    sInputs.ProviderInput.Values[index][ep] = tempMPvalue;
                                }
                                index++;
                            }
                        }
                    }
                }
                loadSuccess = true;
            }
            catch(Exception e)
            {
                SetError(e.Message);
            }
            return loadSuccess;
        }

        #endregion
        
        #region SimulationDB
        //-----------------------------------------------------------------------------
        // Variables used in the Simulation_DB routines
       // OleDbConnection _dbConnect;
        DataTable _dbTable;
        string _DataTableName = "";
        OleDbDataAdapter RawTableAdapter;

        internal bool _Simulation_DB_initialized = false;
        internal bool _inDBRun = false;
        internal int _SimulationDB_Years = 0;
        internal string _ScnName = "";
        //-----------------------------------------------------------------------------        
        internal void fetchData(int year)
        {   
            int datai = 0;
            int index = year - Sim_StartYear; // changed 7 26 11 _SimulationDB_Start_Year; 
            if ((index >= 0) & (index < _SimulationDB_Years))
            {
                datai = 0;
                foreach (ModelParameterClass MP in ParamManager.BaseOutputs())
                {
                    DbRun_AnnualResults[index].BaseOutput[datai] = MP.GetInt();
                    datai++;
                }
                datai = 0;
                foreach (ModelParameterClass MP in ParamManager.BaseInputs())
                {
                    DbRun_AnnualResults[index].BaseInput[datai] = MP.GetInt();
                    datai++;
                }
                datai = 0;
                foreach (ModelParameterClass MP in ParamManager.ProviderOutputs())
                {
                    DbRun_AnnualResults[index].ProviderOutput.Values[datai].Values = MP.GetIntArray();
                    datai++;
                }
                datai = 0;
                foreach (ModelParameterClass MP in ParamManager.ProviderInputs())
                {
                    DbRun_AnnualResults[index].ProviderInput.Values[datai].Values = MP.GetIntArray();
                    datai++;
                }
                DbRun_AnnualResults[index].year = year;
            }
        }
        //-----------------------------------------------------------------------------
        internal void initializeData(int Years)
        {
            if (DbRun_AnnualResults == null)
            {
                DbRun_AnnualResults = new AnnualSimulationResults[Years];
                for (int i = 0; i < Years; i++)
                {
                    DbRun_AnnualResults[i] = new AnnualSimulationResults(ParamManager.NumberOfParameters(modelParamtype.mptOutputBase),
                                                    ParamManager.NumberOfParameters(modelParamtype.mptOutputProvider),
                                                    ParamManager.NumberOfParameters(modelParamtype.mptInputBase),
                                                    ParamManager.NumberOfParameters(modelParamtype.mptInputProvider)); 
                    //for(int j=0;j<DbRun_AnnualResults[i].BaseInput.Length;j++)
                    //    DbRun_AnnualResults[i].BaseInput[j] = 0;
                    //for (int j = 0; j < DbRun_AnnualResults[i].ProviderInput.Length; j++)
                    //    DbRun_AnnualResults[i].ProviderInput[j] = 0;
                    //for (int j = 0; j < DbRun_AnnualResults[i].BaseInput.Length; j++)
                    //    DbRun_AnnualResults[i].BaseInput[j] = 0;
                    //for (int j = 0; j < DbRun_AnnualResults[i].ProviderInput.Length; j++)
                    //    DbRun_AnnualResults[i].ProviderInput[j] = 0;

                }
             }
        }
        //-----------------------------------------------------------------------------        
        internal static bool FindStringNoCase(string value, string[] values, ref int index)
        {
            index = -1;
            bool found = false;
            for (int i = 0; i < values.Length; i++)
            {
                found = (values[i].ToUpper() == value.ToUpper());
                if (found) 
                {
                    index = i;
                    break;
                }
            }
            return found;
        }
        //-----------------------------------------------------------------------------
        internal void initilaizeFieldStrings()
        {
            // load fieldnames
            db_fieldnames = new SimulationStrings(ParamManager.NumberOfParameters(modelParamtype.mptOutputBase),
                                                    ParamManager.NumberOfParameters(modelParamtype.mptOutputProvider),
                                                    ParamManager.NumberOfParameters(modelParamtype.mptInputBase),
                                                    ParamManager.NumberOfParameters(modelParamtype.mptInputProvider));

            // BaseOutputs
            int index = 0;
            foreach (ModelParameterClass MP in ParamManager.BaseOutputs())
            {
                db_fieldnames.BaseOutput[index] = MP.Fieldname;
                index++;
            }
            // Base Inputs
            index = 0;
            foreach (ModelParameterClass MP in ParamManager.BaseInputs())
            {
                db_fieldnames.BaseInput[index] = MP.Fieldname;
                index++;
            }
            // Provider Outputs
            index = 0;
            foreach (ModelParameterClass MP in ParamManager.ProviderOutputs())
            {
                db_fieldnames.ProviderOutput[index] = MP.Fieldname;
                index++;
            }
            // Provider Inputs
            index = 0;
            foreach (ModelParameterClass MP in ParamManager.ProviderInputs())
            {
                db_fieldnames.ProviderInput[index] = MP.Fieldname;
                index++;
            }
        }
        //-----------------------------------------------------------------------------
        internal bool dbAppendAnnualData ()
        {
            bool appenddone = false;
            string ProviderCode = "";
            string ProviderLabel = "";
            string yearStr = "";
            int pindex = 0;
            
            DataRow DR;
            try
            {
                for (int yeari = 0; yeari < _SimulationDB_Years; yeari++)
                {
                    foreach (eProvider ep in ProviderClass.providers())
                    {
                        pindex = ProviderClass.index(ep);
                        ProviderCode = ProviderClass.FieldName(ep);
                        ProviderLabel = ProviderClass.Label(ep);
                        if (ProviderLabel.Length > 25) ProviderLabel = ProviderLabel.Substring(0, 25);
                        yearStr = DbRun_AnnualResults[yeari].year.ToString();

                        DR = _dbTable.NewRow();
                        //        protected string[] RequiredFields = new string[NumberOfRequiredFields] { "ID", "SCN_NAME", "SIMYEAR", "PRVDLABEL", "PRVDCODE" };

                        DR["SCN_NAME"] = _ScnName;
                        DR["SIMYEAR"] = yearStr;
                        DR["PRVDLABEL"] = ProviderLabel;
                        DR["PRVDCODE"] = ProviderCode;

                        // BaseOutput
                        string wsfield = "";
                        string outstr = "";
                        for(int i=0; i<db_fieldnames.BaseOutput.Length;i++)
                        {
                            wsfield = db_fieldnames.BaseOutput[i];
                            if (_dbTable.Columns.Contains(db_fieldnames.BaseOutput[i]))
                            {
                                DR[db_fieldnames.BaseOutput[i]] = DbRun_AnnualResults[yeari].BaseOutput[i].ToString();
                            }
                        }
                        // ProviderOutput
                        for(int i=0; i<db_fieldnames.ProviderOutput.Length;i++)
                        {
                            wsfield = db_fieldnames.ProviderOutput[i];
                            if (_dbTable.Columns.Contains(db_fieldnames.ProviderOutput[i]))
                            {
                                outstr = DbRun_AnnualResults[yeari].ProviderOutput.Values[i].Values[pindex].ToString();
                                DR[db_fieldnames.ProviderOutput[i]] = DbRun_AnnualResults[yeari].ProviderOutput.Values[i].Values[pindex].ToString();
                            }
                        }
                        // BaseInput
                        for(int i=0; i<db_fieldnames.BaseInput.Length;i++)
                        {
                            wsfield = db_fieldnames.BaseInput[i];
                            if (_dbTable.Columns.Contains(db_fieldnames.BaseInput[i]))
                            {
                                DR[db_fieldnames.BaseInput[i]] = DbRun_AnnualResults[yeari].BaseInput[i].ToString();
                            }
                        }
                        // ProviderInput
                        for(int i=0; i<db_fieldnames.ProviderInput.Length;i++)
                        {
                            wsfield = db_fieldnames.ProviderInput[i];
                            if (_dbTable.Columns.Contains(db_fieldnames.ProviderInput[i]))
                            {
                                DR[db_fieldnames.ProviderInput[i]] = DbRun_AnnualResults[yeari].ProviderInput.Values[i].Values[pindex].ToString();
                            }
                        }
                        
                        _dbTable.Rows.Add(DR);

                    }

                }
                return appenddone = true;
            }
            catch (Exception e)
            {
                SetError("DataTable append failed Code " + e.Message);
                return appenddone;
            }
        }
        //-----------------------------------------------------------------------------
        /// <summary>
        /// Initializes a DB Simulation.  Opens the datatable using DbConnection and datatablename,  results will be written to temp DataTable and then updated in the Database.
        /// </summary>
        /// <param name="aDataTablename">a tablename to use in the database for output</param>
        /// <param name="Scenario_Name">the name of the scenario to label the output</param>
        /// <remarks>The scenario name is important becuase it can be used to retrieve the Simulation inputs from the datatable later.</remarks>
        /// <returns> true if datatable was opended and simulation initialized, false otherwise, check isError and GetErrorMessage()</returns>
        /// <exception cref="WaterSim_DB_Exception">if data base is not open</exception>
        /// <seealso cref="LoadScenario_DB"/>
        /// <seealso cref="isError"/>
        /// <seealso cref="GetErrorMessage"/>
        public bool Simulation_Initialize(string aDataTablename, string Scenario_Name)
        {
            bool dbdone = false;
            if (!FDbOpen) throw new WaterSim_DB_Exception("Database must be open to initialize simulation");
            // ok load the table
            // Ok fill the data table with all teh relvant fields
            try
            {
                RawTableAdapter = new OleDbDataAdapter("Select * from " + '[' + aDataTablename + ']', FDbConnection);
                RawTableAdapter.Fill(_dbTable);
                dbdone = Simulation_Initialize(_dbTable, Scenario_Name);
            }
            catch (Exception e)
            {
                SetError("Data Table Open Failed: " + e.Message);
            }
            return dbdone;
        }
        //-----------------------------------------------------------------------------
        /// <summary>
        /// Initializes a DB Simulation. results will be appended to the specified DataTable
        /// </summary>
        /// <param name="aDataTable">a DataTable used for output</param>
        /// <param name="Scenario_Name">the name of the scenario to label the output</param>
        /// <remarks>The datatable must exist in the dbConnection database.  The scenario name is important becuase it can be used to retrieve the Simulation inputs from the datatable later.</remarks>
        /// <returns> true if datatable was opended and simulation initialized, false otherwise, check isError and GetErrorMessage()</returns>
        /// <seealso cref="LoadScenario_DB"/>
        /// <seealso cref="isError"/>
        /// <seealso cref="GetErrorMessage"/>

        public bool Simulation_Initialize(DataTable aDataTable, string Scenario_Name)
        {
            // Verify Table
            if (!Valid_DataTable(aDataTable))
            {
                SetError("WaterSim DB Error: Invalid Data Table Structure, Missing required fields");
                return false;
            }
            else
            {
                _dbTable = aDataTable;
                _ScnName = Scenario_Name;
                // nul this out, just use DataTable, no table update
                _DataTableName = "";
                // initialize the field strings
                initilaizeFieldStrings();

                // initialize the Simulation
                Simulation_Initialize();
                _Simulation_DB_initialized = true;
                _inDBRun = false;

                return true;
            }
        }
        //-----------------------------------------------------------------------------
        bool _inDBSimulationNextYear = false;

        /// <summary>
        /// Runs the next (or first) year of the Simulation.  If initialized with one of the DB Simulation_Initilaize() overloads, will append data to the datatable, otehrwise just runs the model one year.
        /// </summary>
        /// <returns>the year of Simulation</returns>
        public override int Simulation_NextYear()
        {
            int year = 0;
            if (!_inDBSimulationNextYear)
            {
                _inDBSimulationNextYear = true;
                if (_Simulation_DB_initialized)
                {
                    // run s=imulation
                    if (!_inDBRun)
                    {
                        // Set years for Simulation
                        _SimulationDB_Years = (Simulation_End_Year - Simulation_Start_Year) + 1;
                        //_SimulationDB_Start_Year = Simulation_Start_Year;
                        // initialize data structure
                        initializeData(_SimulationDB_Years);
                        // Ok exist and start running years
                        _inDBRun = true;
                    }
                    year = base.Simulation_NextYear();
                    //Fetch Data
                    fetchData(year);
                }
                // Did not use DB version of initialize so just run normal
                else
                    year = base.Simulation_NextYear();
            }
            _inDBSimulationNextYear = false;
            return year;
        }
       //-----------------------------------------------------------------------------
      
        bool _inSimulationAllYears = false;
        /// <summary>
        /// Runs all the years of the Simulation based on Simulation_STart_Year and Simulation_End_year, If initialized with one of the DB Simulation_Initilaize() overloads, will append data to the datatable, otehrwise just runs the model all years.
        /// </summary>
        public override void Simulation_AllYears()
        {
            if (!_inSimulationAllYears)
            {
                _inSimulationAllYears = true;
                foreach (int year in simulationYears())
                {
                    Simulation_NextYear();  // calls DBSimulation which test if in a DB Simulation otherwise just calls root
                }
                _inSimulationAllYears = false;

            }
        }
         //-----------------------------------------------------------------------------
        /// <summary>
        /// Stops the Simulation.  If initialized with one of the DB Simulation_Initilaize() overloads, will update the datatable in the database.
        /// </summary>
        public override void Simulation_Stop()
        {
            // stop simulation
            base.Simulation_Stop();
            // if DB initiualized close it down
            if (_Simulation_DB_initialized)
            {
                _inDBRun = false;
                _Simulation_DB_initialized = false;
               // Now append this data to datatable
                if (dbAppendAnnualData())
                {
                    // check if should uopdate database
                    if (_DataTableName != "")
                    {
                        OleDbCommandBuilder TableUpdateBuilder = new OleDbCommandBuilder(RawTableAdapter);
                        RawTableAdapter.Update(_dbTable);
                    }
                }
                else
                {
                    if (isError)
                    {
                        throw new WaterSim_DB_Exception(" Whille updating database " + GetErrorMessage());
                    }
                }
            }
        }
      
    

    }// End of WaterSim_DB
    internal class WaterSim_DB_Exception : WaterSim_Exception
    {
        new const string Pre = "WaterSim DB Error: ";
        internal WaterSim_DB_Exception(string message) : base(message) { }
        internal WaterSim_DB_Exception(int index) : base(index) { }
    }
#endregion

}
